/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill;

import anthill.storage.*;

import java.io.*;

/**
 *  Interface <CODE>AntView</CODE> contains the set of methods which are
 *  accessible to ants when visiting nests. These methods enable ants to 
 *  cooperate with nests in a controlled way. For example, ants may
 *  retrieve resources from or add resources to a nest; obtain routing 
 *  information related to other nests; move to other nests, and so on.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision$
 */
public interface AntView
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for requesting general information
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the identifier of the nest hosting the
   *  ant using this <CODE>AntView</CODE>.
   */
  public NestId getNestId();

  
  /**
   *  Returns the identifier of the last nest visisted by this
   *  ant.
   */
  public NestId getLastVisited();
  
  /**
   */
  public int getTTL();

  /**
   *  Returns the ant which is associated to this <CODE>AntView</CODE>. Normally not
   *  used by ants, as they may access the same information using
   *  <CODE>this</CODE>.
   */
  public Ant getAnt();

  
  /**
   *  Creates a new ant identifier.
   */
  public AntId createAntId();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for ant movement and creation
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a new ant to this nest. For example, this method
   *  could be used to create a backward ant for returning
   *  information to the requestor nest.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *  
   *  @param ant the ant to be added to the current nest
   */
  public void addAnt(Ant ant);

  /**
   *  Method invoked by an ant in order to be sent to another 
   *  <CODE>Nest</CODE>. It is possible that particular 
   *  <CODE>AntView</CODE> implementation limit the operations 
   *  which can be performed after a move operation. For example, 
   *  movement to more than one nest (i.e., duplication) may not
   *  be enabled.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param id the identifier of the destination nest
   *  @throws IOException raised if the move operation fails,
   *    for example due to the unreachability of the nest
   */
  public void move(NestId id)
    throws IOException;
  
  /**
   *  Store information about the last nest visited by the ant
   *  associated with this <CODE>AntView</CODE> on the current 
   *  nest.  This information is constituted of a pair
   *  (<CODE>antId</CODE>, <CODE>nestId</CODE>) where 
   *  <CODE>antId</CODE> is the identifier of the ant and
   *  <CODE>nestId</CODE> is the identifier of the nest from
   *  which the ant has arrived. During the return trip, this
   *  information will be used to obtain the identifier of the 
   *  previous nest in the incoming path.
   *  
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *  
   *  @return <CODE>true</CODE> if this is the first visit of this ant
   *    to this nest (and thus, the information about the previous nest
   *    has been stored successfully), or <CODE>false</CODE> if this
   *    ant has already visited the nest (and thus, the inforamtion about
   *    the previous nest could not be stored).
   */
  public boolean storeReturnNest(AntId aid, NestId nid);

  /**
   *  Returns the identifier of the previous nest in the incoming path,
   *  as stored through during the forward trip through method  
   *  <CODE>storePreviousNest()</CODE>. If the current nest is the requestor
   *  nest (i.e., the nest which has created this ant), the method returns
   *  <CODE>null</CODE>.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *  
   *  @param remove true if the association (ant id, nest id)
   *    should be removed after retrieval
   *  @throws IllegalStateException raised if no information about the previous
   *    nest visited by this ant is present (for example, due to expiration).
   */
  public NestId getReturnNest(AntId aid, boolean remove);

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for neighbour management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Ants invoke this method to add the identifier of a new neighbour
   *  nest to the set of neighbours known to the current nest. Depending
   *  on the implementation, this operation may require the establishment
   *  of a connection to the new nest. In this case, an <CODE>IOException</CODE> 
   *  may be raised in the case a connection cannot be established.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *  
   *  @throws IOException if a connection with the nest cannot be established.
   */
  public void addNeighbour(NestId id)
    throws IOException;

  /**
   *  Ants invoke this method to remove a neighbor nest from the list of
   *  neighbors known to this nest. Depending on the implementation, this 
   *  operation may require the closure of a previously created nest connection.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   */
  public void removeNeighbour(NestId id);
  
  /**
   *  Ants invoke this method to obtain the list of nest neighbors
   *  known to the current nest.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   */
  public Neighbour[] getNeighbours();

    
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for storage management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return a storage object identified by the given name.
   */
  public Storage getStorage(String name)
    throws StorageNotFoundException;
  
  
  /**
   *  Get the pheromone information object related to this ant.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   */
  public Object getPheromone();

  /**
   *  Set the pheromone information object related to this ant.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @throws CapacityException if the nest cannot store the
   *    the pheromone object.
   */
  public void setPheromone(Object pheromone);

  
  /**
   *  Performs the specified request on the local resource
   *  storage and returns the resources satisfying the
   *  request. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param request The request to be performed
   */
  public Resource[] getResources(Object request)
    throws StorageNotFoundException;

  /**
   *  Performs the specified request on the local resource
   *  storage and returns the resources satisfying the
   *  request. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param request The request to be performed
   */
  public Resource getResource(Object key)
    throws StorageNotFoundException;
  
  /**
   *  Add a resource to the nest. This method enables
   *  ants to move resources from one nest to another. 
   *  <p>
   *  Depending on the implementation of <CODE>Resource</CODE> 
   *  and <CODE>Nest</CODE>, adding a resource to a nest through 
   *  an AntView may result to actual copy the content of the 
   *  resource, or to simply insert a reference to the actual 
   *  position of the resource itself. 
   *  <p>
   *  This method may be invoked more than once from an ant, in 
   *  order to insert more resources. Nevertheless, the actual 
   *  insertion of a resource in a nest depends on the policies 
   *  of the particular <CODE>Nest</CODE> implementation. For 
   *  example, a nest may limit the number of resources originated 
   *  by  another nest, in order to avoid "spamming".
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param resource resource to be added
   */
  public void addResource(Resource resource)
    throws CapacityException, StorageNotFoundException;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for resource management
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  This method is invoked by ants to notify the nest that
   *  the ant associated to this <CODE>AntView</CODE>, created to satisfy
   *  a request, has found the specified set of resources.
   *
   *  @param resources  the resources satisfying the request
   */
  public boolean returnResources(RequestId rid, Resource[] resources);
  
  
  
  /** Read from statistics */
  // FIXME Add comments
  // FIXME Add other methods (?)
  public int intValue(String name);
  
  
  
} // END AntView
