/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill;

/**
 *  Class <CODE>Neighbour</CODE> contains the information related
 *  to another nest related to another nest. The information
 *  stored in <CODE>Neighbour</CODE> objects include the identifier
 *  of the neighbour nest, the communication cost and the reachability
 *  status.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class Neighbour
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Identifier of the nest neighbour */
  private NestId id;
  
  /** Communication cost with this neighbour */
  private float cost;

  /** True if this neighbour is currently reachable */
  private boolean reachable;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a new neighbour with the given identifier.
   *  
   *  @param id identifier of the neighbour nest
   */
  public Neighbour(NestId id)
  {
    this(id, 0, true);
  }
  
  /**
   *  Constructs a new neighbour with the given identifier and with
   *  the specified additional parameters.
   *  
   *  @param id identifier of the neighbour nest
   *  @param cost the cost of the connection
   *  @param reachable true if the neighbor nest is reachable
   */
  public Neighbour(NestId id, float cost, boolean reachable)
  {
    this.id = id;
    this.cost = cost;
    this.reachable = reachable;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the nest identifier of the neighbour 
   */
  public NestId getId()
  {
    return id;
  }
  
  
  /**
   *  Returns a float value representing the cost of
   *  communication with this neighbor.
   */
  public float getCost()
  {
    return cost;
  }

  
  /**
   *  Returns true if this neghbour is considered reachable.
   */
  public boolean isReachable()
  {
    return reachable;
  }
  
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Check whether this <CODE>LocalNid</CODE> is equal to the specified object.
   */
  public boolean equals(Object obj)
  {
    if (obj != null && obj instanceof Neighbour) {
      Neighbour neighbour = (Neighbour) obj;
      return (this.id == neighbour.id);
    } else {
      return false;
    }
  }

  /**
   *  Returns a hash code value for the object.
   */
  public int hashCode()
  {
    return id.hashCode();
  }
  
  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(": ");
    buf.append(id.toString());
    buf.append(", ");
    buf.append(cost);
    buf.append(", ");
    buf.append(reachable);
    buf.append("]");
    return buf.toString();
  }
  
  
} // END Neighbour

