/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antnet;

import anthill.*;
import anthill.util.*;
import anthill.nest.*;

import java.security.*;

/**
 *  Class <CODE>AntIdImpl</CODE> is the implementation of <CODE>AntId</CODE> to be used in 
 *  a deployment environment. These pseudo-unique identifiers are constituted by a sequence 
 *  of bytes generated randomly through <CODE>java.security.SecureRandom</CODE>. Two 
 *  identifiers are equal if they contains the same sequence of bytes.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class AntIdImpl
implements AntId
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Number of random bytes in the identifier */
  public static int SIZE = 20;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Counter used to generate unique identifiers */
  private static SecureRandom random;
  
  static {
    try {
      random = SecureRandom.getInstance("SHA1PRNG");
    } catch (Exception e) {
      Util.panic(e);
    }
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Identifier */
  private byte[] id;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Builds a new identifier, by generating a random sequence of
   *  <CODE>SIZE</CODE> bytes.
   */
  public AntIdImpl()
  {
    id = new byte[SIZE];
    random.nextBytes(id);
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Check whether this <CODE>AntIdImpl</CODE> is equal to the specified object.
   */
  public boolean equals(Object obj)
  {
    if (obj != null && obj instanceof AntIdImpl) {
      AntIdImpl aid = (AntIdImpl) obj;
      for (int i=0; i < SIZE; i++)
        if (aid.id[i] != this.id[i])
          return false;
      return true;
    } else {
      return false;
    }
  }

  /**
   *  Returns a hash code value for the object.
   */
  public int hashCode()
  {
    int hash = id[0];
    for (int i=1; i < SIZE; i++) {
      hash = hash * id[i];
    }
    return hash;
  }
  
  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(": ");
    for (int i=0; i < SIZE; i++) 
      buf.append(Integer.toHexString(id[i]));
    buf.append("]");
    return buf.toString();
  }
  
} // END AntIdImpl

