/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antnet;

import anthill.*;
import anthill.nest.*;

/**
 *  This interface contains method receive(), which should be
 *  implemented by a anthill in order to receive an agent
 *  from other nests. The interface is implemented as a remote
 *  interface (it extends Remote and method receive() throws
 *  RemoteException); however, implementations of this class are
 *  not required to be based on RMI. 
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class Executor
extends Thread
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields 
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Reference to the activity manager to be notified when a
   *  the run() method of an agent has been completed.
   */
  private ConcurrentManager manager;

  /** 
   *  True if this thread is executing the run() method of an 
   *  agent; false otherwise */
  private boolean executing;

  /** 
   *  Reference to the AntView object containing an agent to
   *  be executed and providing the basic API accessible to
   *  agents.
   */
  private AntView view;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Instantiates an executor thread. Initially, the executor is empty,
   *  i.e. it does not execute any thread.
   *
   *  @param anthill the Nest object to be passed to agents through method
   *    run()
   *  @param manager a reference to the object to be notified when an agent
   *    has been completed.
   */
  Executor(ConcurrentManager manager)
  {
    this.manager = manager;
    executing = false;
    this.start();
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods 
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Requires that an agent is executed by the executor thread.
   *  The agent is contained in an AntDescriptor structure.
   *  If the thread is busy executing another agent, an exception
   *  is thrown. Otherwise, the agent is accepted and stored
   *  in the agent field. The execute() method returns immediately,
   *  while the run() method of this thread will take care of 
   *  executing the run() method of the specified object.
   */
  synchronized void execute(AntView view)
  {
    if (executing)
      throw new IllegalStateException("Thread busy");
    this.view = view;
    executing = true;
    notify();
  }
  
  /**
   *  Method run() of the thread. The thread waits for an agent by calling
   *  method waitStart(); executes the run() method of the thread; and then
   *  notifies the thread manager of its completion, in order that the thread
   *  manager can schedule another work.
   */
  public void run()
  {
    while (true) {
      waitStart();
      view.getAnt().run(view);
      notifyStop();
    }
  }
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Waits until executing is true, and store infos about
   *  the executing agent.
   */
  synchronized void waitStart()
  {
    while (executing == false) {
      try { wait(); } catch (InterruptedException e) { }
    }
  }
    
  /**
   *  Cleans variables associated with the agent, and notifies
   *  the thread manager about the completion of the agent.
   */
  synchronized void notifyStop()
  {
    executing = false;
    view = null;
    manager.completed(this);
  }

} // END Executor
