/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antnet;

import anthill.*;
import anthill.nest.*;

import java.io.*;
import java.util.*;

/**
 *  <CODE>FileResource</CODE> represents a local file resource that can be
 *  added by a an anthill clients.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class FileResource
implements Resource
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** String identifier of the resource */
  private String name;

  /** File object associated to the resource */
  private File file;
  
  /** Mapping between attribute names and attribute values */
  private HashMap map;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a new <CODE>FileResource</CODE> identified by the 
   *  specified name and containing the specified file.
   */
  FileResource(String name, File file)
  {
    this.name = name;
    this.file = file;
    map = new HashMap();
  }

    
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Resource
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the key identifier of the object.
   */
  public Object getKey()
  {
    return name;
  }

  /**
   *  Returns the string identifier of the object.
   */
  public String getName()
  {
    return name;
  }
  
  
  /**
   *  Returns the value associated to the specified attribute
   *  name, or null if there is no attribute with that name.
   *
   *  @param name the name of the attribute
   */
  public String getAttribute(String name)
  {
    return (String) map.get(name);
  }

  
  /**
   *  Return an input stream to download the contents of the
   *  resource. Depending on the implementation, the contents
   *  may be read from the local storage, if the resource has
   *  been copied locally; or downloaded remotely, otherwise.
   *
   *  @return an input stream to read the contents of a resource;
   */
  public InputStream getInputStream()
    throws IOException
  {
    return new FileInputStream(file);
  }

  
  /**
   *  Return the lenght of the content of the resource, if
   *  available; otherwise, it returns -1.
   */
  public long getContentLength()
  {
    if (!file.exists())
      return -1;
    else
      return(file.length());
  }

  public long size()
  {
    return name.length() + file.getAbsolutePath().length() + map.size();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods 
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Insert a mapping between the specified attribute
   *  name and the specified attribute value.
   *
   *  @param name the attribute name
   *  @param value the attribute value
   */
  void setAttribute(String name, String value)
  {
    map.put(name, value);
  }

} // END FileResource

