/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antnet;

import anthill.*;
import anthill.nest.*;

import java.io.*;
import java.net.*;


/**
 *  A UdpGate is an implementation of the Gate interface based on
 *  UDP. UdpGate creates a thread listening to a datagram socket,
 *  and delivers received agents to the socket it receives.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class UdpGate
extends Thread
implements Gate
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Maximum packet size */
  private static final int PACKETSIZE = 1024;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Socket */
  private DatagramSocket socket;

  /** Reference to the object listening for agents */
  private AntListener listener;

  /** Id of this anthill, used for identification */
  private UdpNid me;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Constructs a new UdpGate by instantiating a new datagram
   *  socket listening on the port contained in the anthill 
   *  identifier.
   */
  public UdpGate(int port)
    throws IOException
  {
    me = new UdpNid(port);
    socket = new DatagramSocket(port);
    this.start();
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Gate
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the identifier of the Nest to which this
   *  gate is associated.
   *
   */
  public NestId getId()
  {
    return me;
  }

  
  /*
   *  Set the agent listener, i.e. the object that will
   *  accept the agents coming from other nests.
   *
   *  @param listener the listener object that will
   *    accepts the agents coming from other nests
   */
  public void setAntListener(AntListener listener)
  {
    this.listener = listener;
  }
  
  
  /**
   *  Send an agent to the specified anthill. Both the anthill
   *  identifier and the agent must be serializable or
   *  externalizable.
   *
   * @exception ClassCastException
   *   raised if the anthill identifier does not belong to the
   *   expected class (in this case, UdpNid)
   * @exception IOException
   *   raised if something goes wrong during serialization
   *   and sending
   */
  public void send(NestId id, Ant ant, int ttl)
    throws IOException, ClassCastException
  {
    UdpNid dest = (UdpNid) id;
    
    // FIXME: We need a datastructure that does not use copying
    ByteArrayOutputStream bout = new ByteArrayOutputStream(PACKETSIZE);
    ObjectOutputStream out = new ObjectOutputStream(bout);
    out.writeObject(me);
    out.writeObject(ant);
    out.writeInt(ttl);
    out.close();
    byte[] buffer = bout.toByteArray();
    DatagramPacket packet = new DatagramPacket(buffer, buffer.length, 
      dest.getAddress(), dest.getPort());
    socket.send(packet);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Gate
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  While-true loop waiting for a packet arriving from the
   *  datagram socket; the received packet is delivered to
   *  the agent listener object.
   */
  public void run()
  {
    while (true) {
      try {
        
        /*
         *  Listen for a packet
         */
        byte[] buffer = new byte[PACKETSIZE];
        DatagramPacket packet = new DatagramPacket(buffer, PACKETSIZE);
        socket.receive(packet);

        if (listener != null) {
          /*
           *  Unmarshals a anthill id and an agent through serialization
           */
          ByteArrayInputStream bin = new ByteArrayInputStream(buffer);
          ObjectInputStream in = new ObjectInputStream(bin);
          NestId sender = (NestId) in.readObject();
          Ant agent = (Ant) in.readObject();
          int ttl = in.readInt();
          AntDescriptor desc = new AntDescriptor(sender, agent, ttl);
          listener.deliver(desc);
        } else {
          /*
           *  If no listener is present, the packet is discarded.
           *  FIXME: Right?
           */
        }
  
      } catch (Exception e) {
        // FIXME: Better management of this exception
        e.printStackTrace(System.err);
      }
    }
  }

  public void clear()
  {
    // FIXME
  }
  
} // END UdpGate


