/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antnet;

import anthill.*;
import anthill.nest.*;

import java.io.*;
import java.net.*;

/**
 *  An UDP anthill identifier is composed by an IP address plus an UDP port.
 *  This information is used to contact an UDP anthill. Methods equals(),
 *  hashcode() and toString() are implemented opportunely.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class UdpNid
implements NestId, Externalizable
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** IP address of the Nest */
  InetAddress address;
  
  /** UDP port */
  int port;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor needed for externalization.
   */
  public UdpNid()
  {
  }

  
  /**
   *  Instantiates a new UDP anthill id with the specified 
   *  port number, using the local address as IP address.
   */
  public UdpNid(int port)
    throws UnknownHostException
  {
    address = InetAddress.getLocalHost();
    this.address = address;
    this.port = port;
  }

  /**
   *  Instantiates a new UDP anthill id and sets both the address and the port number.
   */
  public UdpNid(InetAddress address, int port)
  {
    this.address = address;
    this.port = port;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the address contained in this id. 
   */
  public InetAddress getAddress()
  {
    return address;
  }
  
  /**
   *  Returns the port number contained in this id. 
   */
  public int getPort()
  {
    return port;
  }

 
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Check if two UdpNid objects are equals.
   */
  public boolean equals(Object obj)
  {
    if (obj instanceof UdpNid) {
      UdpNid id = (UdpNid) obj;
      return (this.address.equals(id.address) && this.port == id.port);
    } else {
      return false;
    }
  }

  /**
   *  Returns a hash code value for the object.
   */
  public int hashCode()
  {
    return address.hashCode()+port;
  }
  
  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    return "[Address:"+address.toString()+",Port:"+port+"]";
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Read a UdpNid from a stream.
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    this.address = (InetAddress) in.readObject();
    this.port = in.readInt();
  }
  
  /**
   *  Write a UdpNid to a stream.
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeObject(address);
    out.writeInt(port);
  }
  
} // END UdpNid

