/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antsim;

import java.util.*;
import anthill.util.*;
import anthill.factory.*;
import anthill.antsim.*;

/**
 *  The <I>Stats</I> class enables programmers to collect information
 *  about events occurring during a simulation. In particular, this
 *  class may be used to count the occurrence of events. Each event
 *  is identified by a string name chosen by the programmer and is 
 *  associated to a counter. The value of the counter may be incremented
 *  by one by invoking method <I>increment()</I>; the same value
 *  may be read by invoking method <I>get()</I>.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class Stats
implements Clearable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Hash table mapping event names and event counters for saving averages.
   *  Note the use of static here, thus multiple instance of Stats objects
   *  will be averaged over.
   */
  private static HashMap avgMap = new HashMap();

  /**
   *  Hash table mapping event names and event counters.
   */
  private HashMap map;

  /**
   *  Array list used to maintain order of elements
   */
  private ArrayList list;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Initializes an empty stats object.
   */
  public Stats()
  {
    map = new HashMap();
    list = new ArrayList();
  }
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Increment by 1 the value of the counter identified by <CODE>name</CODE>.
   *
   */
  public void inc(String name)
  {
    add(name, 1);
  }

  /**
   *  Add the value of <CODE>addend</CODE> to the counter
   *  identified by <CODE>name</CODE>.
   *
   */
  public void add(String name, float addend)
  {
    Wrapper wrapper = (Wrapper) map.get(name);
    if (wrapper == null) {
      list.add(name);
      wrapper = new Wrapper();
      map.put(name, wrapper);
    }
    wrapper.value += addend;
  }

  /**
   *  Return the value of the counter identified  by <CODE>name</CODE>
   */
  public float get(String name)
  {
    Wrapper wrapper = (Wrapper) map.get(name);
    return (wrapper != null ? wrapper.value : -1);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for computing the average
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the average value of the counter identified  by <CODE>name</CODE>
   */
  public static float getAvg(String name)
  {
    Wrapper wrapper = (Wrapper) avgMap.get(name);
    return (wrapper != null ? wrapper.value : -1);
  }

  /**
   *  Computes the average of all the values stored in the map,
   *  and keeps the result in the avgMap. The avgMap should not
   *  be cleared between iterations, therefore it is not inclued
   *  in this class' clear method.
   */
  public void computeAvg()
  {
    if (avgMap.size() > 0) {
      /* Compute the average using the map and avgMap data. */
      Iterator i = map.keySet().iterator();
      while (i.hasNext()) {
        String name = (String) i.next();
        Wrapper w = (Wrapper) map.get(name);
        if (w != null) {
          if (avgMap.containsKey(name)) {
            Wrapper avgw = (Wrapper) avgMap.get(name);
            /* This updates the value in the avgMap. */
            avgw.value += w.value;
            avgw.value /= 2;
            // FIXME? There seems to be some incorrectness in the average result;
            //  Could be due to floating point precision?
          } else {
            /* First iteration, add the value to the avgMap. */
            avgMap.put(name, w);
          }
        }
      }
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for printing
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a semi-colon separated list of values to be printed
   *    
   */
  public String toString(boolean title)
  {
    StringBuffer buffer = new StringBuffer();
    buffer.append("\n");
    if (title) {
      for (int i=0; i < list.size(); i++) {
        buffer.append(list.get(i));
        if (i != list.size()-1)
          buffer.append(" ; ");
      }
      buffer.append("\n");
    }
    for (int i=0; i < list.size(); i++) {
      String name = (String) list.get(i);
      buffer.append(get(name));
      //      buffer.append("(" + getAvg(name) + ")");
      buffer.append(" ; ");
    }
    return buffer.toString();
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the object to its initial state, setting all fields
   *  to zero, except for updating the average.
   */
  public void clear()
  {
    map.clear();
    list.clear();
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Wrapper inner class
  ////////////////////////////////////////////////////////////////////////////////////////////

  private class Wrapper
  {
    /** Float value enclosed in this wrapper */
    private float value;

    /** Default constructor */
    private Wrapper() 
    {
      value = 0;
    }
  } // END Wrapper
  
} // END Stats

