/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antsim.impl;

import anthill.*;
import anthill.factory.*;
import anthill.antsim.*;
import anthill.util.*;

/**
 *  Class <CODE>ExperimentImpl</CODE> is a simple implementation of 
 *  <CODE>Experiment</CODE> which repeat the execution of a single
 *  scenario a certain number of times and returns a single
 *  Stats object containing a summary of the statistics for all the 
 *  executions.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class ExperimentImpl
implements Experiment
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  The identifier of the counter used in a <CODE>Stats</CODE> object to 
   *  count the number of repetitions of the scenario evaluation in a 
   *  experiment.
   */
  public static final String ROUND_NUMBER = "RoundNumber";

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;
  
  /** Number of times a scenario is evaluated  */
  protected int iterations;
  
  /** If true, the scenario will be cleared between iterations of the experiment */
  protected boolean clear;

  /** If true, methods for the analysis of experiment footprint are invoked */
  protected boolean debug;

  /** Scenario on which evaluate the experiment */
  protected Scenario scenario;
  
  /** Version identifier of the scenario; when init params change, the version increase */
  protected int scenarioVersion;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new experiment by setting the given number of
   *  times by which a scenario has to be evaluated.
   *  
   *  @param factory the factory to be used to create scenarios
   *  @param iterations the number of scenario evaluations to be performed
   */
  public ExperimentImpl(Factory factory, int iterations)
  {
    this(factory, iterations, true, false);
  }

  /**
   *  Instantiates a new experiment by setting the given number of
   *  times by which a scenario has to be evaluated.
   *  
   *  @param factory     the factory to be used to create scenarios
   *  @param iterations  the number of scenario evaluations to be performed
   *  @param clear       if true, the scenario will be cleared after each
   *     iteration of the experiment
   */
  public ExperimentImpl(Factory factory, int iterations, boolean clear)
  {
    this(factory, iterations, clear, false);
  }


  /**
   *  Instantiates a new experiment by setting the given number of
   *  times by which a scenario has to be evaluated.
   *  
   *  @param factory the factory to be used to create scenarios
   *  @param iterations the number of scenario evaluations to be performed
   *  @param clear       if true, the scenario will be cleared after each
   *     iteration of the experiment
   *  @param debug       if true, methods to analyze the memory footprint of the 
   *    experiment are activated
   */
  public ExperimentImpl(Factory factory, int iterations, boolean clear, boolean debug)
  {
    this.factory = factory;
    this.iterations = iterations;
    this.debug = debug;
    this.clear = clear;
    clear();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Experiment
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Perform the evaluation of the current ant algorithm. Returns a
   *  <CODE>Stats</CODE> array containing the statistics collected during the
   *  simulation. 
   *  <p>
   *  In this implementation, a single <CODE>Stats</CODE> object is created
   *  and is used to collect statistics for all scenario executions
   *  performed by this <CODE>Experiment</CODE>.
   */
  public Stats[] evaluate()
  {
    long time, elapsed;
    
    Stats[] results = new Stats[iterations];
    for (int i=0; i < iterations; i++) {

      /* Compute time for statistics */
      time = System.currentTimeMillis();
      
      /* 
       *  Clear the scenario; create a new statistics object; evaluate the scenario 
       */
      results[i] = new Stats();
      results[i].add(ROUND_NUMBER, i);
      scenario.setStats(results[i]);
      scenario.evaluate();
      /* This iteration is completed; notify the scenario. */
      //      scenario.complete();
      if (clear == true)
        scenario.clear();

      /*
       *  For debug purposes:
       */
      elapsed = (System.currentTimeMillis()-time);
      time = System.currentTimeMillis();
      results[i].add("Simulation time", elapsed);
      if (debug) {
        //Log.println(Analyzer.printClasses(this));
        // results[i].add("Footprint", Analyzer.footprint(this));
        elapsed = (System.currentTimeMillis()-time);
        results[i].add("Analyzer time", elapsed);
      }

      /* Compute the average; is executed at each iteration. */
      //      results[i].computeAvg();

      /* 
       *  Print statistics; before printing the first row (i==0), we print also the
       *  title of the values 
       */
      Log.print(200, results[i].toString((i == 0)));
      Log.flush();
    }
    return results;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the object to its initial state. 
   */
  public void clear()
  {
    int version = factory.getVersion(Scenario.class);
    if (scenario == null || scenarioVersion != version) {
      /* Creates the scenario */
      scenario = (Scenario) factory.get(Scenario.class);
    } else {
      /* Restore state of existing scenario */
      scenario.clear();
    }
  }

} // END ExperimentImpl

