/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antsim.impl;

import anthill.*;
import anthill.nest.*;
import anthill.factory.*;
import anthill.antsim.*;

import java.util.Random;

/**
 *  This is a basic implementation of the Peernet interface. A <CODE>PeernetImpl</CODE>
 *  is characterized by a certain size (i.e., the number of nests composing
 *  it) and a degree (i.e., the number of links connecting it to other
 *  nests). 
 *  <p>
 *  This <CODE>Peernet</CODE> implementation uses the <CODE>Factory</CODE> class to 
 *  dynamically instantiate objects implementing Nest; the <CODE>Factory</CODE> 
 *  instance must be opportunely instructed with an XML definition for <CODE>Nest</CODE>.
 *  
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class PeernetImpl
implements Peernet
{
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  String name of the parameter used to determine the number of nodes included
   *  in this peernet.
   */
  public static final String PEERNET_SIZE = "PeernetSize";

  /** 
   *  String name of the parameter used to determine the degree of the nodes included
   *  in this peernet.
   */
  public static final String PEERNET_DEGREE = "PeernetDegree";

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator */
  static Random random = new Random();
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Reference to the peernet factory which created this object */
  Factory factory;
  
  /** The size of the network, i.e. the number of nests included in it */
  int size;

  /** The degree of the nests, i.e. the number of links to other nests */
  int degree;

  /** Nest network */
  Nest[] nests;

  /** Nest version */
  int nestVersion;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Initialize a new Peernet factory
   *  
   *  @param factory the factory to be utilized
   */
  public PeernetImpl(Factory factory)
  {
    this.factory = factory;
    clear();
  }
  

  /**
   * Restores the object to its initial state. 
   * New nests are instantiated when this method is called for the first
   * time; if the initialization arguments for the Nest interface have
   * been changed, the old nests are discarded and the new nests are 
   * created. In all other cases, the old nodes are cleared.
   */
  public void clear()
  {
    /*
     *  Read parameter PeernetSize and PeernetDegree from the factory object; if the 
     *  parameter are found, the value are updated. Otherwise, the initialization 
     *  parameter is still valid.
     */
    try {
      size = factory.intValue(PEERNET_SIZE);
    } catch (Exception e) { }
    try {
      degree = factory.intValue(PEERNET_DEGREE);
    } catch (Exception e) { }
    
    
    /*
     * Creates or clears the nodes of the peernet
     */
    int version = factory.getVersion(Nest.class);
    if (nests == null || version != nestVersion) {

      /* Update nestVersion */
      nestVersion = version;
      
      /* Create new nodes */
      nests = new Nest[size];
      for (int i=0; i < size; i++) {
        nests[i] = (Nest) factory.get(Nest.class);
      }

    } else {
      if (nests.length < size) {
        /* 
         *  The new size is greater than the previous size; 
         *  create new nests, clear previous ones
         */
        Nest[] newnests = new Nest[size];
        for (int i=0; i < nests.length; i++) {
          newnests[i] = nests[i];
          newnests[i].clear();
        }
        for (int i=nests.length; i < newnests.length; i++)
          newnests[i] = (Nest) factory.get(Nest.class);
        nests = newnests;
      } else if (nests.length > size) {
        /* 
         *  The new size is less than the previous size; 
         *  discard additional nests, clear remaining ones
         */
        Nest[] newnests = new Nest[size];
        for (int i=0; i < size; i++) {
          newnests[i] = nests[i];
          newnests[i].clear();
        }
        nests = newnests;
      } else {
        for (int i=0; i < nests.length; i++) {
          nests[i].clear();
        }
      }
    }
      
    /*
     *  Create arcs
     */
    for (int i=0; i < size; i++) {
      for (int j=0; j < degree; j++) {
        int k = i;
        while (k == i) 
          k = random.nextInt(size);
        nests[k].addClientNeighbour(nests[i].getId());
        nests[i].addClientNeighbour(nests[k].getId());
      }
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Peernet
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the nodes composing the simulated p2p network
   */
  public Nest[] getNests()
  {
    return nests;
  }

  /**
   *  Returns the size of the simulated p2p network (number of nests)
   *
   *  @return size of the p2p network.
   */
  public int size()
  {
    return size;
  }

} // END PeernetImpl

