/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antsim.impl;

import anthill.*;
import anthill.factory.*;
import anthill.antsim.*;
import anthill.nest.*;
import anthill.nestor.*;
import anthill.util.*;

import java.util.*;

/**
 *  Class <CODE>RangeExperiment</CODE> is a simple implementation of 
 *  <CODE>Experiment</CODE> which repeat the execution of a single
 *  scenario a certain number of times and returns a single
 *  Stats object containing a summary of the statistics for all the 
 *  executions.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class RangeExperiment
implements Experiment
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  The identifier of the counter used in a <CODE>Stats</CODE> object to 
   *  count the number of repetitions of the scenario evaluation in a 
   *  experiment.
   */
  public static final String ROUND_NUMBER_ID = "NumberOfRounds";

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;
  
  /* Name of the property to be set */
  protected String name;  

  /* Maximum value */
  protected int maxvalue;  
  
  /* Minimum TTL value */
  protected int minvalue;  

  /* Value used to increment step */
  protected int step; 

  /** Number of times a scenario is evaluated  */
  protected int iterations;
  
  /** If true, methods for the analysis of experiment footprint are invoked */
  protected boolean debug;

  /** Scenario on which evaluate the experiment */
  protected Scenario scenario;
  
  /** Version identifier of the scenario; when init params change, the version increase */
  protected int scenarioVersion;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new experiment by setting the given number of
   *  times by which a scenario has to be evaluated.
   *  
   *  @param factory the factory to be used to create scenarios
   *  FIXME Comment argument names
   */
  public RangeExperiment(Factory factory, String name, int minvalue, int maxvalue, int step)
  {
    this(factory, name, minvalue, maxvalue, step, 1, false);
  }
  
  /**
   *  Instantiates a new experiment by setting the given number of
   *  times by which a scenario has to be evaluated.
   *  
   *  @param factory the factory to be used to create scenarios
   *  FIXME Comment argument names
   */
  public RangeExperiment(Factory factory, String name, int minvalue, int maxvalue, 
    int step, int iterations)
  {
    this(factory, name, minvalue, maxvalue, step, iterations, false);
  }

  /**
   *  Instantiates a new experiment by setting the given number of
   *  times by which a scenario has to be evaluated.
   *  
   *  @param factory the factory to be used to create scenarios
   *  FIXME Comment argument names
   */
  public RangeExperiment(Factory factory, String name, int minvalue, int maxvalue, 
    int step, int iterations, boolean debug)
  {
    this.factory = factory;
    this.name = name;
    this.minvalue = minvalue;
    this.maxvalue = maxvalue;
    this.step = step;
    this.iterations = iterations;
    this.debug = debug;
  }
   
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Experiment
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Perform the evaluation of the current ant algorithm. Returns a
   *  <CODE>Stats</CODE> array containing the statistics collected during the
   *  simulation. 
   *  <p>
   *  In this implementation, a single <CODE>Stats</CODE> object is created
   *  and is used to collect statistics for all scenario executions
   *  performed by this <CODE>Experiment</CODE>.
   */
  public Stats[] evaluate()
  {
    long time, elapsed;
    
    /* Allocate space for the statistics object to be returned */
    int steps = (maxvalue-minvalue)/step +1;
    Stats[] results = new Stats[steps];

    /* Creates the scenario */
    factory.putValue(name, minvalue);
    scenario = (Scenario) factory.get(Scenario.class);

    int value = minvalue;
    for (int i=0; i < steps; i++) {
      
      /* Update value in the factory; the new value will be set when calling clear */
      factory.putValue(name, value);
      
      /* Compute time for statistics */
      time = System.currentTimeMillis();
      
      /* 
       *  Clear the scenario; create a new statistics object; evaluate the scenario 
       */
      results[i] = new Stats();     
      results[i].add(name, value);
      if (iterations > 1)
        results[i].add("Iterations", iterations);
      for (int j=0; j < iterations; j++) {
        System.out.println("Iteration: " + i);
        scenario.clear(); 
        scenario.setStats(results[i]);
        scenario.evaluate();
      }
      
      /*
       *  For debug purposes:
       */
      elapsed = (System.currentTimeMillis()-time);
      time = System.currentTimeMillis();
      results[i].add("Simulation time", elapsed);
      if (debug) {
        results[i].add("Footprint", Analyzer.footprint(this));
        elapsed = (System.currentTimeMillis()-time);
        results[i].add("Analyzer time", elapsed);
      }
      
      /* 
       *  Print statistics; before printing the first row (i==0), we print also the
       *  title of the values 
       */
      Log.print(200, results[i].toString((i == 0)));

      /* Increase value */
      value += step;
      
    }
    return results;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the object to its initial state. 
   */
  public void clear()
  {
    int version = factory.getVersion(Scenario.class);
    if (scenario == null || scenarioVersion != version) {
      /* Creates the scenario */
    } else {
      /* Restore state of existing scenario */
      scenario.clear();
    }
  }

} // END RangeExperiment

