/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antsim.impl;

import anthill.*;
import anthill.nest.*;
import anthill.factory.*;
import anthill.antsim.*;
import anthill.util.*;

import java.util.*;


/**
 *  Class <CODE>ResponseMonitor</CODE> is 
 *  
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision$
 */
public class ResponseMonitor
implements Clearable
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Stats */
  private Stats stats;
  
  /** Responses */
  private HashMap responses;
  
  /** Time */
  private long time;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor / reinitialization methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Instantiates a new <CODE>ResponseMonitor</CODE>.
   *
   *  @param stats  
   */
  public ResponseMonitor()
  {
    clear();
  }
  
  
  /**
   *  Restores the object to its initial state. The network is re-initialized
   *  as an empty network.
   */
  public void clear()
  {
    if (responses == null) {
      responses = new HashMap();
    } else {
      responses.clear();
    }
    time = 0;
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods 
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  
   */
  public void setStatistics(Stats stats)
  {
    this.stats = stats;
    stats.add("TotalRequests", 0);
    stats.add("SatisfiedRequests", 0);
    stats.add("TotalResponses", 0);
    stats.add("LocalFound", 0);
    stats.add("TotalTime", 0);
  }

  /**
   *  
   */
  public void tick()
  {
    time++;
  }

  /**
   *  
   */
  public void tick(int v)
  {
    time += v;
  }

  /**
   *  
   */
  public void startRequest(RequestId rid)
  {
    ResponseSet rs = (ResponseSet) responses.get(rid);
    if (rs != null)
      throw new IllegalArgumentException("Duplicated request: " + rid);
    rs = new ResponseSet(time);
    responses.put(rid, rs);
    stats.inc("TotalRequests");
  }
  
  /**
   *  Upcall method used to notify this listener that the request 
   *  associated to it has returned the specified set of resources.
   *
   *  @param rid the request identifier for the returned resources.
   *  @param resources the resources found.
   *  @return true  if the timeout has passed.
   */
  public boolean returnResources(RequestId rid, Resource[] resources, int timeout)
  {
    ResponseSet set = (ResponseSet) responses.get(rid);
    if (set != null) { 
      long diff = time - set.time;
      if (set.empty) {
        stats.inc("SatisfiedRequests");
        //System.out.println("Rid: " + rid + " Time: " + time + " Start: " + set.time);
        if (diff == 0)
          stats.inc("LocalFound");
        stats.add("TotalTime", diff);
        set.empty = false;
      }
      for (int i = 0; i < resources.length; i++) {
        if (set.responses.add(resources[i])) {
          /* The resource was not present; increment number of unique resources found. */
          /* FIXME is the above statement correct?? */
          stats.inc("TotalResponses");
        }
      }
      if (diff >= timeout)
        return true;
    }
    return false;
  }

  /**
   *  Returns the responses associated with the specified response identifier.
   *
   *  @param rid the request identifier whose responses to return.
   *  @return the set of responses associated with the specified response identifier.
   */
  public Set responses(RequestId rid)
  {
    ResponseSet set = (ResponseSet) responses.get(rid);
    if (set == null)
      throw new IllegalArgumentException("Unknown request: " + rid);

//     Log.println(600, "no. of resp: " + set.responses.size());
//     Iterator i = set.responses.iterator();
//     while (i.hasNext())
//       Log.println(600, "resp: " + i.next());

    return set.responses;
  }

  /**
   *  Returns the responses associated with the specified response identifier
   *  if the number of responses received so far exceeds limit.
   *
   *  @param rid the request identifier whose responses to return.
   *  @param fraction the fraction of unique resources that must be satisfied
   *    before we return any responses. If the total number of unique responses is
   *    below the limit <code>null</code> is returned.
   *  @return the set of responses associated with the specified response identifier.
   *    Returns null if not enough resources has been found yet.
   */
  public Set responses(RequestId rid, float fraction)
  {
    if (stats.get("TotalResponses")*fraction >= 1)
      return responses(rid);
    else
      return null;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner class 
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Response collection
   */
  private class ResponseSet
  {
    /** Request generation time */
    long time;
    
    /** Unique responses */
    Set responses;

    /** True if we have not found any resource for this request */
    boolean empty;
    
    /** Constructor */
    private ResponseSet(long time)
    {
      this.time = time;
      responses = new HashSet();
      empty = true;
    }
  } // END ResponseSet
    
    
} // END ResponseMonitor
  
