/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antsim.impl;

import anthill.*;
import anthill.nest.*;
import anthill.factory.*;
import anthill.antsim.*;
import anthill.util.*;

import java.util.*;


/**
 *  Class <CODE>ScenarioImpl</CODE> is a simple implementation of
 *  interface <CODE>Scenario</CODE> in which a predefined set of
 *  resources is inserted in the network and then a certain number
 *  of requests are performed and simulate. A <CODE>ScenarioImpl</CODE>
 *  is composed of a <CODE>Peernet</CODE> and a <CODE>ResourceSet</CODE> 
 *  object. Each resource included in the <CODE>ResourceSet</CODE> is inserted
 *  before starting the simulation in the <CODE>Peernet</CODE>. Then,
 *  the <CODE>ResourceSet</CODE> is used to generate a certain number
 *  of requests to be presented to the nests composing the network.
 *  
 *  
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class ScenarioImpl
implements Scenario, ResponseListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  String name of the parameter used to determine the number of rounds which
   *  will be performed in this scenario.
   */
  public static final String SCENARIO_ROUNDS = "ScenarioRounds";
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Random number generator */
  private static Random random = new Random();
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Configuration fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;
  
  // FIXME
  protected int nrequests;
  
  /** 
   *  Initialization rounds to be performed; during this phase, requests are 
   *  generated and statistics are not computed 
   */
  protected int initRounds;

  /** 
   *  Regular rounds to be performed; during this phase, requests are 
   *  generated and statistics are computed
   */
  protected int rounds;

  /** 
   *  Final rounds to be performed; during this phase, requests are not
   *  generated and statistics are computed
   */
  protected int finalRounds;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Configuration fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Stats */
  protected Stats stats;

  /** Responses */
  protected ResponseMonitor monitor;

  /** Peernet on which the genetic code will be evaluated */
  protected Peernet net;
  
  /** Version identifier of the peer network; when init params change, the version increase */
  protected int netVersion;
  
  /** Set of resources stored in the Anthill network */
  protected ResourceSet resources;
  
  /** Version identifier of the resource set; when init params change, the version increase */
  protected int rsetVersion;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor / reinitialization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new ScenarioImpl.
   *
   *  @param factory  the factory object used to generate the peernet
   *    and the resource set contained in a scenario
   *  @param rounds the number of rounds of the run
   */
  public ScenarioImpl(Factory factory, int nrequests, int initRounds, int rounds, int finalRounds)
  {
    /* Initializes variables */
    this.factory = factory;
    this.nrequests = nrequests;
    this.initRounds = initRounds;
    this.rounds = rounds;
    this.finalRounds = finalRounds;
    
    /* Invokes clear*/
    clear();
  }

  
  /**
   *  Restores the object to its initial state. The network is re-initialized
   *  as an empty network.
   */
  public void clear()
  {
    int version; 

    /*
     *  Read parameter ScenarioRounds from the factory object; if the parameter 
     *  is found, the value is updated. Otherwise, the initialization parameter
     *  is still valid.
     */
    try {
      rounds = factory.intValue(SCENARIO_ROUNDS);
    } catch (Exception e) {
    }
    
    /* 
     * Creates a new peernet iff (i) there is no peernet currently; 
     * (ii) the peernet initialization data have changed.
     * Otherwise, clear existing one.
     */
    version = factory.getVersion(Peernet.class);
    if (net == null || netVersion != version ) {
        netVersion = version;
        net = (Peernet) factory.get(Peernet.class);
    } else {
      net.clear();
    }
    
    /* 
     * Creates a new resource set iff (i) there is no resource set currently; 
     * (ii) the resource set initialization data have changed. 
     * Otherwise, clear existing one.
     */
    version = factory.getVersion(ResourceSet.class);
    if (resources == null || rsetVersion != version) {
      rsetVersion = version;
      resources = (ResourceSet) factory.get(ResourceSet.class);
    } else {
      resources.clear();
    }

    /* Inserts generated resources in nests */
    Nest[] nests = net.getNests();
    
    Resource[] rs = resources.getResources();
    for (int i=0; i < rs.length; i++) {
      int pos = random.nextInt(nests.length);  
      nests[pos].insert(rs[i]);
    }

    if (monitor ==  null) {
      monitor = new ResponseMonitor();
    } else {
      monitor.clear();
    }
  }
   
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Scenario
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Set the statistics object which will be used throughout
   *  the scenario evaluation.
   */
  public void setStats(Stats stats)
  {
    this.stats = stats;
    monitor.setStatistics(stats);
  }
  

  /**
   *  Perform the evaluation of this scenario object by simulating it.
   *  The simulation is subdivided in rounds. At each iteration,
   *  a new request is generated and assigned to one of the nest
   *  included in the simulated peer network, which generates an ant
   *  for it. Furthermore, for each  nest, an ant included in the 
   *  manager queue is selected and its <CODE>run()</CODE> method
   *  is executed. Statistics about the simulated execution of the
   *  code are collected in the <CODE>Stats</CODE> object specified
   *  through method <CODE>setStats()</CODE>.
   */
  public void evaluate()
  {
    // Obtain information about the network
    Nest[] nests = net.getNests();
    Manager[] managers = new Manager[nests.length];
    for (int i=0; i < nests.length; i++) {
      managers[i] = nests[i].getManager();
    }
    
    // Execute the initialization rounds 
    for (int i=0; i < initRounds; i++) {
      for (int j=0; j < nrequests; j++) {
        /* Choose random nest and make a request */
        int nest = random.nextInt(nests.length);
        Object request = resources.getRequest();
        Rid rid = new Rid(nests[nest]);
        nests[nest].request(rid, request, this);
      }
      
      /* Each nest executes the code of an ant in its queue */
      for (int j=0; j < managers.length; j++)
        managers[j].execute();
    }

    // Execute the regular rounds 
    for (int i=0; i < rounds; i++) {
      monitor.tick();
      /* Choose random nest and make a request */
      for (int j=0; j < nrequests; j++) {
        int nest = random.nextInt(nests.length);
        Object request = resources.getRequest();
        Rid rid = new Rid(nests[nest]);
        monitor.startRequest(rid);
        nests[nest].request(rid, request, this);
      }
      
      /* Each nest executes the code of an ant in its queue */
      for (int j=0; j < managers.length; j++)
        managers[j].execute();
    }

    // Execute the final rounds 
    for (int i=0; i < finalRounds; i++) {
      monitor.tick();
      /* Each nest executes the code of an ant in its queue */
      for (int j=0; j < managers.length; j++)
        managers[j].execute();
    }
    for (int i=0; i < nests.length; i++) {
      Log.println(400, nests[i]);
    }
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ResponseListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upcall method used to notify this listener that the request 
   *  associated to it has returned the specified set of resources.
   * 
   *  @param resources the resources found.
   */
  public void returnResources(RequestId rid, Resource[] resources)
  {
    monitor.returnResources(rid, resources, 0);
  }

  /**
   *  Upcall method used to notify this listener that the request 
   *  associated to it has finished to return resources.
   */
  public void complete()
  {
  }
  
} // END ScenarioImpl

