/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.factory;

import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

/**
 *  Parser code to read an XML configuration file, using the supplied URL.
 *  The default URL is 'file:$HOME/.anthill/config.xml', which is used if the
 *  non-argument constructor is used. Otherwise, the constructor takes the
 *  URL as an argument.
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public final class ConfigParser {

  /////////////////////////////////////////////////////////////////////////////
  // Constants - Error message and XML tags
  /////////////////////////////////////////////////////////////////////////////

  private static final String cfgRootTagName        = "SystemConfiguration";
  private static final String cfgVersionAttrib      = "version";
  private static final String cfgVersionValue       = "1.0";

  private static final String cfgCreateTagName      = "Create";
  private static final String cfgCreateIfaceAttrib  = "interface";
  private static final String cfgCreateClassAttrib  = "class";

  private static final String cfgConfigTagName      = "Config";

  private static final String cfgArgTagName         = "Argument";
  private static final String cfgArgNameAttrib      = "name";
  private static final String cfgArgTypeAttrib      = "type";
  private static final String cfgArgValueAttrib     = "value";

  /////////////////////////////////////////////////////////////////////////////
  // Static block - Default URL for the configuration file
  /////////////////////////////////////////////////////////////////////////////

  private static final String configURL;

  /*  Initialize the default URL for the configuration file.
   *  Used if no URL is supplied to the constructor. */
  static {
    String url = "file:" + System.getProperty("user.home");
    String s = java.io.File.separator;
    configURL = url.concat(s + ".anthill" + s + "config.xml");
  }

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  /*  Factory object used to invoke this parser object. */
  private Factory factory;

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a parser.
   *
   * @param factory    The factory object invoking the parser. Used to update
   *                   the factory's hash map of interface to <class, arg>
   */
  public ConfigParser(Factory factory)
  {
    this.factory = factory;
  }

  /**
   * Parses the file obtained from the specified URL.
   *
   * @param urlConfig  The URL to the xml configuration file containing
   *                   the description of the distributed system.
   */
  public void read(String urlConfig)
    throws ConfigFormatException, java.io.IOException
  {
    Document doc = null;
    try {
      DocumentBuilderFactory docBF = DocumentBuilderFactory.newInstance();
      docBF.setValidating(true);
      DocumentBuilder db = docBF.newDocumentBuilder();
      doc = db.parse(urlConfig);
    } catch (Exception exception) {
      exception.printStackTrace();
      throw new java.io.IOException("Unable to set up parsing environment");
    }
    Element elm = doc.getDocumentElement();
    checkCfgVersion(elm);
    NodeList nl = elm.getChildNodes();
    for (int i=1; i<=nl.getLength(); i++) {
      Node node = nl.item(i);
      if (node == null || node.getNodeName().startsWith("#")) {
        // Node is not an element, move on to the next node.
        continue;
      }
      elm = (Element)node;
      // This will place all mappings directly into the factory hash table.
      getCreateTag(elm, factory);
    }
  }


  /////////////////////////////////////////////////////////////////////////////
  // Private methods for parsing the XML tree
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Check that the configuration file has the expected
   *  version number and root tag.
   */
  private void checkCfgVersion(Element elm)
    throws ConfigFormatException
  {
    String cfgRoot = elm.getTagName();
    String cfgVersion = elm.getAttribute(cfgVersionAttrib);
    if (cfgRoot.equalsIgnoreCase(cfgRootTagName)) {
      if (cfgVersion.equals(cfgVersionValue)) {
        // All is OK, continue parsing.
        return;
      } else {
        throw new ConfigFormatException("Unexpected version ("
          + cfgVersion + "). Expected version " + cfgVersionValue);
      }
    } else {
      throw new ConfigFormatException("Unexpected root tag ("
        + cfgRoot + "). Expected " + cfgRootTagName);
    }
  }


  /**
   *  Get the content of a <Create> tag, including its <Argument> tag.
   *  
   */
  private void getCreateTag(Element elm, Factory factory)
    throws ConfigFormatException
  {
    if (elm != null && elm.getTagName().equalsIgnoreCase(cfgCreateTagName)) {
      
      /* Tag Create */
      
      String iface = elm.getAttribute(cfgCreateIfaceAttrib);
      String clazz = elm.getAttribute(cfgCreateClassAttrib);
      if (iface.length() == 0) {
        throw new ConfigFormatException(
          "Required 'interface' attribute not specified.");
      }
      if (clazz.length() == 0) {
        throw new ConfigFormatException(
          "Required 'class' attribute not specified.");
      }
      // System.out.println("iface: " + iface + ", class: " + clazz);
      NodeList argList = elm.getChildNodes();

      // Allocate space in the ArrayList to store the complete argument list
      ArrayList aList = new ArrayList(argList.getLength());

      /* Read the list of arguments */
      for (int i=0; i<argList.getLength(); i++) {
        Node node = argList.item(i);
        if (node != null && !node.getNodeName().startsWith("#")) {
          Element elem = (Element)node;
          ArgInfo arg = getArgumentInfo(elem, cfgArgTagName);
          aList.add(arg);
        }
      }
        
      /* Convert the argument list to the opportune arrays */
      String names[] = new String[aList.size()];
      Class classes[] = new Class[aList.size()];
      Object values[] = new Object[aList.size()];
      for (int i=0; i < aList.size(); i++) {
        ArgInfo arg = (ArgInfo) aList.get(i);
        names[i] = arg.name;
        classes[i] = arg.clazz;
        values[i] = arg.value;
      }
      
      /* Put the mapping into the internal factory ifaceMap table. */
      factory.put(iface, clazz, names, classes, values);

    } else if (elm.getTagName().equalsIgnoreCase(cfgConfigTagName)) {
        
      /* Tag Config */
      
      ArgInfo arg = getArgumentInfo(elm, cfgConfigTagName);
      factory.putValue(arg.name, arg.value);

    } else {
      throw new ConfigFormatException("Expected <Create> or <Config> tag.");
    }
  }


  /**
   *  Get the content of a single <Argument> tag.
   *
   *  @param elm  An element in the DOM tree.
   *  @return     Argument converted to its value
   */
  private ArgInfo getArgumentInfo(Element elm, String tagname)
    throws ConfigFormatException
  {
    if (elm.getTagName().equalsIgnoreCase(tagname)) {
      
      /* Read attributes */
      String name  = elm.getAttribute(cfgArgNameAttrib);
      String type  = elm.getAttribute(cfgArgTypeAttrib);
      String value = elm.getAttribute(cfgArgValueAttrib);
      
      /* Perform syntax checks */
      if (name.length() == 0)
        throw new ConfigFormatException("Required 'name' attribute not specified.");
      if (type.length() == 0)
        throw new ConfigFormatException("Required 'type' attribute not specified.");

      /* Return the triple (name, type, value) */
      Class clazz = obtainClass(type);
      Object wrapper = obtainWrapper(clazz, value);
      if (wrapper == null)
        wrapper = factory;
      return new ArgInfo(name, clazz, wrapper);
      
    } else {

      throw new ConfigFormatException("Expected " + tagname + " tag.");

    }
  }

  private Class obtainClass(String type)
  {
    if (type.equals("byte")) {
      return byte.class;
    } else if (type.equals("double")) {
      return double.class;
    } else if (type.equals("float")) {
      return float.class;
    } else if (type.equals("int")) {
      return int.class;
    } else if (type.equals("long")) {
      return long.class;
    } else if (type.equals("short")) {
      return short.class;
    } else if (type.equals("boolean")) {
      return boolean.class;
    } else if (type.equals("String")) {
      return String.class;
    } else if (type.equals("Factory")) {
      return Factory.class;
    } else {
      throw new ConfigFormatException("Only primitive types can be used on"
        + " constructors using the generic factory.");
    }
  }

  private Object obtainWrapper(Class type, String value)
    throws ConfigFormatException
  {
    Object arg;
    if (type == byte.class) {
      arg = (Object) new Byte(value);
    } else if (type == char.class) {
      // Note, we assume the first char of a string value is to be used
      // Maybe we could provide a check on the length of the string
      // and throw an exception if it is longer than 1 character.
      arg = (Object) new Character(value.charAt(0));
    } else if (type == double.class) {
      arg = (Object) new Double(value);
    } else if (type == float.class) {
      arg = (Object) new Float(value);
    } else if (type == int.class) {
      arg = (Object) new Integer(value);
    } else if (type == long.class) {
      arg = (Object) new Long(value);
    } else if (type == short.class) {
      arg = (Object) new Short(value);
    } else if (type == boolean.class) {
      arg = (Object) new Boolean(value);
    } else if (type == String.class) {
      arg = (Object) new String(value);
    } else if (type == Factory.class) {
      arg = null;
    } else {
      throw new ConfigFormatException("Only primitive types can be used on"
        + " constructors using the generic factory.");
    }
    return arg;
  }

  
} // END of class ConfigParser
