/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.genie;

/**
 *  For a nest, the genetic code of an ant is simply a 
 *  sequence of bits. The exact interpretation of the sequence
 *  depends on the class of the ant being used. For this
 *  reason, class <CODE>GeneticCode</CODE> abstracts the 
 *  basic operations that a nest is enabled to perform
 *  on a genetic code, which are (i) to read the contents of 
 *  each bit in the code; (ii) to obtain the total size of
 *  the code; and (iii) to read the content of the sequence
 *  as a list of parameters.
 *  <p> 
 *  For performance reasons, the class is final.
 *  Implementation note: the bits are currently maintained
 *  in an array of booleans. Alternative implementations could
 *  be based on integer array and bit-to-bit arithmetics, if
 *  they prove to be faster than this one.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public final class GeneticCode
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Sequence of bits representing this <CODE>GeneticCode</CODE> */
  private boolean[] bits;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Construct a new <CODE>GeneticCode</CODE> object starting from a sequence
   *  of bits maintained in a boolean array.
   */
  public GeneticCode(boolean[] bits)
  {
    this.bits = new boolean[bits.length];
    for (int i=0; i < bits.length; i++)
      this.bits[i] = bits[i];
  }

  
  /**
   *  Construct a new <CODE>GeneticCode</CODE> object starting from another
   *  <CODE>GeneticCode</CODE> object. Implements the semantics of 
   *  <CODE>Cloneble</CODE> without incurring in interface implementation 
   *  and casting.
   */
  public GeneticCode(GeneticCode code)
  {
    this.bits = new boolean[bits.length];
    for (int i=0; i < bits.length; i++)
      this.bits[i] = bits[i];
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the size of this genetic code.
   */
  public int size()
  {
    return bits.length;
  }

  /**
   *  Returns the content of bit <CODE>i</CODE>. 
   *
   *  @param i the index of the bit to be read
   */
  public boolean get(int i)
  {
    return bits[i];
  }

  /**
   *  Read the content of this genetic code formattted as specified
   *  by the <CODE>sizes</CODE> array. Argument <CODE>sizes</CODE>
   *  contains the size of each of the arguments in which the sequence
   *  of bits must be subdivided. For example, if <CODE>sizes</CODE> 
   *  is equal to <CODE>{ 4, 8, 4 }</CODE>, the first parameter takes
   *  the first four bits; the second parameter takes the next eight bits;
   *  and the last parameter takes the last four bits.
   *
   *  @param sizes the array containing the sizes in bits of the
   *    parameters to be read
   *  @return an array of integers containing the read parameters
   *  @throws IllegalArgumentException if the sum of sizes in the
   *    specified array is not equal to the length of this genetic
   *    code, as returned by <CODE>size()</CODE>.
   *    
   */
  public int[] get(int[] sizes)
  {
    int nparameters = sizes.length;
    int[] parameters = new int[nparameters];
    int index = 0;
    try {
      for (int i=0; i < nparameters; i++) {
        parameters[i] = 0;
        for (int j=0; j < sizes[i]; j++) {
          parameters[i] = parameters[i] * 2;
          if (bits[index])
            parameters[i]++;
          index++;
        }
      }
    } catch (IndexOutOfBoundsException e) {
      /*
       *  The sum of the parameter sizes is greater than the number
       *  of bits in this genetic code. Something went wrong! We
       *  compute the sum here to provide a better exceptione explanation.
       *  Note that computing the sum before would have been expensive
       *  for every computation, even those in which no exception is
       *  thrown.
       */
      int sum = 0;
      for (int i=0; i < nparameters; i++) 
        sum += sizes[i];
      throw new IllegalArgumentException("Invalid genetic code: required " + sum + 
        " bits, found " + index + " bits");
    }
    
    /*
     *  Checks whether the sum of the parameter size is less than the number of 
     *  bits in the genetic code. If so, throws an exception. 
     */
    if (index < bits.length)
      throw new IllegalArgumentException("Invalid genetic code: required " + index + 
        " bits, found " + bits.length + " bits");
    
    return parameters;
  }

  /**
   *  Check whether this <CODE>GeneticCode</CODE> is equal to the specified object.
   */
  public boolean equals(Object obj)
  {
    if (obj != null && obj instanceof GeneticCode) {
      GeneticCode code = (GeneticCode) obj;
      boolean equal = true;
      for (int i=0; i < bits.length && equal; i++) {
        equal = (bits[i] != code.bits[i]);
      }
      return equal;
    } else {
      return false;
    }
  }

  /**
   *  Returns a String containing this genetic code represented as a 0/1 string. 
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(": ");
    for (int i=0; i < bits.length; i++) {
      buf.append(bits[i] ? "1" : "0");
    }
    buf.append("]");
    return buf.toString();
  }
  
} // END GeneticCode

