/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.genie.impl;

import anthill.*;
import anthill.factory.*;
import anthill.genie.*;
import anthill.antsim.*;
import anthill.antsim.impl.*;
import anthill.util.*;

/**
 *  Class ExperimentImpl is a simple implementation of Experiment.
 *  It is constituted by a set of scenarios, on which the fitness
 *  is computed as the average fitness as computed in each scenario.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class GeneticExperimentImpl
implements GeneticExperiment
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  The identifier of the counter used in a <CODE>Stats</CODE> object to 
   *  count the number of repetitions of the scenario evaluation in a 
   *  experiment.
   */
  public static final String ROUND_NUMBER = "RoundNumber";

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;
  
  /** Number of times a scenario is evaluated  */
  protected int iterations;
  
  /** Genetic code to be evaluated */
  private GeneticCode code;
  
  /** Genetic scenario */
  private GeneticScenario scenario;
  
  /** Version identifier of the scenario; when init params change, the version increase */
  protected int scenarioVersion;

  /** Evaluated fitness */
  private double fitness;
  
  /** Boolean flag; true if the fitness has been already computed */
  protected boolean computed;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new experiment by creating the specified
   *  number of scenarios through the specified scenario factory.
   *
   */
  public GeneticExperimentImpl(Factory factory, int iterations)
  {
    this.factory = factory;
    this.iterations = iterations;  
    fitness = 0.0;
    computed = false;
    code = null;
  }
  
   
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Experiment
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Set the genetic code to be evaluated. In this basic implementation,
   *  only one genetic code can be present in the Nest network at a time.
   *  If this method is called more than once, the values is substituted
   *  with the new values.
   */
  public void addGeneticCode(GeneticCode code)
  {
    this.code = code;
    
  }

  /**
   *  Execute the experiment, i.e. evaluate the fitness of the genetic code on
   *  each of the scenarios contained in this experimnt. The fitness of the
   *  genetic code for this experiment is obtained by the sum of the fitness 
   *  values obtained by evaluating them on the scenarios.
   *
   *  @throws IllegalStateException if no genetic code is present
   */
  public Stats[] evaluate()
  {

    if (this.code == null)
      throw new IllegalStateException("No genetic code to be evaluated");

    Stats stats = new Stats();
    stats.add(ROUND_NUMBER, iterations);
    for (int i=0; i < iterations; i++) {
      scenario.clear(); 
      scenario.addGeneticCode(code);
      scenario.setStats(stats);
      scenario.evaluate();
    }
    
    /* Fitness evalation using an external class */
    Fitness fitnessEval = (Fitness) factory.get(Fitness.class);
    fitness = fitnessEval.compute(stats);
    Log.println(200, "Genetic code: " + code + " found " + fitness + "% resources");
    computed = true;

    return new Stats[] { stats };
  }

  
  /**
   *  Returns the evaluated fitness. 
   *
   *  @throws IllegalStateException if the fitness value has not been computed
   *    yet.
   */
  public double getFitness()
  {
    if (!computed)
      throw new IllegalStateException("No fitness value yet!");
    return fitness;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the object to its initial state. Method clear() is invoked on
   *  all scenarios included in this experiment. The genetic code to be
   *  evaluated and the computed fitness are restored to their initial
   *  state.
   */
  public void clear()
  {
    code = null;
    fitness = 0.0;
    computed = false;
    int version = factory.getVersion(GeneticScenario.class);
    if (scenario == null || scenarioVersion != version) {
      /* Creates the scenario */
      scenario = (GeneticScenario) factory.get(GeneticScenario.class);
    } else {
      /* Restore state of existing scenario */
      scenario.clear();
    }
  }

} // END ExperimentImpl

