/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.genie.impl;

import anthill.*;
import anthill.factory.*;
import anthill.antsim.*;
import anthill.genie.*;
import anthill.util.*;
import gajit.*;
import gajit.util.*;

/**
 *  Basic implementation of the GeneticFramework interface.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class GeneticFrameworkImpl
extends Printable
implements GeneticFramework
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;

  /** Number of genetic codes constituting the population */
  protected int     numcodes;
  
  /** Number of genes (bit sequences) constituting a genetic code */
  protected int     numgenes;
  
  /** Size of a single gene */
  protected int     sizegenes;
  
  /** Mutation rate */ 
  protected double  mutRate;

  /** Percentage of population carried forward unchanged from each generation */
  protected double  eliteRate;
  
  /** Percentage deemed unfit for reproduction */
  protected double  cullRate;
  
  /** Number of generations to be computed */
  protected int generations;

  /** Population of evolved genetic codes */
  protected Population people;
  
  /** Fittest genetic code */
  protected GeneticCode bestCode;
  protected double bestFitness;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new genetic framework
   */
  public GeneticFrameworkImpl(Factory factory, int numcodes, int numgenes, int sizegenes, 
    double mutRate, double eliteRate, double cullRate, int generations)
  {
    this.factory = factory;
    this.numcodes = numcodes;
    this.numgenes = numgenes; 
    this.sizegenes = sizegenes; 
    this.mutRate = mutRate; 
    this.eliteRate = eliteRate; 
    this.cullRate = cullRate; 
    this.generations = generations;
    
    // Initialize genetic codes population
    people = new Population(numcodes, sizegenes, numgenes, numgenes);
    people.setEliteRate(eliteRate);
    people.setCullRate(cullRate);
    people.addOp( new XOverPtOp( 1 ), 1 );
    if (mutRate > 0.0)
      people.addOp( new MutOp( mutRate ), 1 );

    // Initialize statistics
    bestCode = null;
    bestFitness = 0.0;
  }
  
  /**
   *  Tries to evolve a population of genetic codes, using the
   *  configuration parameters contained in this instance.
   *
   *  @param out  print stream used to log information about the
   *    evolution; if print stream is null, no log file is generated.
   */
  public void evolve()
  {
    // Compute generations
    for (int i = 0; i < generations; i++) {
      long time = System.currentTimeMillis();
      
      // Initialize Experiment
      GeneticExperiment experiment = 
        (GeneticExperiment) factory.get(GeneticExperiment.class);

      // Iterate over the population of genetic codes
      Iterator censo = people.iteratorNewOnly();
      while (censo.hasNext()) {
        ChromItem tmp = (ChromItem) censo.next();
        GeneticCode code = new GeneticCode(tmp.getChrom().toBits());

        experiment.addGeneticCode(code);
        experiment.evaluate();
        double val = experiment.getFitness();
        experiment.clear();
        tmp.setFitness(val);
        if (val > bestFitness) {
          bestFitness = val;
          bestCode = code;
        }
      }

      // Compute the next generation
      if (i < generations - 1) {
        people.newGeneration();
      }

      // Print statistics
      Log.println(100, i + " ; " + (System.currentTimeMillis() - time) + " ; " +
          bestFitness + " ; " + bestCode);
    }
  }

  /**
   *  Return the population of evolved genetic codes. 
   */
  public Population getPopulation()
  {
    return people;
  }

  
  /**
   *  Return the genetic code with the best fitness.
   */
  public GeneticCode getFittest()
  {
    return bestCode;
  }

} // END GeneticFramework

