/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nest;

import anthill.*;
import anthill.util.*;
import anthill.storage.*;

/**
 *  <CODE>Nest</CODE> is the interface used by nest users
 *  to communicate with nests. The basic methods contained in this 
 *  interface enable the user to search resources, to insert 
 *  resources in the Anthill network and to add knowledge
 *  about other nests.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public interface Nest
extends Clearable
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nest management
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the identifier of this nest
   */
  public NestId getId();

  /**
   *  Returns the <CODE>Manager</CODE> object associated to this nest
   */
  public Manager getManager();
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Resource management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Obtains a storage reference from the nest.
   *
   *  @param name  name of the storage to obtain a reference to.
   *  @return a <code>Storage</code> reference
   *  @exception StorageNotFoundException if the specified name does not
   *    correspond to a storage currently available from this nest.
   */
  public Storage getStorage(String name)
    throws StorageNotFoundException;

  /**
   *  Accepts a resource request originated by the user of this nest.
   *  The interpretation of the request object is left to the ant(s)
   *  created to satisfy the response. This method returns immediately.
   *  When the request is completed, one or more <CODE>Resource</CODE> 
   *  object will be returned to the specified resource listener.
   *
   *  @param request the request to be satisfied
   *  @param listener The listener object which will be notified
   *    when the request is completed.
   */
  // FIXME: Have this method return the RequestId instead.
  public void request(RequestId rid, Object request, ResponseListener listener);
  
  /**
   *  Inserts a new resource in this nest. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  resource storage contained in the nest. Please refer to the documentation 
   *  of the resource storage implementation for more details.
   * 
   *  @param resource the resource to be inserted.
   */
  public void insert(Resource resource);
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Neighbour list management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Adds the identifier of a new neighbour nest to the set of 
   *  neighbours known to this nest, on behalf of the nest user. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  gate object contained in the nest. Please refer to the documentation 
   *  of the resource storage implementation for more details.
   *
   *  @param nid the identifier of the neighbour to be added
   */
  public void addClientNeighbour(NestId nid);

  /**
   *  Adds the identifier of a new neighbour nest to the set of 
   *  neighbours known to this nest, on behalf of an ant visiting
   *  the nest.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  gate object contained in the nest. Please refer to the documentation 
   *  of the resource storage implementation for more details.
   *
   *  @param nid the identifier of the neighbour to be added
   */
  public void addAntNeighbour(NestId nid);

  /**
   *  Removes the identifier of a neighbour nest from the set of 
   *  neighbours known to this Nest. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  gate object contained in the nest. Please refer to the documentation 
   *  of the resource storage implementation for more details.
   *
   *  @param nid the identifier of the neighbour to be removed
   */
  public void removeNeighbour(NestId nid);
  
  /**
   * Returns information about the neighbours known to this nest.
   */
  public Neighbour[] getNeighbours();

  
} // END Nest
