/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import java.io.*;
import anthill.*;
import anthill.nest.*;
import anthill.storage.*;

/**
 *  Class LocalAntView is a simple implementation of interface 
 *  <CODE>AntView</CODE>, which implements methods included in
 *  it by invoking the opportune methods on the local nest.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class LocalAntView
implements AntView
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
 
  /** Reference to the nest object hosting this view */
  private LocalNest nest;
  
  /** The descriptor of the ant currently associated with this view */
  private AntDescriptor desc;

  /** True if the resource method move() has been invoked successfully */
  private boolean moved;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  Instantiates an AntView and stores in it the references to
   *  which the operations are delegated.
   *  <p>
   *  The constructor has package visibility to avoid
   *  instances not builded through the appropriate factory.
   *
   *  @param storage  the resource database
   *  @param gate     the communication handler
   *  @param desc     the descriptor of the ant to which this AntView
   *                  is associated
   */
  LocalAntView(LocalNest nest, AntDescriptor desc)
  {
    this.nest = nest;
    this.desc = desc;
    moved = false;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for requesting general information
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*  
   *  Javadoc comment inherited by interface anthill.AntView
   *  Asks the nest for its identifier
   */
  public NestId getNestId()
  {
    return nest.getId();
  }
  
  /*  
   *  Javadoc comment inherited by interface anthill.AntView
   *  Asks the associated ant descriptor for the nest identifier
   */
  public NestId getLastVisited()
  {
    return desc.getLastVisited();
  }

  /*  
   *  
   */
  public int getTTL()
  {
    return desc.getTTL();
  }

  /*
   *  Javadoc comment inherited by interface anthill.AntView
   *  Asks the associated ant descriptor for the ant reference
   */
  public Ant getAnt()
  {
    return desc.getAnt();
  }

  /**
   *  Creates a new ant identifier.
   */
  public AntId createAntId()
  {
    return new LocalAid();
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for ant movement and creation
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  FIXME: Implement.
   *  <p>
   *  The exact behavior of this method depends on the class implementing
   *  the <CODE>Manager</CODE> interface currently used in the current nest.
   *  Please refer to the implementation documentation for more details.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   */
  public void addAnt(Ant ant)
  {
    AntDescriptor newdesc = 
      new AntDescriptor(nest.getId(), ant, desc.getTTL());
    LocalAntView view = new LocalAntView(nest, newdesc);
    nest.getManager().add(view);
  }

  /**
   *  Moves an ant to the specified nest. In this particular implementation,
   *  an ant can perform at most one move operation when in a nest. This 
   *  restriction is imposed in order to avoid unbounded ant replication.
   *  <p>
   *  The exact behavior of this method depends on the class implementing
   *  the <CODE>Gate</CODE> interface currently used in the current nest.
   *  Please refer to the implementation documentation for more details.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   *
   *  @param id the id of the destination nest
   *  @throws IOException if the ant cannot move to the
   *    destination nest for communication problems.
   */
  public void move(NestId id)
    throws IOException
  {
    nest.getGate().send(id, desc.getAnt(), desc.getTTL());
    /*
     *  Note that moved is updated only if we are
     *  sure that the send operation has not
     *  raised an exception.
     */
    moved = true;
  }

  /**
   *  Leave trail information on the <CODE>Trail</CODE> object stored
   *  in the nest. If absent, the trail is created. Return true if the
   *  the information is left
   *  <p>
   *  The exact behavior of this method depends on the class implementing
   *  the <CODE>Trail</CODE> interface currently used in the current nest.
   *  Please refer to the implementation documentation for more details.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   */
  public boolean storeReturnNest(AntId aid, NestId nid)
  {
    Trail trail = nest.getPreviousNest(true);
    return trail.add(aid, desc.getLastVisited());
  }

  /**
   *  Lookup trail information on the <CODE>Trail</CODE> object stored
   *  in the nest and move to the obtained nest. If the trail is absent, 
   *  <CODE>IllegalStateException</CODE> is raised.
   *  <p>
   *  The exact behavior of this method depends on the class implementing
   *  the <CODE>Trail</CODE> interface currently used in the current nest.
   *  Please refer to the implementation documentation for more details.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   *  @param remove true if the association (ant id, nest id)
   *    should be removed after retrieval
   */
  public NestId getReturnNest(AntId aid, boolean remove)
  {
    Trail trail = nest.getPreviousNest(false);
    if (trail != null) {
      return trail.lookup(aid, remove);
    } else {
      throw new IllegalStateException("No trail information in this nest");
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for neighbour management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add the new neighbor information to the <CODE>Gate</CODE> object contained
   *  in the current nest. FIXME: correct implementation by removing addAntNeighbour
   *  and adding the method to Gate.
   *  <p>
   *  The exact behavior of this method depends on the class implementing
   *  the <CODE>Gate</CODE> interface currently used in the current nest.
   *  Please refer to the implementation documentation for more details.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   */
  public void addNeighbour(NestId id)
    throws IOException
  {
    nest.addAntNeighbour(id);
  }
    
  /**
   *  FIXME: Implement.
   *  <p>
   *  The exact behavior of this method depends on the class implementing
   *  the <CODE>Gate</CODE> interface currently used in the current nest.
   *  Please refer to the class documentation for more details.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   */
  public void removeNeighbour(NestId id)
  {
    // nest.removeAntNeighbour(id);
  }
    
  
  /**
   *  Ants invoke this method to obtain the list of nest neighbors
   *  known to the current nest. FIXME: correct implementation to refer to Gate
   *  <p>
   *  The exact behavior of this method depends on the class implementing
   *  the <CODE>Gate</CODE> interface currently used in the current nest.
   *  Please refer to the implementation documentation for more details.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   */
  public Neighbour[] getNeighbours()
  {
    return nest.getNeighbours();
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for pheromone management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return a storage object identified by the given name.
   */
  public Storage getStorage(String name)
    throws StorageNotFoundException
  {
    return nest.getStorage(name);
  }
  
  /**
   *  Performs the specified request on the local resource
   *  storage and returns the resources satisfying the
   *  request. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param request The request to be performed
   */
  public Resource[] getResources(Object request)
    throws StorageNotFoundException
  {
    return nest.getResources(request);
  }

  /**
   *  Performs the specified request on the local resource
   *  storage and returns the resources satisfying the
   *  request. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param request The request to be performed
   */
  public Resource getResource(Object key)
    throws StorageNotFoundException
  {
    return nest.getResource(key);
  }
  
  /**
   *  Add a resource to the nest. This method enables
   *  ants to move resources from one nest to another. 
   *  <p>
   *  Depending on the implementation of <CODE>Resource</CODE> 
   *  and <CODE>Nest</CODE>, adding a resource to a nest through 
   *  an AntView may result to actual copy the content of the 
   *  resource, or to simply insert a reference to the actual 
   *  position of the resource itself. 
   *  <p>
   *  This method may be invoked more than once from an ant, in 
   *  order to insert more resources. Nevertheless, the actual 
   *  insertion of a resource in a nest depends on the policies 
   *  of the particular <CODE>Nest</CODE> implementation. For 
   *  example, a nest may limit the number of resources originated 
   *  by  another nest, in order to avoid "spamming".
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param resource resource to be added
   */
  public void addResource(Resource resource)
    throws CapacityException, StorageNotFoundException
  {
    nest.addResource(resource);
  }
  
  
  /**
   *  Get the pheromone information object related to the species of the
   *  ant associated to this <CODE>AntView</CODE> object. Different
   *  ant species are associated to different pheromone objects.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   */
  public Object getPheromone()
  {
    return nest.getPheromone(desc.getAnt().getClass());
  }

  
  /**
   *  Set the pheromone information object related to the species of the
   *  ant associated to this <CODE>AntView</CODE> object. Different
   *  ant species are associated to different pheromone objects.
   *  
   *  @param pheromone the pheromone object to be set.
   *  <p>
   *  Please see the corresponding method comment in interface 
   *  <CODE>AntView</CODE> for additional information.
   */
  public void setPheromone(Object pheromone)
  {
    nest.setPheromone(pheromone, desc.getAnt().getClass());
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for resource management
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  This method is invoked by ants to notify the nest that
   *  the ant associated to this AntView, created to satisfy
   *  a request, is returning the specified set of resources.
   *
   *  @param request the request to be performed
   *  @param resources  the resources satisfying the query
   */
  public boolean returnResources(RequestId rid, Resource[] resources)
  {
    return nest.getRequestSet().returnResources(rid, resources);
  }
  
  /** Read from statistics */
  // FIXME Add comments
  // FIXME Add other methods (?)
  public int intValue(String name)
  {
    return nest.intValue(name);
  }
  
  
} // END LocalAntView

