/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import anthill.*;
import anthill.nest.*;
import anthill.factory.*;
import anthill.genie.*;
import java.util.*;

/**
 *  A Nest class collects the components of a Nest site: (i) a AntView
 *  object, which is the view of a Nest for an Ant; (ii) a Gate object,
 *  which manage the communication between nests; (iii) an Manager,
 *  i.e. the object which manage the activities performed by different agents.
 *  <p>
 *  Implementation note: this implementation is supposed to be used as
 *  simple implementation of <CODE>Nest</CODE> in the off-line
 *  environment. Here, we assume that the Manager object used is single-threaded;
 *  in this way, we avoid the costs of synchronized methods.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class LocalGeneticNest
extends LocalNest
implements GeneticNest
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Genetic code of the ant used in this environment*/
  private GeneticCode code;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a new nest.
   */
  public LocalGeneticNest(Factory factory)
  {
    super(factory);
    code = null;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Add a new genetic code to the set of genetic codes known to this anthill.
   *  When creating a new ant, an anthill may choose from one of the
   *  available genetic codes. This implementation of <CODE>Nest</CODE> 
   *  does not enable the use of more than one genetic code; thus, duplicated
   *  invocations of <CODE>addGeneticCode()</CODE> throws an 
   *  <CODE>IllegalStateException</CODE>.
   * 
   *  @param code the genetic code to be added
   *  @throws IllegalStateException if no other genetic code can be added to
   *    the anthill
   */
  public void addGeneticCode(GeneticCode code)
  {
    if (this.code == null)
      this.code = code;
    else
      throw new IllegalStateException("Maximum one genetic code in " + 
        this.getClass().getName());
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    super.clear();
    code = null;
  }
  
} // END Genie
