/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import anthill.*;
import anthill.nest.*;
import anthill.storage.*;
import anthill.factory.*;
import anthill.genie.*;
import anthill.util.*;

import java.util.*;

/**
 *  A Nest class collects the components of a Nest site: (i) a AntView
 *  object, which is the view of a Nest for an Ant; (ii) a Gate object,
 *  which manage the communication between nests; (iii) an Manager,
 *  i.e. the object which manage the activities performed by different agents.
 *  <p>
 *  Implementation note: this implementation is supposed to be used as
 *  simple implementation of <CODE>Nest</CODE> in the off-line
 *  environment. Here, we assume that the Manager object used is single-threaded;
 *  in this way, we avoid the costs of synchronized methods.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class LocalNest
implements Nest, AntListener
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** True if we have verified the */
  static private boolean checked = false;
  
  /**
   *  Verify that the factory contains information for building at least a document
   *  storage and url storage
   */
  static private void checkFactory(Factory factory)
  {
    try {
      QueryStorage storage = (QueryStorage) factory.get(Storage.DOC_STORAGE);
    } catch (ConstructorException e) {
      System.err.println("Document Storage not defined: ");
      System.err.println(e.getTargetException().getMessage());
      //e.getTargetException().printStackTrace(System.err);
    } catch (Exception e) {
      System.err.println("Document Storage not defined: ");
      System.err.println(e.getMessage());
      //e.printStackTrace(System.err);
    }
    try {
      QueryStorage storage = (QueryStorage) factory.get(Storage.URL_STORAGE);
    } catch (ConstructorException e) {
      System.err.println("Url Storage not defined: ");
      System.err.println(e.getTargetException().getMessage());
      //e.getTargetException().printStackTrace(System.err);
    } catch (Exception e) {
      System.err.println("Url Storage not defined: ");
      System.err.println(e.getMessage());
      //e.printStackTrace(System.err);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  String name of the parameter used to determine the ttl used by these nests.
   */
  public static final String NEST_TTL = "NestTTL";

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initialization fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  private Factory factory;
  
  /** Time-to-live used to initialize ants */
  private int ttl;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Ant factory used to generate ants for request */
  private AntFactory antf;
  
  /** Communication handler to be used to communicate with other nests */
  private Gate gate;
  
  /** Hash table mapping request to response listeners */
  private RequestSet requests;

  /** Thread manager needed to be used to manage concurrent agents */
  private Manager manager;

  /** Back trail information */
  private Trail trail;
  
  /** Neighbours hash table, with NestId as key (FIXME move in Gate) */
  private HashMap map;

  /** Neighbours array (FIXME fixed, should be configurable) */
  private Neighbour[] neighbours;

  /** Storage map */
  HashMap storageMap;
  
  /** Pheromone object managed by ants */
  private Object pheromone;
  
  /** Storage */
  private QueryStorage docStorage;
  
  /** Storage */
  private QueryStorage urlStorage;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a new nest.
   */
  public LocalNest(Factory factory)
  {
    if (!checked) {
      checkFactory(factory);
      checked = true;
    }
    
    /* Update factories */
    this.factory = factory;
    clear();
  }

    
  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    /*
     *  Read parameter TTL from the factory object; if the parameter is found, 
     *  the value are updated. Otherwise, the initialization parameter is still valid.
     */
    try {
      this.ttl = factory.intValue(NEST_TTL);
    } catch (Exception e) { }

    /* 
     *  Obtain implementations of basic interfaces. Note that in this class
     *  we do not manage the possibility of changing implementations of
     *  the basic components of nests at runtime. So, we do not need to
     *  maintain the versions of the implementations classes stored here.
     */
    
    /* Gate */
    if (gate == null) {
      gate = (Gate) factory.get(Gate.class);
    } else {
      gate.clear();
    }
    gate.setAntListener(this);
    
    /* RequestSet */
    if (requests == null) {
      requests = (RequestSet) factory.get(RequestSet.class);
    } else {
      requests.clear();
    }

    /* Manager */
    if (manager == null) {
      manager = (Manager) factory.get(Manager.class);
    } else {
      manager.clear();
    }
    
    /* Ant factory */
    if (antf == null) {
      antf = (AntFactory) factory.get(AntFactory.class);
    } else {
      antf.clear();
    }
    
    if (trail == null) {
      trail = (Trail) factory.get(Trail.class);
    } else {
      trail.clear();
    }

    /* Pheromone */
    pheromone = null;

    /* Storage management */ 
    if (storageMap == null) {
      storageMap = new HashMap();
      try {
        docStorage = (QueryStorage) getStorage(Storage.DOC_STORAGE);
        urlStorage = (QueryStorage) getStorage(Storage.URL_STORAGE);
      } catch (Exception e) {
        // Do nothing; we have already done a check
      }
    } else {
      Iterator i = storageMap.values().iterator();
      while (i.hasNext()) {
        Storage storage = (Storage) i.next();
        storage.clear();
      }
    }

    // FIXME Move this part into Gate
    if (map == null) {
      map = new HashMap();
    } else {
      map.clear();
    }
    neighbours = null;
  
  }
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from AntListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Receive an agent from outside and deliver it to the
   *  to the activity manager.
   */
  public void deliver(AntDescriptor desc)
  {
    AntView view = new LocalAntView(this, desc); 
    manager.add(view);
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Nest
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the identifier of this Nest
   */
  public NestId getId()
  {
    return gate.getId();
  }
  
  /**
   *  Return the Manager associated to this Nest
   */
  public Manager getManager()
  {
    return manager;
  }

  /**
   *  Accepts a request from the nest user. Possible responses will be
   *  returned to the specified listener. One or more ants are generated
   *  and assigned to the request.
   *  Accept a resource request originated by an anthill client.
   *  The resource to be searched is identified by name. The method
   *  returns immediately. When the resource is found, a Resource
   *  object will be returned to the specified resource listener.
   *  <p>
   *  Implementation note: If the resource can be found locally, 
   *  the listener is immediately contacted. Otherwise, a new 
   *  <CODE>Ant</CODE> is generated and is delivered to the ant 
   *  manager.
   *
   *  @param request the request to be performed
   *  @param listener The listener object which will be notified
   *    when the request is completed.
   */
  public void request(RequestId rid, Object request, ResponseListener listener)
  {
    /* Creates a set of ants and insert them in the local nest */
    Ant[] ants = antf.getAnts(rid, request, gate.getId());
    for (int i=0; i < ants.length; i++) {
      AntDescriptor desc = new AntDescriptor(gate.getId(), ants[i], ttl);
      AntView view = new LocalAntView(this, desc);
      ants[i].init(view);
      manager.add(view);
    } 
    requests.add(rid, listener);
  }
  
  /**
   *  Insert a new resource in this nest. Delagates
   *  the insertion to the storage.
   * 
   *  @param resource the resource to be inserted.
   */
  public void insert(Resource resource)
  {
    if (docStorage != null) {
      try {
        docStorage.addResource(resource.getKey(), resource);
      } catch (Exception e) {
        e.printStackTrace();
        // There is no enough space in the entire storage!
      }
    }
  }

  
  /**
   *  Add the identifier of a new neighbour nest to the set of 
   *  neighbours known to this <CODE>Nest</CODE>, on behalf of 
   *  an Nest client. 
   *
   *  @param aid neighbour identifier
   */
  public void addClientNeighbour(NestId aid)
  {
    map.put(aid, new Neighbour(aid));
    neighbours = null;
  }
  
  
  /**
   *  Add the identifier of a new neighbour nest to the set of 
   *  neighbours known to this <CODE>Nest</CODE>, on behalf of 
   *  an ant. 
   *
   *  @param aid neighbour identifier
   */
  public void addAntNeighbour(NestId aid)
  {
    map.put(aid, new Neighbour(aid));
    neighbours = null;
  }

  
  /**
   * Remove a neighbour identifier from the set of neighbour
   * @param aid neighbour identifier
   */
  public void removeNeighbour(NestId aid)
  {
    map.remove(aid);
    neighbours = null;
  }
  
  /**
   * Return the neighbours known to this Nest
   */
  public Neighbour[] getNeighbours()
  {
    if (neighbours == null) {
      neighbours = new Neighbour[map.size()];
      Iterator iterator = map.values().iterator();
      for (int i=0; i < neighbours.length; i++) 
        neighbours[i] = (Neighbour) iterator.next();
    }
    return neighbours;
  }
  
  /**
   *  Return the pheromone associated to the link to this
   *  neighbour. The class of the pheromone is left unspecified.
   */
  public Object getPheromone(Class cl)
  {
    // FIXME Add a per-class table
    return pheromone;
  }

  
  /**
   *  Return the pheromone associated to the link to this
   *  neighbour. The class of the pheromone is left unspecified.
   */
  public void setPheromone(Object pheromone, Class cl)
  {
    // FIXME Add a per-class table
    this.pheromone = pheromone;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the <CODE>Gate</CODE> object used by this <CODE>Nest</CODE>
   */
  Gate getGate()
  {
    return gate;
  }
    
  /**
   *  Return a storage object identified by the given name.
   */
  public Storage getStorage(String name)
    throws StorageNotFoundException
  {
    Storage storage = (Storage) storageMap.get(name);
    if (storage == null) {
      try { 
        storage = (Storage) factory.get(name);
        storageMap.put(name, storage);
      } catch (Exception e) { 
        throw new StorageNotFoundException(e); 
      }
    }
    return storage;
  }

  /**
   *  Performs the specified request on the local resource
   *  storage and returns the resources satisfying the
   *  request. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param request The request to be performed
   */
  public Resource[] getResources(Object request)
  {
    if (docStorage != null) {
      return docStorage.getResources(request);
    } else {
      return null;
    }
  }

  /**
   *  Performs the specified request on the local resource
   *  storage and returns the resources satisfying the
   *  request. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param request The request to be performed
   */
  public Resource getResource(Object key)
  {
    if (docStorage != null) {
      return docStorage.getResource(key);
    } else {
      return null;
    }
  }
  
  /**
   *  Add a resource to the nest. This method enables
   *  ants to move resources from one nest to another. 
   *  <p>
   *  Depending on the implementation of <CODE>Resource</CODE> 
   *  and <CODE>Nest</CODE>, adding a resource to a nest through 
   *  an AntView may result to actual copy the content of the 
   *  resource, or to simply insert a reference to the actual 
   *  position of the resource itself. 
   *  <p>
   *  This method may be invoked more than once from an ant, in 
   *  order to insert more resources. Nevertheless, the actual 
   *  insertion of a resource in a nest depends on the policies 
   *  of the particular <CODE>Nest</CODE> implementation. For 
   *  example, a nest may limit the number of resources originated 
   *  by  another nest, in order to avoid "spamming".
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param resource resource to be added
   */
  public void addResource(Resource resource)
    throws CapacityException
  {
    if (docStorage != null) 
      docStorage.addResource(resource.getKey(), resource);
  }

  /**
   *  Return the <CODE>RequestSet</CODE> object used by this <CODE>Nest</CODE>
   */
  RequestSet getRequestSet()
  {
    return requests;
  }
  
  /**
   *  Return the <CODE>RequestSet</CODE> object used by this <CODE>Nest</CODE>
   */
  Trail getPreviousNest(boolean create)
  {
    if (trail == null && create) {
      trail = (Trail) factory.get(Trail.class);
    }
    return trail;
  }

  public String toString()
  {
    StringBuffer buffer = new StringBuffer();
    buffer.append("[Nest: ");
    buffer.append(trail);
    buffer.append(", ");
    buffer.append(requests);
    buffer.append(", [Docs: ");
    buffer.append(docStorage.getCapacity()-docStorage.getRemainingCapacity());
    buffer.append(", Urls: ");
    buffer.append(urlStorage.getCapacity()-urlStorage.getRemainingCapacity());
    buffer.append("]]");
    return buffer.toString();
  }

  /** Read from statistics */
  // FIXME Add comments
  // FIXME Add other methods (?)
  public int intValue(String name)
  {
    return factory.intValue(name);
  }
  
  
} // END LocalNest
