/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import anthill.*;
import anthill.nest.*;
import anthill.factory.*;
import anthill.util.*;

import java.util.*;


/**
 *  Class <CODE>LocalTrail</CODE> is the implementation of <CODE>Trail</CODE>
 *  to be used during off-line simulations. Programmers are enabled to specify
 *  the maximum number of associations (ant identifier, previous nest identifier)
 *  which can be stored in a <CODE>LocalTrail</CODE> object, or the maximum duration 
 *  of such an associations in a <CODE>LocalTrail</CODE> object, or both.
 *  Assocations which exceed the maximum capacity or are older than the maximum
 *  duration are discarded from this <CODE>LocalTrail</CODE>.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class LocalTrail
implements Trail
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  The identifier of the configuration parameter containing the 
   *  maximum number of pairs (ant id, nest id) that may be stored
   *  in a <CODE>Trail</CODE> object.
   */
  public static final String TRAIL_MAXSIZE = "TrailMaxSize";

  /** 
   *  The identifier of the configuration parameter containing the
   *  maximum time of permanence of a pair (ant id, nest id) in a
   *  <CODE>Trail</CODE> object.
   */
  public static final String TRAIL_MAXTIME = "TrailMaxTime";

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;
  
  /** Max size of the HashList */
  private int maxsize;

  /** Max duration for storing a key in the HashList */
  private int maxtime;

  /** Hash list implementing the LRU policy */
  HashList list;
 
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Instantiate a new empty <CODE>Trail</CODE> object. The maximum number of
   *  pairs (ant id, source) that may be contained in it, and the maximum duration
   *  of pairs are read through the factory object.
   *
   *  @param factory the factory object used to obtain the configuration
   *    values
   */
  public LocalTrail(Factory factory)
  {
    this.factory = factory;
    clear();
  }
  
  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    /*
     *  Read maximum time and size from the factory
     */
    maxtime = factory.intValue(TRAIL_MAXTIME);
    maxsize = factory.intValue(TRAIL_MAXSIZE);
  
    if (list == null) {
      /*  Construct a new instance */ 
      list = new HashList(maxsize, maxtime);
    } else {
      /*  Clear current instance */
      list.clear();
      list.setMaxsize(maxsize);
      list.setMaxtime(maxtime);
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Add a pair (id, source) to the trail object containing
   *  the host. Returns true if the ant identifier is not already
   *  included in this Trail object; return false otherwise.
   *  The "false" notification can be used to detect duplicated
   *  visit to a nest.
   */
  public synchronized boolean add(AntId id, NestId source)
  {
    if (id == null)
      throw new IllegalArgumentException("Method LocalTrail.add: The ant identifier could not be null");
    if (list.get(id) == null) {
      list.put(id, source);
      return true;
    } else {
      return false;
    }
  }

  
  /**
   *  Lookup this <CODE>LocalTrail</CODE> object in order to obtain the 
   *  identifier of the nest from which the specified 
   *  ant arrived. The pair associating the ant
   *  identifier to the source nest is removed from
   *  both the hash table and the linked list.
   */
  public synchronized NestId lookup(AntId id, boolean remove)
  {
    if (remove) {
      return (NestId) list.remove(id);
    } else {
      return (NestId) list.get(id);
    }
  }
  

  public String toString()
  {
    StringBuffer buffer = new StringBuffer();
    buffer.append("[Trail Size: ");
    buffer.append(list.size());
    buffer.append("]");
    return buffer.toString();
  }
  
} // END LocalTrail

