/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.storage;

import anthill.*;
import java.util.*;

/**
 *  The <CODE>GenericStorage</CODE> represents a generic storage
 *  which enforce a simple policy based on a maximum capacity.
 *  It contains methods to add, remove and get resources from the
 *  storage.  You may also inspect the content of the storage using
 *  the iterator and check its capacity and usage.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision$
 */
public interface GenericStorage
extends Storage
{

  /**
   *  Returns the resource associated with the specified key. If the key
   *  is not associated with any resource, a null value is returned.
   *
   *  @param key  whose associated resource to retreive.
   *  @return resource with which the specified key is associated.
   */
  public Resource getResource(Object key);

  /**
   *  Returns true if the resource associated to <CODE>key</CODE> is contained in
   *  the storage, false otherwise.
   *
   *  @param key  the key to be searched
   *  @return true if the resource associated to <CODE>key</CODE> is contained in
   *    the storage
   */
  public boolean containsResource(Object key);

  /**
   *  Adds a resource to the storage and associates it to the specified
   *  key. If the remaining capacity of the storage is not sufficient to
   *  store the resource, a <CODE>CapacityException</CODE> is thrown
   *  containing information about the remaining capacity. Method
   *  <CODE>getContentLength()</CODE> may be invoked on the resource
   *  to determine the amount of space required by the resource. 
   *  In order to be stored on persistent storage, resource must 
   *  implement the <CODE>Serializable</CODE> interface.
   *
   *  @param key   the key identifying the resource
   *  @param resource  the resource to be added
   *  @throws CapacityException  raised if the given resource object
   *    exceeds the total available capacity for this storage.
   */
  public void addResource(Object key, Resource resource)
    throws CapacityException;


  /**
   *  Removes the resource associated with the specified key
   *  from the storage.
   *
   *  @param key the identifier of the resource to be removed
   */
  public void removeResource(Object key);

  
  /**
   *  Returns the complete set of resources stored in this storage object
   *  using an iterator.
   *
   *  @return iterator representing this storage object, allowing to
   *    inspect its content.
   */
  public Iterator iterator();

  /**
   *  Returns the initial capacity of this generic storage.
   */
  public long getCapacity();

  /**
   *  Returns the remaining capacity of this generic storage.
   */
  public long getRemainingCapacity();

} // END GenericStorage
