/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.storage.impl;

import anthill.*;
import anthill.util.*;
import anthill.factory.*;
import anthill.storage.*;

import java.util.*;

/**
 *  Implementation of an LRU based key storage.  This is a wrapper class
 *  for storing resources based on using a key to access the resources.
 *
 *  @author Hein Meling
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class LRUKeyStorage
implements GenericStorage
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** A generic storage. */
  protected GenericStorage storage;

  /** HashList use to implement the LRU policy of the LRU key storage. */
  protected HashList list;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Contructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public LRUKeyStorage(Factory factory, String storageName)
  {
    /* Obtains the back-end storage object to be used to implement LRU */ 
    storage = (GenericStorage) factory.get(storageName);
    /*  Creates a new LRU list with unbounded capacity */
    list = new HashList(0); // Capacity 0, because this is a mapping
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the GenericStorage interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns true if the resource associated to <CODE>key</CODE> is contained in
   *  the storage, false otherwise.
   *
   *  @param key  the key to be searched
   *  @return true if the resource associated to <CODE>key</CODE> is contained in
   *    the storage
   */
  public boolean containsResource(Object key)
  {
    return list.containsKey(key);
  }

  /**
   *  Returns the resource associated with the specified key. If the key
   *  is not associated with any resource, a null value is returned.
   *
   *  @param key  whose associated resource to retreive.
   *  @return resource with which the specified key is associated.
   */
  public Resource getResource(Object key)
  {
    list.get(key); /* Move the key in the LRU queue */
    return storage.getResource(key);
  }


  /**
   *  Adds a resource to the storage and associates it to the specified
   *  key. If the remaining capacity of the storage is not sufficient to
   *  store the resource, a <CODE>CapacityException</CODE> is thrown
   *  containing information about the remaining capacity. Method
   *  <CODE>size()</CODE> may be invoked on the resource
   *  to determine the amount of space required by the resource. 
   *  In order to be stored on persistent storage, resource must 
   *  implement the <CODE>Serializable</CODE> interface.
   *
   *  @param key   the key identifying the resource
   *  @param resource  the resource to be added
   *  @throws CapacityException  raised if the given resource object
   *    exceeds the total available capacity for this storage.
   */
  public void addResource(Object key, Resource resource)
    throws CapacityException
  {
    boolean inserted = false;
    while (!inserted) {
      /* Iterate, removing resources if needed to fit the new resource,
       * until the resource has been inserted into the storage. */
      try {
        storage.addResource(key, resource);
        list.add(key);
        inserted = true;
      } catch (CapacityException e) {
        if (resource.size() > e.getCapacity()) {
          System.out.println("Sono qui; " + resource.size() + " " + e.getCapacity() + resource);
          throw e;
        } else {
          /* Removing the head of the list corresponds to removing the
           * least recently used entry in the list. */
          Object headKey = list.getHeadKey();
          list.remove(headKey);
          storage.removeResource(headKey);
        }
      }
    }
  }


  /**
   *  Removes the resource associated with the specified key
   *  from the storage.
   *
   *  @param key the identifier of the resource to be removed
   */
  public void removeResource(Object key)
  {
    list.remove(key);
    storage.removeResource(key);
  }


  /**
   *  Returns the complete set of resources stored in this storage object
   *  using an iterator.
   *
   *  @return iterator representing this storage object, allowing to
   *    inspect its content.
   */
  public Iterator iterator()
  {
    return storage.iterator();
  }


  /**
   *  Returns the initial capacity of this generic storage.
   */
  public long getCapacity()
  {
    return storage.getCapacity();
  }


  /**
   *  Returns the remaining capacity of this generic storage.
   */
  public long getRemainingCapacity()
  {
    return storage.getRemainingCapacity();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the Clearable interface (inherited from the Storage marker interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restore the object to its initial state
   */
  public void clear()
  {
    storage.clear();
    list.clear();
  }

} // END LRUKeyStorage

