/*
 * Copyright (c) 1998-2000 The Anthill Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.util;

import java.security.*;
import java.lang.reflect.*;
import java.util.*;

/**
 *  This class can be used to analyze the memory footprint of an object
 *  reference. Memory footprint is computed recursively, by adding the
 *  memory footprint of the fields of the referenced object and the
 *  memory footprint of the objects referenced by these objects.
 *  Furthermore, this class can be used to obtain information about
 *  the classes referenced (directly or indirectly) by an object and 
 *  the number of instances for each classes.
 *  
 *  @author  Alberto Montresor
 *  @version $Revision$
 */
public final class Analyzer
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /*
   *  NOTE: These values have been estimated by reading the white paper
   *  on the Java Hotspot Virtual Machine; they may be not correct, and
   *  not valid for different JVMs.
   */
  
  /** Size of an int */
  private static final int INT_SIZE = 4;
  
  /** Size of a long */
  private static final int LONG_SIZE = 8;

  /** Size of a short */
  private static final int SHORT_SIZE = 2;
  
  /** Size of a byte */
  private static final int BYTE_SIZE = 1;

  /** Size of a char */
  private static final int CHAR_SIZE = 2;

  /** Size of a boolean */
  private static final int BOOLEAN_SIZE = 1;

  /** Size of a float */
  private static final int FLOAT_SIZE = 4;

  /** Size of a double */
  private static final int DOUBLE_SIZE = 8;

  /** Size of a reference */
  private static final int REF_SIZE = 4;

  /** Size of an array length */
  private static final int LENGTH_SIZE = 4;

  /** Size of a additional information associated to an object */
  private static final int OBJECT_SIZE = 0;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the memory footprint of the specified object, computed 
   *  recursively by adding the footprint of the fields of the
   *  object and the footprint of each of the objects referenced
   *  by it.
   */
  public static int footprint(Object obj)
  {
    /*
     *  Invokes doComputeClasses() in order to
     *  obtain information about the footprint; no
     *  information about classes are computed.
     */
    HashSet set = new HashSet();
    return doComputeClasses(obj, set, null);
  }

  /**
   *  Returns a collection of <CODE>InstanceCount</CODE> objects, one for
   *  each of the classes referenced (directly or indirectly) 
   *  by the specified object. <CODE>InstanceCount</CODE> objects may
   *  be used to obtain the number of instances for each of the
   *  classes reference by the specified object.
   */
  public static Collection countClasses(Object obj)
  {
    /*
     *  Invokes doComputeClasses in order to
     *  obtain information about referenced classes
     *  and number of instances.
     */
    HashSet set = new HashSet();
    HashMap map = new HashMap();
    int val = doComputeClasses(obj, set, map);
    return map.values();
  }
  
  /**
   *  Returns a string containing information about the classes 
   *  referenced (directly or indirectly) by the specified object 
   *  and the number of instances for each classes. 
   */
  public static String printClasses(Object obj)
  {
    Collection c = countClasses(obj);
    
    /*
     *  Returns information in the form of a string
     */
    StringBuffer buffer = new StringBuffer();
    Iterator i = c.iterator();
    while (i.hasNext()) {
      InstanceCount ic = (InstanceCount) i.next();
      buffer.append("Class ");
      buffer.append(ic.getType().getName());
      buffer.append(": ");
      buffer.append(ic.getCounter());
      buffer.append(" instances\n");
    }
    return buffer.toString();
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Recursively computes the footprint of an object and returns it.
   *  
   *  @param obj the object to be inspected
   *  @param set the collection of objects which have been already
   *    visited and counted
   *  @param set the map that contains the <CODE>InstanceCount</CODE> objects
   *    used to count the number of instances
   */
  private static int doComputeClasses(Object obj, Set set, Map map)
  {
    if (obj == null)
      return 0;

    /* Total number of bytes reference by this object */
    int tot = OBJECT_SIZE;
    
    if (!set.contains(obj)) {
      set.add(obj);
      Class objClass = obj.getClass();
      
      /*
       *  This code is executed if the list of classes and instances
       *  is requested; 
       */
      if (map != null) {
        InstanceCount ic = (InstanceCount) map.get(objClass);
        if (ic == null) {
          map.put(objClass, new InstanceCount(objClass));
        } else {
          ic.inc();
        }
      } 
      
      if (objClass.isArray()) {
        /* 
         *  Array 
         */
        int len = Array.getLength(obj);
        Class componentType = objClass.getComponentType();
        if (componentType.isPrimitive()) {
          /* 
           *  Array of primitive elements; the size is given by
           *  the length value, plus "len" times the footprint
           *  of the primitive component type
           */
          tot += primitiveLen(componentType) * len + LENGTH_SIZE;
        } else {
          /* 
           *  Array of primitive elements; the size is given by
           *  the length value, plus "len" times the footprint
           *  of a reference, plus the footprint of the objects.
           */
          tot += REF_SIZE * len + LENGTH_SIZE; 
          for (int i=0; i<len; i++) {
            tot += doComputeClasses(Array.get(obj, i), set, map);
          }
        }
      } else {
        /**
         *  The object we are analyzing is a normal object;
         *  we compute the length of each of its fields, obtained
         *  by invoking method getDeclaredFields on the class and
         *  each of its superclasses.
         */
        while (objClass != null) {
          Field[] fields = objClass.getDeclaredFields();
          for (int i=0; i < fields.length; i++) {
            if (!Modifier.isStatic(fields[i].getModifiers())) {
              Class clazz = fields[i].getType();
              if (clazz.isPrimitive()) {
                /* Primitive value */
                tot += primitiveLen(clazz);
              } else {
                /* Reference */
                final Field field = fields[i];
                AccessController.doPrivileged(
                  new PrivilegedAction() {
                    public Object run() {
                      field.setAccessible(true);
                      return null;
                    }
                  }
                );    
                try {
                  Object ref = field.get(obj);
                  tot += REF_SIZE + doComputeClasses(ref, set, map); 
                } catch (IllegalAccessException e) {
                  System.out.println(e.getMessage());
                  // e.printStackTrace();
                }
              }
            }
          }
          /* 
           *  Get declared fields return also private fields, but only those
           *  declared in this class; for the other fields, we need to recurse
           *  class definitions using getSuperclass()
           */
          objClass = objClass.getSuperclass();
        }
      }
    }
    return tot;
  }

  /**
   *  Returns the footprint of a primitive class 
   */
  private static int primitiveLen(Class clazz) 
  {
    if (clazz.equals(int.class)) {
      return INT_SIZE;
    } else if (clazz.equals(long.class)) {
      return LONG_SIZE;
    } else if (clazz.equals(short.class)) {
      return SHORT_SIZE;
    } else if (clazz.equals(byte.class)) {
      return BYTE_SIZE;
    } else if (clazz.equals(char.class)) {
      return CHAR_SIZE;
    } else if (clazz.equals(boolean.class)) {
      return BOOLEAN_SIZE;
    } else if (clazz.equals(float.class)) {
      return FLOAT_SIZE;
    } else if (clazz.equals(double.class)) {
      return DOUBLE_SIZE;
    } else if (!clazz.isPrimitive()) {
      throw new IllegalArgumentException("Class " + clazz.getName() + " is not primitive");
    }
    // Never reached
    return 0;
  }

  
  /**
   *  Recursively computes the footprint of an object and returns it.
   *  
   *  @param obj the object to be inspected
   *  @param set the collection of objects which have been already
   *    visited and counted
   *  @param set the map that contains the <CODE>InstanceCount</CODE> objects
   *    used to count the number of instances
   */
  private static int doComputeClasses(Object obj, Map map)
  {
    Set done = new HashSet();
    Set todo = new HashSet();
    
    if (obj == null)
      return 0;

    /* Total number of bytes reference by this object */
    int tot = OBJECT_SIZE;
    
    todo.add(obj);
    while (todo.size() > 0) {
      System.out.println("Todo" + todo.size() + " Done: " + done.size() + " Tot " + tot);
      Object scan = todo.iterator().next();
      todo.remove(scan);
      if (!done.contains(scan)) {
        done.add(scan);
        Class objClass = scan.getClass();
        //System.out.println(scan);
        /*
         *  This code is executed if the list of classes and instances
         *  is requested; 
         */
        if (map != null) {
          InstanceCount ic = (InstanceCount) map.get(objClass);
          if (ic == null) {
            map.put(objClass, new InstanceCount(objClass));
          } else {
            ic.inc();
          }
        } 
      
        if (objClass.isArray()) {
          /* 
           *  Array 
           */
          int len = Array.getLength(scan);
          Class componentType = objClass.getComponentType();
          if (componentType.isPrimitive()) {
            /* 
             *  Array of primitive elements; the size is given by
             *  the length value, plus "len" times the footprint
             *  of the primitive component type
             */
            tot += primitiveLen(componentType) * len + LENGTH_SIZE;
          } else {
            /* 
             *  Array of primitive elements; the size is given by
             *  the length value, plus "len" times the footprint
             *  of a reference, plus the footprint of the objects.
             */
            tot += REF_SIZE * len + LENGTH_SIZE; 
            for (int i=0; i<len; i++) {
              todo.add(Array.get(scan, i));
            }
          }
        } else {
          /**
           *  The object we are analyzing is a normal object;
           *  we compute the length of each of its fields, obtained
           *  by invoking method getDeclaredFields on the class and
           *  each of its superclasses.
           */
          while (objClass != null) {
            Field[] fields = objClass.getDeclaredFields();
            for (int i=0; i < fields.length; i++) {
              if (!Modifier.isStatic(fields[i].getModifiers())) {
                Class clazz = fields[i].getType();
                if (clazz.isPrimitive()) {
                  /* Primitive value */
                  tot += primitiveLen(clazz);
                } else {
                  /* Reference */
                  final Field field = fields[i];
                  AccessController.doPrivileged(
                    new PrivilegedAction() {
                      public Object run() {
                        field.setAccessible(true);
                        return null;
                      }
                    }
                  );    
                  try {
                    Object ref = field.get(scan);
                    tot += REF_SIZE;
                    todo.add(ref);
                  } catch (IllegalAccessException e) {
                    System.out.println(e.getMessage());
                    // e.printStackTrace();
                  }
                }
              }
            }
            /* 
             *  Get declared fields return also private fields, but only those
             *  declared in this class; for the other fields, we need to recurse
             *  class definitions using getSuperclass()
             */
            objClass = objClass.getSuperclass();
          }
        }
      }
    }
    return tot;
  }
  
} // END Analyzer


