/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.util;

import java.io.*;

/**
 *  The class <CODE>KeyHash</CODE> computes the digest (or hash)
 *  of a specified string keyword, and enables various other classes
 *  to compare keyword hashes and store them in a <CODE>HashMap</CODE> 
 *  or other collections in need of a hashCode method.
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class KeyHash
extends SHADigest
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // State Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The keyword for this <code>KeyHash</code> object. */
  protected String keyword;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public KeyHash()
  {
  }

  /**
   *  Creates a new <code>KeyHash</code> instance based on a randomly
   *  generated SHA-1 hash key.
   *
   *  @param secure if true, the secure random generator will be used;
   *    otherwise the standard random generator is used.
   */
  public KeyHash(boolean secure) 
  {
    this(secure, null);
  }

  /**
   *  Compute a SHA-1 hash value for the specified keyword, or compute
   *  a random SHA-1 identifier if the keyword is equal to null.
   *
   *  @param keyword  A string keyword to be hashed and stored in this object.
   */
  public KeyHash(String keyword)
  {
    this(false, keyword);
  }

  /**
   *  Compute a SHA-1 hash value for the specified keyword, or compute
   *  a random SHA-1 identifier if the keyword is equal to null.
   *
   *  @param secure if true, the secure random generator will be used;
   *    otherwise the standard random generator is used.
   *  @param keyword  a string keyword to be hashed and stored in this object.
   */
  public KeyHash(boolean secure, String keyword)
  {
    if (keyword != null) {
      computeDigest(keyword);
      this.keyword = keyword;
    } else {
      initRandom(secure);
      computeRandomSHA();
      /*
       * Use the computed random digest as stringified keyword,
       * since the given keyword was <code>null</code>.
       */
      this.keyword = rdigest.toString(16);
    }
    computeHash();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the keyword associated with this <code>KeyHash</code> object.
   *
   *  @return keyword  the keyword string associated with this <code>KeyHash</code>.
   */
  public String getKeyword()
  {
    return keyword;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(": ");
    buf.append(keyword);
    buf.append(" -> ");
    buf.append(rdigest.toString(16));
    buf.append("]");
    return buf.toString();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals this <CODE>KeyHash</CODE> to the specified output stream.
   *
   *  @param in  an <code>ObjectInput</code> stream object
   *  @exception IOException  if an error occurs
   */
  public void readExternal(ObjectInput in)
    throws IOException
  {
    keyword = in.readUTF();
    super.readExternal(in);
  }

  /**
   *  Marshals this <CODE>KeyHash</CODE> to the specified output stream.
   *
   *  @param out  an <code>ObjectOutput</code> stream object
   *  @exception IOException  if an error occurs
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeUTF(keyword);
    super.writeExternal(out);
  }

} // END KeyHash
