/*
 * Copyright (c) 1998-2000 The Anthill Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.util;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  This class contains a set of static methods for the management
 *  of log files. The log information can be printed on both the
 *  standard output and a log file, whose name is generated using
 *  the name of the host in which the log class is running and
 *  the current time. Different log level may be associated to
 *  both the standard output and the log file, in order to obtain
 *  different set of information.
 *
 *  @author  Alberto Montresor
 */
public class Log
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Debug activation / deactivation
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * True if the log function is active.
   * Putting this variable to false and using an optimizing compiler such as Jikes
   * guarantee that all log code will be removed from the generated bytecode; this to
   * avoid overhead in the execution when log is not needed.
   */
  public static final boolean DEBUG = true;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** True if the log is activated (default) */
  private static boolean activated = true;

  /** Debugging level (console) */
  private static int clevel = Integer.MAX_VALUE;

  /** Debugging level (file) */
  private static int flevel = Integer.MAX_VALUE;

  /** Log writer */
  private static PrintWriter out;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initializer
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  Reads the log properties and initializes the opportune variables
   */
  static {
    
    // Read if the log function is activated or not
    String disableString = System.getProperty("anthill.log.disable");
    if (disableString != null)
      activated = false;

    String cString = System.getProperty("anthill.log.outLevel");
    if (cString != null) {
      try {
        clevel = Integer.parseInt(cString);
      } catch (NumberFormatException e) {
      }
    }
    String fString = System.getProperty("anthill.log.fileLevel");
    if (fString != null) {
      try {
        flevel = Integer.parseInt(fString);
      } catch (NumberFormatException e) {
      }
    }
    String dirString = System.getProperty("anthill.log.dir");
    if (dirString == null) 
      dirString = ".";
    GregorianCalendar gc = new GregorianCalendar();
    StringBuffer buffer = new StringBuffer();
    buffer.append("log-");
    try {
      String name = InetAddress.getLocalHost().toString();
      int t1 = name.indexOf('.');
      int t2 = name.indexOf('/');
      t1 = (t1 < t2 ? t1 : t2);
      buffer.append(name.substring(0, t1));
      buffer.append("-");
    } catch (Exception e) {
      // Problems when determining the host name; we skip it
    }
    buffer.append(Integer.toString(gc.get(GregorianCalendar.YEAR)));
    buffer.append(Integer.toString(gc.get(GregorianCalendar.MONTH)+100).substring(1));
    buffer.append(Integer.toString(gc.get(GregorianCalendar.DAY_OF_MONTH)+100).substring(1));
    buffer.append("-");
    buffer.append(Integer.toString(gc.get(GregorianCalendar.HOUR)+100).substring(1));
    buffer.append(Integer.toString(gc.get(GregorianCalendar.MINUTE)+100).substring(1));
    buffer.append(".txt");

    try {
      File dir = new File(dirString);
      if (!dir.exists()) {
        /* Directory does not exist, create it. */
        if (!dir.mkdir())
          throw new IOException("Couldn't create the specified log directory");
      }
      out = new PrintWriter(new FileOutputStream(new File(dir, buffer.toString())));
    } catch (Exception e) {
      e.printStackTrace();
      activated = false;
    }
  }
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initializer
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static void printError()
  {
    try {
      throw new Exception("Log method called when Log is not active");
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Set the file debug level to <t>level</t>.
   * @param level the new logging level.
   */
  public static void setFileLevel(int level)
  {
    if (!DEBUG) {
      printError();
      return;
    }
    if (out != null)
      flevel = level;
  }

  /**
   * Set the console log level to <t>level</t>.
   * @param level the new logging level.
   */
  public static void setOutLevel(int level)
  {
    if (!DEBUG) {
      printError();
      return;
    }
    if (out != null)
      clevel = level;
  }
  
  
  /**
   * Prints the specified information on the standard output. If the log is
   * active, prints also the information on the log file.
   *
   * @param string the string to be printed.
   */
  public static void printFile(Object object)
  {
    if (DEBUG && activated) {
      if (out != null) {
        out.print(object);
        out.flush();
      }
    }
  }

  /**
   * Prints the specified information on the standard output. If the log is
   * active, prints also the information on the log file.
   *
   * @param obj the object to be printed.
   */
  public static void print(Object obj)
  {
    if (DEBUG && activated) {
      System.out.print(obj);
      if (out != null) {
        out.print(obj);
        out.flush();
      }
    }
  }

  /**
   * Prints the specified information on the standard output and terminates
   * the line. If the log is active, prints also the information on the
   * log file and terminates the line.
   *
   * @param string the string to be printed.
   */
  public static void printlnFile(Object object)
  {
    if (DEBUG && activated) {
      if (out != null) {
        out.println(object);
        out.flush();
      }
    }
  }

  /**
   * Prints the specified information on the standard output and terminates
   * the line. If the log is active, prints also the information on the
   * log file and terminates the line.
   *
   * @param obj the object to be printed.
   */
  public static void println(Object obj)
  {
    if (DEBUG && activated) {
      System.out.println(obj);
      if (out != null) {
        out.println(obj);
        out.flush();
      }
    }
  }

  /**
   * If the log is not active, print an exception on the standard output.
   * If the log is active, prints the specified information on the log
   * file; if <t>level</t> is greater or equal than the current log level,
   * prints the specified information on the standard output as well.
   *
   * @param level the log level of this information
   * @param string the string to be printed.
   */
  public static void printFile(int level, Object object)
  {
    if (!DEBUG)
      printError();
    if (DEBUG && activated) {
      if (out != null) {
        if (level <= flevel) {
          out.print(object);
        out.flush();
        }
      }
    }
  }

  /**
   * If the log is not active, print an exception on the standard output.
   * If the log is active, prints the specified information on the log
   * file; if <t>level</t> is greater or equal than the current log level,
   * prints the specified information on the standard output as well.
   *
   * @param level the log level of this information
   * @param obj the object to be printed.
   */
  public static void print(int level, Object obj)
  {
    if (!DEBUG)
      printError();
    if (DEBUG && activated) {
      if (out != null) {
        if (level <= clevel)
          System.out.print(obj);
        if (level <= flevel) {
          out.print(obj);
          out.flush();
        }
      }
    }
  }

  /**
   * If the log is not active, print an exception on the standard output.
   * If the log is active, prints the specified information on the log
   * file and terminates the line; if <t>level</t> is greater or equal than
   * the current log level, prints the specified information on the standard
   * output as well and terminates the line.
   *
   * @param level the log level of this information
   * @param string the string to be printed.
   */
  public static void printlnFile(int level, Object object)
  {
    if (!DEBUG)
      printError();
    if (DEBUG && activated) {
      if (out != null) {
        if (level <= flevel) {
          out.println(object);
          out.flush();
        }
      }
    }
  }

  /**
   * If the log is not active, print an exception on the standard output.
   * If the log is active, prints the specified information on the log
   * file and terminates the line; if <t>level</t> is greater or equal than
   * the current log level, prints the specified information on the standard
   * output as well and terminates the line.
   *
   * @param level the log level of this information
   * @param obj the object to be printed.
   */
  public static void println(int level, Object obj)
  {
    if (!DEBUG)
      printError();
    if (DEBUG && activated) {
      if (out != null) {
        if (level <= clevel)
          System.out.println(obj);
        if (level <= flevel) {
          out.println(obj);
          out.flush();
        }
      }
    }
  }

  /**
   * Print the stack trace of the specified exception on the standard output;
   * if the log function is active, print the stack trace on the log file
   * as well.
   *
   * @param e the exception to be printed.
   */
  public static void printStackTrace(Exception e)
  {
    e.printStackTrace();
    if (DEBUG && activated) {
      if (out != null) {
        e.printStackTrace(out);
        out.flush();
      }
    }
  }

  public static void flush()
  {
    out.flush();
  }
  
  /**
   * If the log is not active, prints an exception on the standard output.
   * If the log is active, prints the stack trace of the specified exception
   * on the log file; if <t>level</t> is greater or equal than the current
   * log level, prints the specified the stack trace on the standard output
   * as well.
   *
   * @param e the exception to be printed.
   */
  public static void printStackTrace(int level, Exception e)
  {
    if (!DEBUG) 
      printError();
    if (DEBUG && activated) {
      if (out != null) {
        if (level <= clevel)
          e.printStackTrace();
        if (level <= flevel) {
          e.printStackTrace(out);
          out.flush();
        }
      }
    }
  }

} // END Debug

