/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.util;

import java.util.*;


/**
 *  This class implements a HashList, which is a combination of a HashMap
 *  and a doubly linked list. The linked listed is always in sorted order
 *  according to the least recently used (LRU) scheme.
 *
 *  @author Hein Meling
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class MapList
implements Map, Runnable, Clearable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Max size of the map */
  private int maxsize;

  /** Size of the map */
  private int size;

  /** 
   *  A map mapping a key identifier to a composite <CODE>Pair</CODE>
   *  object implementing a doubly linked list.
   */
  protected Map map;
  
  /** Head of the list connecting the pairs; pairs will be removed here */
  private Pair head;
  
  /** Tail of the list connecting the pairs; pairs will be added here */
  private Pair tail;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Thread related fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Max duration for storing a key in the map */
  private int maxtime;
  
  /** Polling interval */
  private int interval;

  /** True if the pooling thread has to be stopped */
  private boolean ended;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiate a new MapList object by setting the maximum number of
   *  composite <CODE>Pair</CODE> objects which may be contained in it.
   *
   *  @param maxsize  maximum number of pair objects which may
   *    be contained in this MapList. If zero, there is no upper
   *    limit to the the size.
   *  @param maxtime  maximum number
   */
  public MapList(int maxsize, int maxtime)
  {
    size = 0;
    this.maxsize = maxsize;
    this.maxtime = maxtime;
    this.interval = maxtime / 10;
    ended = false;
    if (maxtime > 0) {
      Thread thread = new Thread(this);
      thread.start();
    }
  }


  /**
   *  Instantiate a new MapList object by setting the maximum number of
   *  key, value pairs which may be contained in it.
   *
   *  @param maxsize  Maximum number of objects which may be contained in
   *    this MapList. If zero, there is no upper limit to the the size.
   */
  public MapList(int maxsize)
  {
    size = 0;
    this.maxsize = maxsize;
    this.maxtime = 0;
    this.interval = 0;
  }
    
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  // Below are some required Map methods that must be implemented.
  // These are not correct or completly implemented yet.
  public boolean isEmpty()
  {
    return map.isEmpty();
  }
  
  public boolean containsKey(Object key)
  {
    return map.containsKey(key);
  }
  
  public boolean containsValue(Object val)
  {
    // THIS IS STRAIGHT FORWARD. MUST LOOK THROUGH THE LINKED LIST.
    // BECAUSE THE val IS HIDDEN IN THE PAIR OBJECT.
    throw new UnsupportedOperationException("Not implemented yet.");
  }
  
  public void putAll(Map t)
  {
    // Iterate the Map object and use our put.
    throw new UnsupportedOperationException("Not implemented yet.");
  }
  
  public void clear()
  {
    while(head != null) {
      map.remove(head.key);
      removePair(head);
      size--;
    }
  }   

  public ArrayList keyList()
  {
    ArrayList lruOrder = new ArrayList(size);
    Pair current = tail;
    for (int i = 0; i < size; i++) {
      lruOrder.add(current.key);
      current = current.prev;
    }
    return lruOrder;
  }


  public Set keySet()
  {
    return map.keySet();
  }

  public Collection values()
  {
    //Note, it might be more efficient to iterate the linked list?
    return map.values();
  }


  public Set entrySet()
  {
    return map.entrySet();
  }

  /**
   *  Add the given key to the map with <CODE>null</CODE> as value.
   *
   *  @param key  key with which to associate null.
   *  @return  previous value associated with the specified key, or null.
   */
  public Object add(Object key)
  {
    return put(key, null);
  }


  /**
   *  Add the <CODE>key, value</CODE> pair to the map.
   *
   *  @param key    key with which the specified value is to be associated.
   *  @param value  value to be associated with the specified key.
   *  @return  previous value associated with specified key, or <CODE>null</CODE>
   *    if there was no mapping for key.  A <CODE>null</CODE> return can also
   *    indicate that the map previously associated <CODE>null</CODE> with the
   *    specified key, if the implementation supports <CODE>null</CODE> values.
   * 
   */
  public synchronized Object put(Object key, Object value)
  {
    Pair pair = (Pair) map.get(key);
    if (pair == null) {
      /* Check if there is a maximum size for the data structure */
      if (maxsize > 0) {
        while(size >= maxsize) {
          /* If the list is not empty, remove the oldest element */
          if (head != null) {
            map.remove(head.key);
            removePair(head);
            size--;
          }
        }
      }
      
      /* Add a pair containing the value object to the list */
      pair = new Pair(key, value);
      addPair(pair);
      size++;

      /* Add the corresponding association to the map */
      map.put(key, pair);
      
      return null;
    } else {
      // FIXME What behavior here?
      return pair.value;
    }
  }


  /**
   *  Gets the <CODE>value</CODE> associated with the specified <CODE>key</CODE>.
   *  <P>
   *  This method moves the accessed object to the tail of the
   *  linked listed enforcing the LRU policy for elements in the Map.
   *  This means that for every access to a given key, its element
   *  is moved to the tail of the list. This is because the most
   *  recent keys will appear at the tail part of the list, while the
   *  least recently used keys will appear at the head part of the list.
   *  Hence elements should be removed from the head of the list in
   *  order to reclaim memory consumed by the key item.
   *
   *  @param key  key whose associated value is to be returned.
   *  @return the value to which this map maps the specified key.
   */
  public synchronized Object get(Object key)
  {
    Pair pair = (Pair) map.get(key);
    if (pair != null) {
      /* Update the access time for the specified 'pair' */
      // pair.time = System.currentTimeMillis();
      /* Remove 'pair' from its current position in the list */
      removePair(pair);
      /* Reinsert 'pair' at the tail. */
      pair.next = null;
      pair.prev = tail;
      addPair(pair);
      /* Return the value object associated with the specified key */
      return pair.value;
    } else {
      return null;
    }
  }

  /**
   *  Removes the specify key and its associated value from the map list.
   *
   *  @param key  The key whose assoicated value should be removed
   *    from the map list.
   *  @return  The removed objects value, null if the specified
   *    key is not stored in the map list.
   */
  public synchronized Object remove(Object key)
  {
    Pair pair = (Pair) map.get(key);
    if (pair != null) {
      /* Remove the key from the Map */
      map.remove(key);
      /* Remove the pair from the list */
      removePair(pair);
      /* Update size */
      size--;
    }
    /* Return the value object associated with the specified key */
    return pair.value;
  }

  /**
   *  The key of head in the list, which can be used to remove the head
   *  from the list; suitable for implementing an LRU policy.
   *
   *  @return key of the head element, which will be null of the list is empty.
   *
   */
  public synchronized Object getHeadKey()
  {
    return head.key;
  }

  /**
   *  Removes the specify number of elements from the head
   *  of the linked list and also from the map.
   *
   *  @param elemToRemove  The number of elements to be removed.
   *
   */
  public synchronized Object[] removeChunck(int elemToRemove)
  {
    Object[] ret = new Object[elemToRemove];
    for (int i = 0; i < elemToRemove; i++) {
      ret[i] = remove(head.key);
    }
    return ret;
  }


  /**
   *  Returns the number of key-value mappings in this map.
   *
   *  @return the number of key-value mappings in this map.
   */
  public int size()
  {
    return size;
  }

  /**
   *  Modify the maximum size of this <CODE>MapList</CODE>.
   *  If this MapList contains more elements than the
   *  new maximum size, removes the additional elements.
   *  
   *  @param maxsize the new maximum size
   */
  public void setMaxsize(int maxsize)
  {
    this.maxsize = maxsize;
    if (maxsize > 0) {
      while(size >= maxsize) {
        /* If the list is not empty, removes the oldest elements */
        if (head != null) {
          map.remove(head.key);
          removePair(head);
          size--;
        }
      }
    }
  }
  
  /**
   *  Return the maximum size of the current site.
   */
  public int getMaxsize()
  {
    return maxsize;
  }

  /**
   *  Modify the maximum time of this <CODE>MapList</CODE>.
   *  
   *  @param maxtime the new maximum time
   */
  public void setMaxtime(int maxtime)
  {
    boolean start = (maxtime > 0 && this.maxtime == 0);
    boolean stop  = (maxtime == 0 && this.maxtime > 0);
    this.maxtime = maxtime;
    if (start) {
      ended = false;
      Thread thread = new Thread(this);
      thread.start();
    } else if (stop) {
      ended = true;
    } 
  }
  
  /**
   *  Return the maximum size of the current site.
   */
  public int getMaxtime()
  {
    return maxtime;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Run and related methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Periodically cleans up the staled elements in the map.
   */
  public void run()
  {
    while (!ended) {
      clean();
      try {
        Thread.sleep(interval);
      } catch (InterruptedException e) {
        
      }
    }
  }

  /**
   *  Close the thread.
   */
  public void end()
  {
    ended = true;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add the pair at the end of the list 
   */
  private void addPair(Pair pair)
  {
    pair.next = null;
    pair.prev = tail;
    if (head == null)
      head = pair;
    if (tail != null) 
      tail.next = pair;
    tail = pair;
  }
  
  /**
   *  Remove the pair from the list
   */
  private void removePair(Pair pair)
  {
    if (pair.prev != null) {
      /* Add a link to the next in the previous */
      pair.prev.next = pair.next;
    } else {
      /* No prev, modify the head */
      head = pair.next;
    }
    if (pair.next != null) {
      /* Add a link to the previous in the next */
      pair.next.prev = pair.prev;
    } else {
      /* No next, modify the tail */
      tail = pair.prev;
    }
  }


  /** 
   * Scans the linked list of pairs in order
   * to remove staled pairs (key, value).
   */
  private synchronized void clean()
  {
    //long expiration = System.currentTimeMillis() - maxtime;
    //while (head != null && head.time < expiration) {
    //  remove(head.key);
    //}
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner class
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  The Pair class is an inner class used to link objects
   *  stored in the map into a list.
   */
  protected class Pair
  {
    
    /** The key used for accessing the value. */
    protected Object key;
    
    /** The value stored by this object */
    protected Object value;
    
    /** Next pair in the linked list */
    protected Pair next;
    
    /** Previous pair in the linked list */
    protected Pair prev;

    /** Creation time; used to detect stale objects */
    // protected long time;
    
    /**
     *  Initializing constructor, to be inserted at the tail,
     *  thus requires you to specify the previous tail.
     */
    protected Pair(Object key, Object value)
    {
      this.key = key;
      this.value = value;
      // this.time = System.currentTimeMillis();
    }
    
  } // END Pair
  
} // END MapList

