/*
 * Copyright (c) 1998-2000 The Anthill Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.util;

import java.lang.reflect.*;
import java.security.*;

/**
 *  Classes extending <CODE>Printable</CODE> automatically inherit
 *  a <CODE>toString()</CODE> method which prints out their contents.
 *  i.e. the value of the fields contained in the object instances.
 *  Note that <CODE>toString()</CODE> in a <CODE>Printable</CODE> object 
 *  may generate infinite loops if the reference graph obtained starting 
 *  from the object contains a cycle of <CODE>Printable</CODE> objects.
 *
 *  @author  Alberto Montresor
 *  @version $Revision$
 */
public class Printable
{

  /**
   *  <CODE>toString</CODE> method to print out the content of
   *  an object.
   */
  public String toString()
  {
    return toString("");
  }
  
  /**
   *  Auto-indenting, recursive <CODE>toString()</CODE> method to 
   *  print out the content of an object.
   */
  public String toString(String indent)
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    Class cl = this.getClass();
    buf.append(cl.getName());
    Field[] fields = cl.getDeclaredFields();
    if (fields.length > 0) {
      buf.append(":\n");
      for (int i=0; i < fields.length; i++) {
        buf.append(indent+"  ");
        buf.append(fields[i].getName());
        buf.append(" = ");

        /*
         *  Read the content of fields, even if private, with
         *  a privileged action.
         */
        final Field field = fields[i];
        Object obj = null;
        AccessController.doPrivileged(new PrivilegedAction() {
          public Object run() {
            field.setAccessible(true);
            return null;
          }
        });
        try {
          obj = field.get(this);
        } catch (Exception e) {
          e.printStackTrace();
        }
        try {
          Printable prn = (Printable) obj;
          buf.append(prn.toString(indent+"  "));
        } catch (Exception e) {
          buf.append(obj);
        }
        buf.append("\n");
      }
      buf.append(indent);
    }
    buf.append("]");    
    return buf.toString();
  }

} // END Printable

