/*
 * Copyright (c) 1998-2000 The Anthill Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.util;

/**
 *  The <code> Queue</code> class implements a queue with last-in, first-out
 *  semantics. The queue is implemented as a linked list of buckets, each of
 *  them contains a fixed number of objects. Appended objects are inserted in
 *  the last bucket (if full, a new bucket is allocated), while removed buckets
 *  are extracted from the first bucket.
 *
 *  @author  Alberto Montresor
 *  @version $Revision$
 */
public final class Queue
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private final static int QUEUE_BUCKET_SIZE = 32;      // Bucket size

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private Bucket                head;           // First queue bucket
  private Bucket                tail;           // Last queue bucket
  private int                   size;           // Number of object in a single bucket
  private int                   first;          // Current object index in the first bucket
  private int                   last;           // Current object index in the last elem
  private int                   ncount;         // N. of elements inserted in the queue


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Allocates a queue with bucket size equal to <code> size <\code>.
   *
   * @param size        the size of internal buckets.
   */
  public Queue(int size)
  {
    this.size = size;
    head   = new Bucket(size);
    tail   = head;
    first  = 0;
    last   = 0;
    ncount = 0;
  }

  /**
   * Allocates a queue with default bucket size
   */
  public Queue()
  {
    this(QUEUE_BUCKET_SIZE);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Appends <code> object <\code> at the end of the queue.
   * @param object      the object to be inserted
   */
  public void add(Object object)
  {
    tail.objects[last] = object;
    last++;
    if (last == size) {
      if (tail.next == null)
        tail.next = new Bucket(size);
      tail = tail.next;
      last = 0;
    }
    ncount++;
  }

  /**
   * Returns true if the queue is empty.
   *
   * @return true if the queue is empty
   */
  public boolean isEmpty()
  {
    return (ncount == 0);
  }

  /**
   * Returns the number of elements contained in this queue
   */
  public int size()
  {
    return ncount;
  }

  /**
   * Returns the first element of the queue.
   *
   * @return the first element of the queue; null if the queue is empty.
   */
  public Object get()
  {
    if (ncount == 0)
      return null;
    return head.objects[first];
  }

  /**
   * Removes the first element of the queue.
   *
   * @return the first element of the queue; null if the queue is empty.
   */
  public Object remove()
  {
    Object              ret;            // Return value
    Bucket              temp;           // Temporary reference

    if (ncount == 0)
      return null;
    ret = head.objects[first];
    head.objects[first] = null;
    ncount--;
    first++;
    if (first == size) {
      temp = tail.next;
      tail.next = head;                 // This avoid useless allocations
      head = head.next;
      tail.next.next = temp;
      first = 0;
    }
    return ret;
  }

  /**
   * Add the elements of the addend queue at the end of this queue.
   *
   * @param addend      the queue to be added
   */
  public void merge(Queue addend)
  {
    Object obj;

    while ((obj = addend.remove()) != null)
      add(obj);
  }


  /**
   * Removes all the elements from the queue.
   */
  public void clear()
  {
    for (int i=first; i < size; i++)
      head.objects[i] = null;
    head.next = null;
    tail = head;
    first  = 0;
    last   = 0;
    ncount = 0;
  }

  /**
   * Copies the element of the queue in an array of objects, in the queue order.
   */
  public Object[] toArray()
  {
    Object[] ret = new Object[ncount];
    Bucket scan = head;
    int pos = first;
    for (int i=0; i < ncount; i++) {
      ret[i] = scan.objects[pos];
      pos++;
      if (pos == size) {
        pos = 0;
        scan = scan.next;
      }
    }
    return ret;
  }



  /**
   *  The <code> Bucket</code> class represents the container class used in
   *  queue to avoid expensive allocations and to save memory.
   *
   *  @author  Alberto Montresor
   *  @version $Revision$
   */
  class Bucket {

    //////////////////////////////////////////////////////////////////////////////////////////
    // Fields
    //////////////////////////////////////////////////////////////////////////////////////////

    /** Stored objects */
    Object[]            objects;

    /** Next element */
    Bucket              next;


    //////////////////////////////////////////////////////////////////////////////////////////
    // Constructor
    //////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Constructs a QueueBucket with the specified size
     *
     * @param size      number of elements in a bucket.
     */
    Bucket(int size)
    {
      this.objects = new Object[size];
      this.next = null;
    }

  }  // END Bucket

  //////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String [] args)
  {
    Queue queue = new Queue();
    for (int i=0; i < 20000; i++) {
      System.out.println("Inserted " + i + " elements... list size " + queue.size());
      for (int j=0; j < 1000; j++) {
        Integer in = new Integer(i);
        queue.add(in);
      }
      System.out.println("Inserted " + i + " elements... list size " + queue.size());
      for (int j=0; j < 1000; j++) {
        queue.remove();
      }
    }
    
  }

} // END Queue

