/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.util;

import java.math.*;
import java.util.*;
import java.security.*;
import java.io.*;


/**
 *  The class <CODE>SHADigest</CODE> computes the digest (or hash)
 *  of the specified content, and enables various other classes
 *  to compare the digests and store them in a <CODE>HashMap</CODE>
 *  or other collections in need of a hashCode method.
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class SHADigest
implements Comparable, Externalizable
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // State Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** A big integer for the hash code of the specified keyword. */
  protected BigInteger rdigest;

  /** To avoid recomputing the hash of a big integer we save it */
  protected int hashcode;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Name of the random function to be used */
  public static final String RANDOM_TYPE = "SHA1PRNG";

  /** Name of the digest function to be used */
  public static final String DIGEST_TYPE = "SHA";
  
  /** Byte size of the digest values */
  public static final int DIGEST_SIZE = 20;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields and block
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Counter used to generate unique identifiers; secure version is slow. */
  private static SecureRandom secureRandom;

  /** Counter used to generate unique identifiers; unsecure version initialize faster */
  private static Random unsecureRandom;

  /** Proxy for obtaining random numbers. */
  protected static Random random;

  /** Message digest generator used to produce hash values for strings */
  protected static MessageDigest digest;
  
  static {
    try {
      digest = MessageDigest.getInstance(DIGEST_TYPE);
    } catch (NoSuchAlgorithmException e) {
      Util.panic(e);
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public SHADigest()
  {
  }

  /**
   *  Creates a new <code>SHADigest</code> instance based on a randomly
   *  generated SHA hash.
   *
   *  @param secure if true, the secure random generator will be used;
   *    otherwise the standard random generator is used.
   */
  public SHADigest(boolean secure)
  {
    initRandom(secure);
    computeRandomSHA();
    computeHash();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Protected Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Compute a SHA-1 hash value for the specified string
   *
   *  @param str  The string for which to compute a SHA-1 hash.
   */
  protected void computeDigest(String str)
  {
    digest.reset();
    digest.update(str.getBytes()); // FIXME Should we consider character encoding
    /* Convert the digest to BigInteger for computing the hash code. */
    rdigest = new BigInteger(digest.digest());
  }

  /**
   *  Compute a random SHA-1 hash value for a simulated hash.
   */
  protected void computeRandomSHA()
  {
    byte bytes[] = new byte[DIGEST_SIZE]; 
    random.nextBytes(bytes);
    rdigest = new BigInteger(bytes);
  }

  /**
   *  Computes the hash code for the object.
   */
  protected void computeHash()
  {
    hashcode = rdigest.hashCode();
  }

  /**
   *  Initialize a random generator.
   *
   *  @param secure if true, the secure random generator will be used;
   *    otherwise the standard random generator is used.
   */
  protected void initRandom(boolean secure)
  {
    if (random == null) {
      if (secure) {
        if (secureRandom == null) {
          try {
            secureRandom = SecureRandom.getInstance(RANDOM_TYPE);
          } catch (NoSuchAlgorithmException e) {
            Util.panic(e);
          }
        }
        random = secureRandom;
      } else {
        if (unsecureRandom == null)
          unsecureRandom = new Random();
        random = unsecureRandom;
      }
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Comparable
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int compareTo(Object obj)
    throws ClassCastException
  {
    if (obj != null && obj instanceof SHADigest)
      return rdigest.compareTo(((SHADigest) obj).rdigest);
    else 
      throw new ClassCastException("Object is not a " + this.getClass().getName() + " instance");
  }

  public int compareTo(SHADigest kwHash)
  {
    return rdigest.compareTo(kwHash.rdigest);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Check whether this <CODE>SHADigest</CODE> is equal to the specified object.
   */
  public boolean equals(Object obj)
  {
    if (obj != null && obj instanceof SHADigest) {
      return ((SHADigest) obj).rdigest.equals(this.rdigest);
    } else {
      return false;
    }
  }

  /**
   *  Returns a hash code value for the object.
   */
  public int hashCode()
  {
    return hashcode;
  }
  
  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(": ");
    buf.append(rdigest.toString(16));
    buf.append("]");
    return buf.toString();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals this <CODE>SHADigest</CODE> to the specified input stream.
   *
   *  @param in  an <code>ObjectInput</code> stream object
   *  @exception IOException  if an error occurs
   */
  public void readExternal(ObjectInput in)
    throws IOException
  {
    byte[] b = new byte[in.readInt()];
    while (in.read(b) != -1) ;
    rdigest = new BigInteger(b);
    /* Recompute the hashcode after unmarshaling. */
    computeHash();
  }

  /**
   *  Marshals this <CODE>SHADigest</CODE> to the specified output stream.
   *
   *  @param out  an <code>ObjectOutput</code> stream object
   *  @exception IOException  if an error occurs
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    byte[] b = rdigest.toByteArray();
    out.writeInt(b.length);
    out.write(b);
  }

} // END SHADigest
