/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.util;

import java.util.*;


/**
 *  This class implements a TreeList, which is a combination of a TreeMap
 *  and a doubly linked list. The linked listed is always in sorted order
 *  according to the least recently used (LRU) scheme, while the TreeMap
 *  is sorted according to the keys of the map.
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class TreeList
extends MapList
{

  /* Maximum number of values to be returned from the get(key, valcnt) method */
  private static final int MAX_VALUES = 20;

  /**
   *  A proxy for the actual map, used to access methods
   *  from the SortedMap interface.
   */
  protected SortedMap proxyMap;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiate a new TreeList object by setting the maximum number of
   *  key, value pairs which may be contained in it.
   *
   *  @param maxsize  Maximum number of objects which may be contained in
   *    this TreeList. If zero, there is no upper limit to the the size.
   */
  public TreeList(int maxsize, int maxtime)
  {
    super(maxsize, maxtime);
    map = new TreeMap();
    proxyMap = (SortedMap) map;
  }


  /**
   *  Instantiate a new TreeList object by setting the maximum number of
   *  key, value pairs which may be contained in it.
   *
   *  @param maxsize  Maximum number of objects which may be contained in
   *    this TreeList. If zero, there is no upper limit to the the size.
   */
  public TreeList(int maxsize)
  {
    super(maxsize);
    map = new TreeMap();
    proxyMap = (SortedMap) map;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Extended methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a set of object values whose keys closely match that of key.
   *  The valuecount specifies how may values around the key should be
   *  returned. If not enough keys/values can be found, a null entry
   *  is returned in the remaining positions in the returned array.
   *  <P>
   *  Note that the valuecount is bounded within MAX_VALUES.
   *
   *  @param key  key whose closely associated values is to be returned.
   *  @param valuecount  how many values to return.
   *  @return value set matching or closely matching the specified key.
   */
  public synchronized Object[] get(Object key, int valuecount)
  {
    if (valuecount > MAX_VALUES)
      valuecount = MAX_VALUES;
    Object[] val = new Object[valuecount];
    SortedMap hMap = proxyMap.headMap(key);
    SortedMap tMap = proxyMap.tailMap(key);
    Iterator titer = tMap.values().iterator();
    for (int i = 0; i < valuecount; i++) {
      Object lastkey = null;
      try {
        lastkey = hMap.lastKey();
      } catch (NoSuchElementException exception) {
        // The hMap set is empty, do nothing.
      }

      /* First choice: Alternate, selecting keys
       * from the tail or head map, respectively. */
      if ((i % 2) == 0) {
        val[i] = titer.hasNext() ? ((Pair) titer.next()).value : null;
        if (lastkey != null && val[i] == null) {
          /* If there are no more keys in the tMap try the hMap instead. */
          val[i] = ((Pair) hMap.get(lastkey)).value;
          /* Update the hMap to view only unused keys */
          hMap = hMap.headMap(lastkey);
        }
      } else {
        if (lastkey != null) {
          val[i] = ((Pair) hMap.get(lastkey)).value;
          /* Update the hMap to view only unused keys */
          hMap = hMap.headMap(lastkey);
        }
        if (val[i] == null && titer.hasNext())
          val[i] = ((Pair) titer.next()).value;
      }

    } // end of for (int i = 0; i < valuecount; i++)

    return val;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main method -- used for testing purposes
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main (String[] args) {
    TreeList tl = new TreeList(300);
    tl.put("A", "HeinA");
    tl.put("B", "HeinB");
    tl.put("C", "HeinC");
    tl.put("D", "HeinD");
    tl.put("E", "HeinE");
    tl.put("F", "HeinF");
    tl.put("G", "HeinG");
    tl.put("H", "HeinH");
    tl.put("I", "HeinI");
    tl.put("J", "HeinJ");
    tl.put("K", "HeinK");
    tl.put("L", "HeinL");
    tl.put("M", "HeinM");
    tl.put("N", "HeinN");
    tl.put("O", "HeinO");
    System.out.println("get(K) = " + (String) tl.get("K"));
    Object[] vals = tl.get("G", 23);
    for (int i = 0; i < vals.length; i++) {
      System.out.println("val" + i + ": " + (String) vals[i]);
    }
  } // end of main ()

} // END TreeList

