/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package freenetant;
  
import anthill.*;
import anthill.util.*;
import anthill.storage.*;

import java.io.*;
import java.util.*;

/**
 *  @author Andrea Piergallini
 *  @author Luca Vagnozzi
 *  @version $Revision$
 */
public class FreenetAnt implements Ant, Serializable {

  ////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////

  /** Count the number of created ants */
  private static int counter;
  
  /** NHISTORY */
  private static int NHISTORY = 700;
  
  /**
   * Returns an history object. Only 1 out of NHISTORY history objects created
   * will actually perform history logging
   */
  private static History createHistory() {
    ++counter;
    return new History((counter % NHISTORY) == 0);
  }


  ////////////////////////////////////////////////////////////////////////////
  // Initialization Fields
  ////////////////////////////////////////////////////////////////////////////

  /** The identifier of the request */
  private RequestId rid;

  /** Originating nest of this ant */
  private NestId requestor;
  
  /** The query string associated with this ant */
  private String query;

  /** The hash value of the keyword searched by this ant */
  private KeyHash keyhash;

  ////////////////////////////////////////////////////////////////////////////
  // State Fields
  ////////////////////////////////////////////////////////////////////////////

  /** True if this ant is returning to the requestor nest */
  private boolean backward;

  /** True if we are backtracking because we reached a node already visited */
  private boolean backtracking;

  /** The identifier of this ant */
  private AntId aid;
  
  /** The resource found */
  private Resource resource;
  
  /** Copy Probability */
  private int copyProb;

  /** Fake Probability */
  private int fakeProb;

  /** History objects containing the history of the actions performed by this ant*/
  private History history;

  /** Backtrack information */
  Backtrack backtrack;
  
  /** NestId of provider of the resource used with the fake probability. */
  NestId provider;


  ////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////

  /**
   * Instantiate a new FreeAnt by setting the query to be satisfied and the
   * hash value to be searched.
   */
  public FreenetAnt(RequestId rid, String query, int copyProb,
                    int fakeProb, NestId requestor) {
    this.rid = rid;
    this.query = query;
    this.requestor = requestor;

    /* Hash value of the query */
    keyhash = new KeyHash(query);
    
    /** Set state parameters */
    this.copyProb = copyProb;
    this.fakeProb = fakeProb;

    provider = null;
    
    resource = null;
    backward = false;
    backtracking = false;

    backtrack = new Backtrack();

    /* Print debug information */
    history = createHistory();
  }



  ////////////////////////////////////////////////////////////////////////////
  // Methods from Ant
  ////////////////////////////////////////////////////////////////////////////


  /** Configuration parameters for this ant. */
  public void init(AntView view) {
    /** Create an Id for this ant. */
    aid = view.createAntId();
  }

  // Documentation inherited from interface
  public String getSpecies() {
    return "FreenetAnt";
  }


  /**
   *  This method encapsulate the computation performed by the
   *  ant. 
   *
   *  @param view reference to the AntView object which
   *   constitutes the controlled environment of the ant
   */
  public void run(AntView view) {

    /* Debug */
    //history.add("--------------------------------------------------------------");
    //history.add("Search query: " + query + " Current Nest: " + view.getNestId() + 
    //  " Last Nest: " + view.getLastVisited() + " TTL: " +view.getTTL());

    /* Checks of the Ant */
    if (!backward) {
      /* The Ant go forward */
      goForward(view);
    } else {
      /* The Ant go backward to the requestor nest */
      goBackward(view);
    }
  }


  /**
   * Algorithm used when moving forward.
   */
  public void goForward(AntView view) {

    if (backtracking) {
      /*
       * An ant is backtracking if it reaches a node which
       * has already visited or if it reaches a dead node.
       */ 
      doBacktracking(view);
      return;
    } else {
      /*
       * If the freenetant is not backtracking, it has to check if it
       * has already visited this node.
       */
      boolean firsttime = view.storeReturnNest(aid, view.getLastVisited());
      if (!firsttime) {
        /* We have alredy visited this nest */
        try {
          backtracking = true;
          view.move(view.getLastVisited());
        }
        catch (Exception e) {
          // Do nothing
        }
        return;
      }
    }
    
    /* The ant can only go forward. */
    backtracking = false;

    /** Local query of the database. */
    resource = view.getResource(query);
    if (resource != null) {
      /* Found resource; set the provider variable and go backward. */
      provider = view.getNestId();
      //history.add("Found resource: " + resource +" at Nest: " +provider);
      backward = true;
      goBackward(view);
      return;
    }     

    /*                                                                  
     * Obtain the local pheromone in this nest; if not found,          
     * create a new pheromone object and install it locally            
     */                                                                 
    FreenetAntPheromone freenetPheromone = (FreenetAntPheromone) view.getPheromone();
    if (freenetPheromone == null) {
      freenetPheromone = new FreenetAntPheromone(); // default 30
      freenetPheromone.fillRoutingTable(view.getNeighbours());    
      view.setPheromone((Object) freenetPheromone);
    }

    Object[] closeNests = freenetPheromone.getCloseNests(keyhash);

    /**
     * Just forward the ant. This is the case in which the ant is the 
     * first time visiting this nest, and then the nextNestAndex is 
     * surely empty. Then we send the ant to the first element of 
     * closeNests and push the index value of this element in the 
     * nextNestIndex array.
     */
    if (closeNests[0] != null) {
      /* Add a position to the backtrack object. */
      backtrack.push();
      
      /* Set the last element to 0 */
      backtrack.set(0);
      
      //history.add("Forward to: " +closeNests[0]);
      try {
        view.move((NestId) closeNests[0]);
      } catch (Exception e) {
        // Do nothing
      }
    } else {
      /** 
       * This nests have no neigbours. We must do backtracking.
       * FIXME: Attention. This case will never verified because this means 
       * that this nest is sconnected to the others nests of the PeerNet. This
       * is impossible because in the PeerNet simulation the nests have always 
       * a neighbour. I suppose. By the way i leave this peace of code for 
       * completeness. :P
       */
      backtracking = true;
      doBacktracking(view);
    }
    return;

  }//END goForward()


  /* Perform the Backtracking mechanism of the ant. */
  public void doBacktracking(AntView view) {

    // FIXME: Aggiunta ricerca della risorsa in fase di backtracking. 
    /** Local query of the database. */
    resource = view.getResource(query);
    if (resource != null) {
      System.out.println("Qualcosa si trova...:");
      /* Found resource; set the provider variable and go backward. */
      provider = view.getNestId();
      //history.add("Found resource: " + resource +" at Nest: " +view.getNestId());
      backward = true;
      goBackward(view);
      return;
    }     


    /**
     * Obtain the local pheromone in this nest. If not found (impossible, see
     * below), create a new pheromone object and install it locally. 
     */                                                                 
    FreenetAntPheromone freenetPheromone = (FreenetAntPheromone) view.getPheromone();
    if (freenetPheromone == null) { 
      /* This doesn't never be possible because the ant is backtracking. */
      freenetPheromone = new FreenetAntPheromone(); // default 30
      freenetPheromone.fillRoutingTable(view.getNeighbours());    
      view.setPheromone((Object) freenetPheromone);
    }                                                                   

    Object[] closeNests = freenetPheromone.getCloseNests(keyhash);

    int i = backtrack.get();
    i++;
    

    if ((i < closeNests.length) && (closeNests[i] != null)) {
      /* The ant can move to another (next) nest. */
      backtracking = false;
      backtrack.set(i);
      //history.add("Forwarding to: " +closeNests[i]);
      try {
        view.move((NestId) closeNests[i]);
      } catch (Exception e) {
        // Do nothing
      }
    } else {
      /* There aren't others (next) nests. Backtracking. */
      backtracking = true;
      try {
        /* Pop of the last element in the backtrack object. */
        backtrack.pop();
        NestId prev = view.getReturnNest(aid, true);
        //history.add("Backtracking to: " +prev);
        view.move(prev);
      } catch (Exception e) {
        // Do nothing
      }
    }
  } // END doBacktracking()



  /**
   * Algorithm used when the ant go backward.
   */
  public void goBackward(AntView view) {

    /* 
     * The ant come back to the requestor nest following the return path, and
     * update the routing table in each nest in the return path.
     */
    NestId prev = view.getReturnNest(aid, true);
    
    if (prev == view.getNestId()) {
      /**
       * We have reached the requestor nest. Store the resource in the
       * datastore.
       */

      //history.add("Requestor Nest reached.");
      //history.print();    
      
      FreenetAntPheromone freenetPheromone = (FreenetAntPheromone) view.getPheromone();

      if (freenetPheromone != null) {
        /* This is always true. */
        if ((provider != null) && (provider != view.getNestId())){
          /* Add the entry for this keyhash in the routing table. */
          freenetPheromone.addKeyhash(keyhash, provider);
        } else {
          /*
           * FIXME: This maybe impossible trough the fact that provider was
           * set when the resource was found.
           */
          System.out.println("Non e' impossibile " + provider);
          freenetPheromone.addKeyhash(keyhash, view.getLastVisited());
        }
      }

      view.returnResources(rid, new Resource[] { resource });
        try { 
          view.addResource(resource); 
        } catch (CapacityException e) {
          // Never thrown
        }
    } else if (prev != view.getNestId()) {

      /**
       * Update the routing table in this nest. Update the table with
       * the hash value of the query string and the NestId of the last nest
       * visited by this ant.
       */
      FreenetAntPheromone freenetPheromone = (FreenetAntPheromone) view.getPheromone();

      /**
       * We must control if freenetPheromone is null. This happens when the
       * ant founds the resource in a nest and it is the _first_ ant in this
       * nest. In this case the ant must only go backward.
       */
      if (freenetPheromone != null) {

        if ((provider != null) && (provider != view.getNestId())) {
          /* Add the entry for this keyhash in the routing table. */
          freenetPheromone.addKeyhash(keyhash, provider);
       
          /* Update the routing table adding the entry for this keyhash. */      
          //history.add("Added keyhash: " + keyhash + " from provider: " +
          //            provider);
        } else {
          /*
           * FIXME: This maybe impossible trough the fact that provider was
           * set when the resource was found.
           */
          System.out.println("Non e' impossibile 2" + provider);
          freenetPheromone.addKeyhash(keyhash, view.getLastVisited());
        }
      
        /** Code for fakeProbability */
        if (fakeProb > 0) {
          /* Choose a random integer in the range [1,100]. */
          int coin = (new Random().nextInt(100))+1;

          if (coin <= fakeProb) {
            /* Set the new provider of this resource. */
            provider = view.getNestId();
          } else {
            if (provider == null) {
              /* 
               * FIXME: This maybe impossible trough the fact that the
               * provider was set when the resource was found. Anywhere we set
               * it to the last nest visited by this ant.
               */
              System.out.println("Doveva essere impossibile...");
              provider = view.getLastVisited();
            } else {
              /* Leave untouched the previous provider. */
            }
          }
        } else {
          /**
           * Fake Probability is non positive. Is the same thing if it is 0.
           * Do not set the provider variable to this nest.
           */
        }

        /**
         * If we have a copyProb > 0 then we choose a random integer
         * (coin) in the range [1,100]. If coin <= copyProb then the
         * ant copy the resource in this nest, otherwise do nothing.
         * Then continue backward.
         */
        if (copyProb > 0) {
          
          /* Choose a random integer in the range [1,100]. */
          int coin = (new Random().nextInt(100))+1;

          if (coin <= copyProb) {
            view.returnResources(rid, new Resource[] { resource });
            try { 
              /* Set this nest as provider for this resource. */
              provider = view.getNestId();
              view.addResource(resource);
              //history.add("Copied resource: " + resource + " at Nest: " + view.getNestId()); 
            } catch (CapacityException e) {
              // Never thrown
            }
          } else {
            /* Don't copy the resource in local Storage. */
            //history.add("Resource: " + resource + " NOT COPIED at Nest: " + view.getNestId()); 
          }

        } else {
           /**
           * Copy Probability is non positive. Is the same thing if
           * (copyProb = 0). Do not perform a copy in the intermediate
           * nests.
           */
           //history.add("Resource: " + resource + " NOT COPIED at Nest: " + view.getNestId()); 
        }
      }

      
      /* Pop the first element in the next Nest index. */
      backtrack.pop();
      
      /**
       * Move to the previous hop.
       */
      //history.add("Backward to: " +prev);
      try {
        view.move(prev);
      } catch (Exception e) {
        // Do nothing
      }
    } else {
      /**
       * We have lost the trail.
       */
      System.out.println("Trail lost");
      //history.add("The Ant: "+ aid + " have lost the trail.");
      //history.add("ReturnNest: " +view.getReturnNest(aid, false));
      //history.print();
    }
  }
  

  ////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////

  public void writeExternal(ObjectOutput out)
    throws IOException {
    // FIXME To be implemented
  }

  public void readExternal(ObjectInput in) {
    // FIXME To be implemented
  }

}




