/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package freenetant;

import anthill.*;
import anthill.util.*;
import anthill.factory.*;

import java.util.*;

/**
 *  @author Andrea Piergallini
 *  @author Luca Vagnozzi
 *  @version $Revision$
 */
public class FreenetAntPheromone
{
  /** Default capacity for the routing table. */
  private static final int DEFAULT_CAPACITY  = 60; //16384;

  private int capacity;
  
  /**
   * Oggetto Feromone caratterizzante la Routing Table.
   */
  private TreeList pheromoneRT;

  public FreenetAntPheromone()
  {
    this(DEFAULT_CAPACITY);
  }

  public FreenetAntPheromone(int capacity) {
    /*
     * TreeList associating keyword hash codes and NestIds.
     * TreeList implements a LRU policy, mean that keywords
     * that have been stored may be removed based on the
     * least recently used scheme. This is to make room for
     * new entries in the index. The underlying map is a
     * TreeMap, thus ensuring that keys (key hash codes) are
     * in stored in sorted order.
     */

    /* Create and inizialize the routing table. */
    pheromoneRT = new TreeList(capacity);
    this.capacity = capacity;
  }
  
  
  public void fillRoutingTable(Neighbour[] neighbours) {
    KeyHash keyhash;

    int n = neighbours.length;
    for (int i = 0; i < n; i++) {
      // Random SHA-1 Key    
      keyhash = new KeyHash(null);
      pheromoneRT.put(keyhash, neighbours[i].getId());
    }
  } //End fillRoutingTable 
  
  
  /** 
   * Convert the query string in the corresponding hash value. Then insert
   * this value in the routing table.
   */
  public boolean addKeyword(String keyword, NestId nestid) {
    return addKeyhash(new KeyHash(keyword), nestid);
  }
 
  /**
   *  Add a new nest containing a document that match the specified
   *  keyword, represented by a hash of the keyword.
   *  <P>
   *  !!! Note that this method should perhaps be synchronized because
   *  there may be a race condition if trying to simultaneously add
   *  the same <CODE>keyhash</CODE> with different <CODE>nestid</CODE>s.
   *
   *  @param keyhash  Hash of the keyword represented by a
   *    <CODE>KeywordHash</CODE> object.
   *  @param nestid   Identifier for the nest holding a document
   *    represented by the specified keyword.
   *  @return  <CODE>true</CODE> if the given keyword is already known.
   *    That is, other nests store information about this keyword.
   *    <CODE>false</CODE> is returned if the keyword is unknown.
   */
  public boolean addKeyhash(KeyHash keyhash, NestId nestid) {
    if (pheromoneRT.containsKey(keyhash)) {
      /* Keyword already exists, update this entry with this nest. */
      pheromoneRT.remove(keyhash);
      pheromoneRT.put(keyhash, nestid);
      return true;
    } else {
      /* Keyword doesn't exists, add this entry in the Routing Table */
      pheromoneRT.put(keyhash, nestid);
      return false;
    }
  }


  /**
   * Construct and return an array containing the neighbours nests for this
   * node ordered by closeness of the value of the key (keyhash).
   */
  public Object[] getCloseNests(KeyHash keyhash) {
    
    /*
     * Obtain a set of nests satisfying the closeness relation,
     * sorted according to their closeness to keyhash. The first
     * element of the array may be equal to the keyhash. 
     */
    Object[] closes = pheromoneRT.get(keyhash, pheromoneRT.size());
    ArrayList list = new ArrayList();
    
    for (int i = 0; i < closes.length; i++) {
      list.add(closes[i]);
    }

    Object[] closeNests = list.toArray();
    
    // Stampa di prova. Stampa OK.
    
//    System.out.println("CLOSENESTS");
//    for (int i = 0; i < closeNests.length; i++) {
//      System.out.println("i: " +i +" elem: " +closeNests[i]);
//    }

//    System.out.println("CLOSES");
//    for (int i = 0; i < closes.length; i++) {
//      System.out.println("i: " +i +" elem: " +closes[i]);
//    }
    
    return closeNests;
  }
 

  /** Check if the routing table is empty or not. */  
  public boolean isEmpty() {
    return pheromoneRT.isEmpty();
  }
} // END FreenetAntPheromone
                                                                                       
