/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.storage.*;
import anthill.util.*;
import java.io.*;
import java.util.*;

/**
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class BackwardGnutant
implements Ant, Serializable
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The provider of these resources; originator of this ant */  
  private NestId provider;

  /** The hash value of the keyword searched by this ant */
  private KeyHash keyhash;

  /** Resources that we wish to put update the distributed index with */
  private Resource[] resources;

  /** 
   *  Variable path contains the list of visited nodes, allowing the ant
   *  to avoid visiting the same nests multiple times.
   */
  private ArrayList path;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization
   */
  public BackwardGnutant()
  {
  }

  /**
   *  Initializes a new BackwardGnutant with the set of resources to
   *  be used for updating the distributed index.
   */
  public BackwardGnutant(KeyHash keyhash, Resource[] resources, ArrayList path)
  {
    /*
     *  We remove the provider; since we don't want to move to the
     *  originator of this ant.
     */
    this.provider = (NestId) path.remove(path.size()-1);
    this.keyhash = keyhash;
    this.resources = resources;
    this.path = path;
    //System.out.println("Going back with " + resources.length + " resources!");
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  // Documentation inherited by interface
  public String getSpecies()
  {
    return "Gnutant";
  }

  public void init(AntView view)
  {
  }
  
  /**
   *  This method encapsulate the computation performed by the ant.
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
    /*
     *  Updates the routing information for the keyhash it is searching.
     */
    GnutantPheromone pher = (GnutantPheromone) view.getPheromone();
    if (pher != null && !provider.equals(view.getNestId())) {
      /* Add the keyhash and the provder nest to the local pheromone table. */
      pher.addKeyhash(keyhash, provider);
    }
    
    /*
     *  Add resources to the system  
     */
    QueryStorage storage = (QueryStorage) view.getStorage(Storage.URL_STORAGE);
    if (storage == null)
      throw new IllegalStateException("URL Storage not found");
    for (int i=0; i < resources.length; i++) {
      try {
        storage.addResource(resources[i].getKey(), resources[i]);
      } catch (CapacityException e) {
        System.out.println("Size: " + resources[i].size());
        System.err.println(e);
      }
    }

    /*
     *  Move to the predecessor of the current nest by removing the last entry
     *  in the path for each move. This will shrink the path as we backtrack.
     */
    int pathsize = path.size();
    if (pathsize > 0) {
      try { 
        view.move((NestId) path.remove(pathsize-1)); 
        return;
      } catch (Exception e) { 
        /* Do nothing; should be improved in a real implementation to
         * move to the next nest in the path. But since our simulation
         * currently does not deal with faulty nests this should never
         * happen. */
      } 
    }

    /*
     * We have reached the originating nest; we commit suicide. 
     */
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////////////////////


} // END BackwardGnutant

