/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.util.*;

import java.io.*;
import java.net.*;
import java.math.*;
import java.security.*;

/**
 *  Probabilistic identifier for a document.
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
class DocumentKey
extends SHADigest
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // State Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Length of the document */
  protected int length;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**  The buffer size to use for computing the message digest. */
  private static final int BUFSIZE = 8192;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  DocumentKey()
  {
  }

  /**
   *  Creates a <code>DocumentKey</code> instance for real deployment purposes.
   *  It will read the content of the given URL and compute the digest and length
   *  components comprising the <code>DocumentKey</code>.
   *
   *  @param url  the URL for which we wish to compute a <code>DocumentKey</code>.
   */
  DocumentKey(URL url)
    throws IOException
  {
    computeDocumentKey(url);
    computeHash();
  }


  /**
   *  Creates a deterministic <code>DocumentKey</code> instance for simulation
   *  purposes. The URL path string is used to compute the digest and
   *  length, which comprises the document key. This computation is done
   *  deterministically.
   *
   *  @param urlPath  the URL path; can be fake since no checks are made.
   */
  DocumentKey(String urlPath)
  {
    computeDocumentKey(urlPath);
    computeHash();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Protected Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Reads data from the specified URL and computes the length and digest
   *  components that comprise the <code>DocumentKey</code>.
   *
   *  @param url  the URL from which to read the data needed to compute
   *    the <code>DocumentKey</code>.
   *  @exception IOException if an error occurs
   */
  protected void computeDocumentKey(URL url)
    throws IOException
  {
    digest.reset();
    DigestInputStream dis = new DigestInputStream(url.openStream(), digest);
    /* Read the resource content into the scratch array,
     * while updating the digest; updating is done automagically. */
    byte[] b = new byte[BUFSIZE];
    for (int len = 0; (len = dis.read(b, 0, BUFSIZE)) != -1; length += len) ;
    if (length <= 0)
      throw new IOException("The specified URL has zero length: " + url);
    /* Convert the digest to BigInteger for computing the hash code. */
    rdigest = new BigInteger(digest.digest());
  }

  /**
   *  Computes a fake <code>DocumentKey</code> based on the path component
   *  of the specified URL. This is useful for simulation purposes only.
   *
   *  @param urlPath  the path component of a URL.
   */
  protected void computeDocumentKey(String urlPath)
  {
    computeDigest(urlPath);
    length = urlPath.length() * 10000;
  }

  /**
   *  Computes the hash code for this <code>DocumentKey</code>.
   *  <P>
   *  Note that we wish to avoid computing the hashcode of a big integer
   *  every time, so therefore we compute the hashcode prior to its use
   *  and store it for future calls to the hashcode method.
   */
  protected void computeHash()
  {
    hashcode = rdigest.hashCode() ^ (new Integer(length)).hashCode();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the length component of the <code>DocumentKey</code>.
   *
   *  @return the length component of the <code>DocumentKey</code>.
   */
  public int getLength()
  {
    return length;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object - Overriding those of SHADigest
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Check whether this <CODE>DocumentKey</CODE> is equal
   *  to the specified object.
   *
   *  @param obj  the reference object with which to compare.
   *  @return true  if this object is the same as the obj
   *     argument; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (obj != null && obj instanceof DocumentKey) {
      DocumentKey dockey = (DocumentKey) obj;
      if (dockey.length != length)
        return false;
      return dockey.rdigest.equals(this.rdigest);
    } else {
      return false;
    }
  }

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(": ");
    buf.append(length);
    buf.append(", ");
    buf.append(rdigest.toString(16));
    buf.append("]");
    return buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals this <CODE>DocumentKey</CODE> to the specified output stream.
   *
   *  @param in  an <code>ObjectInput</code> stream object
   *  @exception IOException  if an error occurs
   */
  public void readExternal(ObjectInput in)
    throws IOException
  {
    length = in.readInt();
    super.readExternal(in);
  }

  /**
   *  Marshals this <CODE>DocumentKey</CODE> to the specified output stream.
   *
   *  @param out  an <code>ObjectOutput</code> stream object
   *  @exception IOException  if an error occurs
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(length);
    super.writeExternal(out);
  }

} // END DocumentKey
