/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.storage.*;
import anthill.util.*;
import java.io.*;
import java.util.*;

/**
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision$
 */
public class Gnutant
implements Ant, Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Count the number of created ants */
  private static int counter;
  
  /** NHISTORY */
  // private static int NHISTORY = 10000;
  
  /**
   *  Returns an history object. Only 1 out of NHISTORY history objects created
   *  will actually perform history logging
  private static History createHistory()
  {
    ++counter;
    return new History((counter % NHISTORY) == 0);
  }
  */

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initialization fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The identifier of the request */
  private RequestId rid;

  /** The query string associated with this ant */
  private String query;
  
  /** The hash value of the keyword searched by this ant */
  private KeyHash key;
  
  /** Originating nest of this ant */
  private NestId requestor;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // State fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Current ant state regarding movement */
  private boolean forward;
    
  /** The Ant identifier associated to the response ant */
  private AntId responseId;
  
  /** The Ant identifier associated to the response ant */
  private AntId trailId;

  /** Backtrack information */
  Backtrack backtrack;

  /** Resource to be returned */
  HashSet resources;

  /** First nests to provide new info FIXME currently just one */
  NestId provider;
  
  /** History objects containing the history of the actions performed by this ant */
  // History history;
  
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  This parameter represents the probability for the requestor identifier
   *  to be substituted with the current nest, in order to provide anonymity.
   */ 
  private int fakeProb;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiate a new Gnutant by setting the query to be satisfied and the
   *  hash value to be searched.
   */
  public Gnutant(RequestId rid, String query, KeyHash key, NestId requestor)
  {
    
    /* Store initialization parameters */
    this.rid = rid;
    this.query = query;
    this.key = key;
    this.requestor = requestor;
    
    /* Set state parameters */
    forward = true;
    backtrack = new Backtrack();
    resources = new HashSet();

    // /* Print debug information */
    // history = createHistory();
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Ant
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  // Documentation inherited from interface
  public String getSpecies()
  {
    return "Gnutant";
  }

  
  /**
   *  FIXME: add this method to AntView; add the invocation of this
   *  method in the local nest. Note that this method should be used only for
   *  ant configuration, so we may want to disable move and addAnt
   *  in view.
   */
  public void init(AntView view)
  {
    trailId = view.createAntId();
  }
  
  /**
   *  This method encapsulate the computation performed by the
   *  ant. 
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
    /* Debug */
    // history.add("------------------------------------------------------------------------------");
    // history.add("Search " + query + " Current: " + view.getNestId() + 
    //  " Last: " + view.getLastVisited() + " TTL " +view.getTTL());

    /* Elements from which to start when moving */
    int start = 0; 
    
    if (forward) {
      /********************************************************************************
       *  We arrived to this nest by moving forward
       ********************************************************************************/

      /*  Store trail information to return back; if we have already
       *  visited this node, return back immediately 
       */
      boolean firsttime = view.storeReturnNest(trailId, view.getLastVisited());
      if (!firsttime) {
        backtrackBase(view);
      } else {
        /* Perform a local check of the storages */
        checkStorage(view);
  
        /* (Possibly) store information for ResponseGnutant */
        doStoreResponseInfo(view);
      }
  
    } else {
      /********************************************************************************
       *  We arrived to this nest by backtracking
       ********************************************************************************/

      /* We start from the next nest, whatever it is */
      start = backtrack.get()+1;
      forward = true;
    }

    if (view.getTTL() == 0) {
      /* If we are at the end of the path, send back responses */
      sendResponse(view, resources);
    } else if (forward) {
      /* Otherwise, select the next nest and performs a move to it */
      selectNest(view, start);
    }
    
  }
  

  /**
   *  Method invoked when going forward looking for
   *  documents
   */
  private void selectNest(AntView view, int start)
  {
    /*
     *  Obtains the local pheromone contained in this nest; if not found,
     *  creates a new pheromone object and installs it locally
     */
    GnutantPheromone pher = (GnutantPheromone) view.getPheromone();
    if (pher == null) {
      pher = new GnutantPheromone(view);
      view.setPheromone(pher);
    }
    
    /* 
     *  Get the a set of potential next nests --
     *  based on the existance of the given keyhash
     *  or according to the closeness relation to the keyhash 
     */
    ArrayList list = pher.getNextNest(key, null);
    StringBuffer buf = new StringBuffer();
    for (int i=0; i < list.size(); i++)
      buf.append(list.get(i)+" ");
    // history.add(buf.toString());
    // history.add(pher);
    
    /*
     *  If start is equal to zero, this is the first time
     *  we are visiting the site. So, we need to add a
     *  position to the nest backtrack list.
     */
    if (start == 0)
      backtrack.push();
    
    for (int i=start; i < list.size(); i++) {
      backtrack.set(i);
      // history.add("backtrack: " + i);
      NestId nest = (NestId) list.get(i);
      if (nest != view.getNestId()) {
        try {
          view.move(nest);
          //history.add("GoForward: moved to " + nest);
          return;
        } catch (IOException e) { 
          // history.add("GoForward: move failed to " + nest);
        }
      } else {
        // history.add("WARNING: move to the same nest");
      }
    }
    
    /*
     *  If the ant execution reaches this position, this means that all of the
     *  nests reachable from this nest are dead ends; thus, the node itself is
     *  a dead end and we must backtrack.
     */
    backtrackRecursive(view);
  }
    
  /**
   *  This method is invoked by ants when they reach a node
   *  which has been already visited; in this case, the ants simply
   *  returns to the nest from which it arrived (obtained through
   *  <CODE>view.getLastVisited()</CODE>)
   */
  private void backtrackBase(AntView view)
  {
    /* We put forward to false; this information will be used in the next nest */
    forward = false;

    /* We return to the previous nest */
    NestId prev = view.getLastVisited();
    if (prev != view.getNestId()) {
      try {
        view.move(prev);
        // history.add("backtrackBase: moved to " + view.getLastVisited());
      } catch (Exception e) { 
        // history.add("backtrackBase: move failed to " + view.getLastVisited());
        // history.print();
      }
    } else {
      // history.add("The originating nest of this ant is a dead end");
      // history.print();
    }
  }
  
  
  /**
   *  This method is invoked by ants when all the nests reachable from
   *  its current nest are dead-ends. In this case, the current nest
   *  is a dead end and we must return back.
   */
  private void backtrackRecursive(AntView view)
  {
    /*  We put forward to false; this information will be used in the next nest,
     *  that is, the one we are backtracking to. */
    forward = false;
    
    /*  
     *  We remove backtrack information related to this nest, as we can't reach
     *  any other nest from here 
     */
    try {
      backtrack.pop();
    } catch (IllegalStateException e) {
      // history.add("Illegal pop");
      // history.print(); 
    }
    /* We return to the return nest, as stored in the nest */
    NestId prev = view.getReturnNest(trailId, false);
    if (prev == view.getNestId()) {
      // history.print();
      // history.add("The originating nest of this ant is a dead end");
    } else if (prev == null) {
      // history.add("Ant " + trailId + " has lost the trail!");
      // history.print();
    } else {
      try {
        view.move(prev);
        // history.add("backtrackRec: moved to " + view.getLastVisited());
      } catch (Exception e) { 
        // history.add("backtrackRec: move failed to " + view.getLastVisited());
        // history.print();
      }
    }
  }

  /**
   *  
   */
  private void checkStorage(AntView view)
  {
    QueryStorage docStorage = (QueryStorage) view.getStorage(Storage.DOC_STORAGE);
    QueryStorage urlStorage = (QueryStorage) view.getStorage(Storage.URL_STORAGE);

    /* Debug*/
    // history.add("Doc size: " + docStorage.getRemainingCapacity() + " Url size" + urlStorage.getRemainingCapacity());
    // Iterator it = urlStorage.iterator();
    // while (it.hasNext()) {
    //  history.add(it.next());
    //}

    /* Create an empty array list for the new resources */
    HashSet newResources = new HashSet();
    
    /* Add resources found in the doc storage */
    if (docStorage != null) {
      Resource[] docs = docStorage.getResources(query);
      if (docs != null) {
        for (int i=0; i < docs.length; i++) {
          if (resources.add(docs[i]))
            newResources.add(docs[i]);
        }
      }
    }
    
    /* Add resources found in the url storage */
    if (urlStorage != null) {
      Resource[] urls = urlStorage.getResources(query);
      if (urls != null) {
        for (int i=0; i < urls.length; i++) {
          if (resources.add(urls[i]))
            newResources.add(urls[i]);
        }
      }
    }

    /*  
     *  If there are new resources, send a ResponseGnutant and (possibly) set 
     *  the originating nest
     */
    if (newResources.size() > 0) {
      //if (provider == null)
        provider = view.getNestId();
      // sendResponse(view, newResources);
      Resource[] news = new Resource[newResources.size()];
      newResources.toArray(news);
      ResponseGnutant rgant = new ResponseGnutant(rid, news, requestor, responseId);
      view.addAnt(rgant);
    }
  }

  private void sendResponse(AntView view, HashSet resources)
  {
    if (resources.size() > 0) {
      Resource[] res = new Resource[resources.size()];
      resources.toArray(res);
      // history.add("Resources found:");
      // for (int i=0; i < res.length; i++)
      //  history.add(res[i]);
      // TrailGnutant tgant = new TrailGnutant(provider, key, res, trailId, history);
      TrailGnutant tgant = new TrailGnutant(provider, key, res, trailId, null);
      view.addAnt(tgant);
    }
  }
  
  
  /**
   *  Leave information for the ResponseGnutant ant to
   *  return to the requestor nest. Note that the
   *  response gnutant may follow a shorter path than
   *  the gnutant going back 
   */
  private void doStoreResponseInfo(AntView view)
  {
    if (fakeProb > 0) {
      if (responseId == null) {
        /* 
         *  We create another id for returning the response ant 
         */
        responseId = view.createAntId();
      }
      /*  
       *  Before substituing the requestor, stores some information
       *  needed to return to it
       */
      view.storeReturnNest(responseId, requestor);
      /* 
       *  Substitute the requestor
       */
      requestor = view.getNestId();
    }
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    // FIXME To be implemented
  }
  
  public void readExternal(ObjectInput in)
  {
    // FIXME To be implemented
  }
  
} // END Gnutant


