/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.nest.*;
import anthill.nestor.*;
import anthill.factory.*;
import anthill.storage.*;
import anthill.antsim.*;
import anthill.antsim.impl.*;
import anthill.util.*;

import java.util.*;

/**
 *  Class <CODE>ScenarioImpl</CODE> is a simple implementation of
 *  interface <CODE>Scenario</CODE> in which a predefined set of
 *  resources is inserted in the network and then a certain number
 *  of requests are performed and simulate. A <CODE>ScenarioImpl</CODE>
 *  is composed of a <CODE>Peernet</CODE> and a <CODE>ResourceSet</CODE> 
 *  object. Each resource included in the <CODE>ResourceSet</CODE> is inserted
 *  before starting the simulation in the <CODE>Peernet</CODE>. Then,
 *  the <CODE>ResourceSet</CODE> is used to generate a certain number
 *  of requests to be presented to the nests composing the network.
 *  
 *  
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision$
 */
public class GnutantScenario
implements Scenario, ResponseListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  String name of the parameter used to determine the number of rounds which
   *  will be performed in this scenario.
   */
  public static final String SCENARIO_ROUNDS = "ScenarioRounds";

  /** 
   *  String name of the parameter used to determine the inital number of documents
   *  to be installed at each nest.
   */
  public static final String INITAL_DOCS = "InitialDocumentsPerNest";

  /** 
   *  String name of the parameter used to determine the percentage of documents
   *  to download upon receiving a set of documents as reply from the SearchGnutant.
   */
  public static final String DOCS_TO_DL = "PercentageOfDocumentsToDownload";

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator */
  private static Random random = new Random();

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;
  
  /** Number of rounds to be performed */
  protected int rounds;
  
  /** Number of requests to be issued per round */
  protected int nrequests;

  /** Keywords filename */
  protected String kwFile;

  /** Initial documents per nest */
  protected int initDocsPrNest;

  /** Percentage of the satisfied requests to download */
  protected int docsDL;

  /** Stats */
  protected Stats stats;

  /** Responses */
  protected ResponseMonitor monitor;

  /** Peernet on which the genetic code will be evaluated */
  protected Peernet net;
  
  /** Version identifier of the peer network; when init params change, the version increase */
  protected int netVersion;

  /** Keywords holder. */
  protected Keywords keywords;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor / reinitialization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates a new <code>GnutantScenario</code> instance.
   *
   *  @param factory  the factory object used to generate the peernet
   *    and the resource set contained in a scenario.
   *  @param rounds  the number of rounds of the run.
   *  @param kwFile  keyword file path.
   */
  public GnutantScenario(Factory factory, int rounds, String kwFile)
  {
    this(factory, rounds, 1, kwFile);
  }

  /**
   *  Creates a new <code>GnutantScenario</code> instance.
   *
   *  @param factory  the factory object used to generate the peernet
   *    and the resource set contained in a scenario.
   *  @param rounds  the number of rounds of the run.
   *  @param kwFile  keyword file path.
   */
  public GnutantScenario(Factory factory, int rounds, int nrequests, String kwFile)
  {
    /* Initializes variables */
    this.factory = factory;
    this.rounds = rounds;
    this.nrequests = nrequests;
    this.kwFile = kwFile;

    /* Invokes clear*/
    clear();
  }

  /**
   *  Restores the object to its initial state. The network is re-initialized
   *  as an empty network.
   */
  public void clear()
  {
    int version; 

    /*
     *  Read parameter ScenarioRounds from the factory object; if the parameter 
     *  is found, the value is updated. Otherwise, the initialization parameter
     *  is still valid.
     */
    try {
      rounds = factory.intValue(SCENARIO_ROUNDS);
    } catch (Exception e) {
    }
    /*
     *  Read parameter InitialDocumentsPerNest from the factory object; if the
     *  parameter is found, the value is updated. Otherwise, the initialization
     *  parameter is still valid.
     */
    try {
      initDocsPrNest = factory.intValue(INITAL_DOCS);
    } catch (Exception e) {
    }
    /*
     *  Read parameter PercentageOfDocumentsToDownload from the factory object;
     *  if the parameter is found, the value is updated. Otherwise, the 
     *  initialization parameter is still valid.
     */
    try {
      docsDL = factory.intValue(DOCS_TO_DL);
    } catch (Exception e) {
    }

    /* 
     * Creates a new peernet iff (i) there is no peernet currently; 
     * (ii) the peernet initialization data have changed.
     * Otherwise, clear existing one.
     */
    version = factory.getVersion(Peernet.class);
    if (net == null || netVersion != version ) {
      netVersion = version;
      net = (Peernet) factory.get(Peernet.class);
    } else {
      net.clear();
    }

    /*
     * Compute the total number of documents to insert.
     * and read the keywords file, with at least that number of
     * lines.  We re-read the keywords file for each call to the
     * clear method since the network topology may have changed
     * and also the inital number of documents per nest; thus
     * we may need more keywords.
     */
    int documents = net.size() * initDocsPrNest;
    try {
      keywords = new Keywords(kwFile, documents);
    } catch (java.io.FileNotFoundException e) {
      /* Well, I suppose we should abort. */
      Util.panic(e);
    }

    /* Insert a set of documents at each nest to start with. */
    insertPhase(false, false);

    if (monitor ==  null) {
      monitor = new ResponseMonitor();
    } else {
      monitor.clear();
    }
  }
   
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void insertPhase(boolean insant, boolean random)
  {
    /* Inserts generated resources in nests */
    Nest[] nests = net.getNests();

    for (int index = 0, i = 0; i < nests.length; i++) {
      for (int j = 0; j < initDocsPrNest; j++) {

        String keyw;
        if (random) {
          /* Get a keyword string at random */
          keyw = keywords.getRndKeyword();
        } else {
          /* Get a keyword string sequencially */
          keyw = keywords.getKeyword(index++);
        }

        /* Set the filename to insert. */
        //        String fileName = Integer.toString(index);
        String fileName = keyw;

        /* Construct URL path string. */
        StringBuffer buf = new StringBuffer();
        buf.append("anthill://");
        buf.append(nests[i].getId());
        buf.append("/");
        buf.append(fileName);
        String url = buf.toString();

        /* Insert documents in the nests document storage */
        UrlResource dr = new UrlResource(url);
        nests[i].insert(dr);

        if (insant) {
          /*
           * Perform initial insertion on the P2P network; document upload.
           * Create insert request based on a url and keywords
           */
          Object request = new InsantRequest(url, keyw);
          nests[i].request(null, request, null);
        }
      }
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Scenario
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Set the statistics object which will be used throughout
   *  the scenario evaluation.
   */
  public void setStats(Stats stats)
  {
    this.stats = stats;
    monitor.setStatistics(stats);
  }
  

  /**
   *  Perform the evaluation of this scenario object by simulating it.
   *  The simulation is subdivided in rounds. At each iteration,
   *  a new request is generated and assigned to one of the nest
   *  included in the simulated peer network, which generates an ant
   *  for it. Furthermore, for each  nest, an ant included in the 
   *  manager queue is selected and its <CODE>run()</CODE> method
   *  is executed. Statistics about the simulated execution of the
   *  code are collected in the <CODE>Stats</CODE> object specified
   *  through method <CODE>setStats()</CODE>.
   */
  public void evaluate()
  {
    // Obtain information about the network
    Nest[] nests = net.getNests();
    Manager[] managers = new Manager[nests.length];
    for (int i=0; i < nests.length; i++) {
      managers[i] = nests[i].getManager();
    }

    // Execute the required rounds 
    for (int i=0; i < rounds; i++) {
      monitor.tick();
      
      for (int j=0; j < nrequests; j++) {
        /* Choose random nest and make a request */
        int nest = random.nextInt(nests.length);
        /* We generate a request based on the keywords file. */
        Object request = (Object) keywords.getRndKeyword();
        Rid rid = new Rid(nests[nest]);
        monitor.startRequest(rid);
        nests[nest].request(rid, request, this);
      }
      
      /* Each nest executes the code of an ant in its queue */
      for (int j=0; j < managers.length; j++)
        managers[j].execute();
    }
    
    // Debug information
    for (int i=0; i < nests.length; i++) {
      Log.println(400, nests[i]);
    }
    // logStorageContent(false, false, false);
  }

  /**
   *  Log the content of the UrlStorage and/or Pheromone table at all the
   *  nests in the network.
   *
   *  @param doc if true, log the content of the DocumentStorage tables.
   *  @param url if true, log the content of the UrlStorage tables.
   *  @param phero if true, log the content of the pheromone tables.
   */
  private void logStorageContent(boolean doc, boolean url, boolean phero)
  {
    if (doc || url || phero) {
      // Obtain information about the network
      Nest[] nests = net.getNests();

      /* Log the content of the pheromone table and/or UrlStorage at each nest. */
      for (int i = 0; i < nests.length; i++) {
        LocalNest nest = (LocalNest) nests[i];

        if (doc) {
          Log.println(600, "DOCSTORE " + nest.getId());
          GenericStorage docStore = (GenericStorage) nest.getStorage(Storage.DOC_STORAGE);
          Iterator iter = docStore.iterator();
          while (iter.hasNext())
            Log.println(600, "  " + iter.next());
        }

        if (url) {
          Log.println(600, "URLSTORE " + nest.getId());
          GenericStorage urlStore = (GenericStorage) nest.getStorage(Storage.URL_STORAGE);
          Iterator iter = urlStore.iterator();
          while (iter.hasNext())
            Log.println(600, "  " + iter.next());
        }

        if (phero) {
          Log.println(600, "PHEROMONE " + nest.getId());
          Log.println(600, nest.getPheromone(GnutantPheromone.class));
        }
      }
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ResponseListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upcall method used to notify this listener that the request 
   *  associated to it has returned the specified set of resources.
   * 
   *  @param resources the resources found.
   */
  public void returnResources(RequestId rid, Resource[] resources)
  {
    /*
     * The timeout should be less than the rounds because after
     * 'rounds' iterations we may not receive more responses for
     * the specified 'rid'.
     */
    int timeout = rounds / 4;
    if (monitor.returnResources(rid, resources, timeout))
      downloadResources(rid);
  }

  private void downloadResources(RequestId rid)
  {
    Nest nest = ((Rid) rid).getNest();
    GenericStorage docStore = (GenericStorage) nest.getStorage(Storage.DOC_STORAGE);
    GenericStorage urlStore = (GenericStorage) nest.getStorage(Storage.URL_STORAGE);

    /* Eliminate resources with identical document id; keep only one */
    HashMap uniqueResp = new HashMap();
    Set responses = monitor.responses(rid);
    Iterator iter = responses.iterator();
    while (iter.hasNext()) {
      Resource r = (Resource) iter.next();
      DocumentKey dockey = (DocumentKey) r.getKey();
      uniqueResp.put(dockey, r);
    }
    Collection uresp = uniqueResp.values();
    iter = uresp.iterator();

    /* Compute the number of resources to download;
     * docsDL gives the percentage of resources to download. */
    for (int downloads = (responses.size() * docsDL) / 100;
         iter.hasNext() && downloads > 0; downloads--) {

      /* Look for resources to download; enter the loop only if there are more
       * downloads to be performed (downloads > 0) */
      Resource r = (Resource) iter.next();
      DocumentKey dockey = (DocumentKey) r.getKey();
      String fileName = ((UrlResource) r).getFile();
      if (docStore.containsResource(dockey) || urlStore.containsResource(dockey)) {
        Log.println(600, "LR: " + fileName);
        /*
         * Since we don't download local resources, we must compensate for the
         * <code>downloads</code> counter being decremented in the for-loop;
         * due to the continue statement.  Note that this does not cause
         * a potentially infinite loop since we exit the for-loop if the
         * <code>iter</code> is empty.
         */
        downloads++;
        /* The resource was found locally, we skip to the next resource. */
        continue;
      }

      /* Construct URL path string. */
      StringBuffer buf = new StringBuffer();
      buf.append(nest.getId());
      buf.append("/");
      buf.append(fileName);
      String urlPath = buf.toString();

      /* Insert the downloaded document in the local nests document storage */
      UrlResource dr = new UrlResource("local://" + urlPath, dockey);
      nest.insert(dr);
      Log.println(600, "IR: " + dr);
      /* InsertGnutant phase to notify other nests that I have downloaded a document. */
      boolean insant = true;
      if (insant) {
        /*
         * Create insert request based on a url and keywords
         */
        String keyw = fileName;
        Object request = new InsantRequest("insant://" + urlPath, keyw);
        nest.request(null, request, null);
      }
    }
  }


  /**
   *  Upcall method used to notify this listener that the request 
   *  associated to it has finished to return resources.
   *  HEIN FIXME: How can we determine this?? ^^^
   */
  public void complete()
  {
  }
  
} // END GnutantScenario
