/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.util.*;
import anthill.storage.*;

import java.io.*;
import java.util.*;

/**
 *  Ant implementation for inserting a document into the anthill network.
 *  An insertion involves a user make a document available in her upload
 *  repository and sending out a number of ants to make the anthill network
 *  aware of the document. The user must supply a couple of keywords that
 *  describes the document, and this will generate one ant for each supplied
 *  keyword. Each of these ants are routed along a single trail, leaving
 *  information about the document at each nest along the trail. Routing
 *  will follow the lexicographical closeness of the keywords.
 *
 *  The Insant is of the Gnutant species.
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class Insant
implements Ant, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Variable path contains the list of visited nodes, allowing the ant
   *  to avoid visiting the same nests multiple times.
   */
  private ArrayList path;

  /**
   *  The <code>KeyHash</code> object associated with this InsertGnutant.
   */
  private KeyHash keyhash;

  /**
   *  The URL associated with this InsertGnutant.
   */
  private UrlResource url;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public Insant()
  {
    keyhash = new KeyHash();
    url = new UrlResource();
    path = new ArrayList();
  }

  /**
   *  Instantiate a new Gnutant by setting the query to be satisfied and the
   *  hash value to be searched.
   */
  public Insant(KeyHash keyhash, UrlResource url)
  {
    this.keyhash = keyhash;
    this.url = url;
    /* We could initialize the path length to the ttl value... */
    path = new ArrayList();
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Ant
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  // Documentation inherited from interface
  public String getSpecies()
  {
    return "Gnutant";
  }

  /**
   *  Initialize the InsertGnutant at the originating nest.
   *
   *  @param view  an <code>AntView</code> value
   */
  public void init(AntView view)
  {
  }

  /**
   *  This method encapsulate the computation performed by the ant.
   *  This ant implementation will route itself along a path with
   *  identical or similar (closeness) hashed keywords. At each nest
   *  along this path, it will insert a URL reference to the document
   *  uploaded at the originating nest.
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
    path.add(view.getNestId());

    /*
     * Add the URL of this ant to the local URL storage.
     * The local URL storage avoid storing duplicate URLs,
     * while also taking care of storing multiple URL references
     * to the same document (same DocumentKey, but different URL).
     */
    UrlStorage urlStore = (UrlStorage) view.getStorage(Storage.URL_STORAGE);
    try {
      urlStore.addResource(url.getKey(), url);
    } catch (CapacityException e) {
      /* Commit suicide for now; could move on to the next nest instead. */
      System.out.println("Ciao" + url);
      Util.panic(e);
    }

    /* Update the local pheromone in this nest. */
    GnutantPheromone gnutPhero = (GnutantPheromone) view.getPheromone();
    if (gnutPhero == null) {
      gnutPhero = new GnutantPheromone(view);
      view.setPheromone((Object) gnutPhero);
    }

    /* Add the keyhash and the originating nest to the local pheromone table. */
    gnutPhero.addKeyhash(keyhash, (NestId) path.get(0));

    /* 
     * Get the a set of potential next nests --
     * based on the existance of the given keyhash
     * or according to the closeness relation to the keyhash.
     */
    ArrayList nxtList = gnutPhero.getNextNest(keyhash, path);

    boolean notMoved = true;
    for (int i = 0; notMoved && (i < nxtList.size()); i++) {
      try {
        view.move((NestId) nxtList.get(i));
        /* Successful move */
        return;
      } catch (IOException e) {
        /* Continue trying to move to an alternative nest. */
        continue;
      }
    } // end of for
    
    if (notMoved) {
      Log.println(1000, "NOT MOVED!");
      /* FIXME We should goBack() or just commit suicide. */
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Describe <code>readExternal</code> method here.
   *
   * @param in an <code>ObjectInput</code> value
   * @exception IOException if an error occurs
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    int size = in.readInt();
    for (int i = 0; i < size; i++) 
      path.add(in.readObject());
    keyhash.readExternal(in);
    url.readExternal(in);
  }

  /**
   * Describe <code>writeExternal</code> method here.
   *
   * @param out an <code>ObjectOutput</code> value
   * @exception IOException if an error occurs
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(path.size());
    for (int i = 0; i < path.size(); i++)
      out.writeObject(path.get(i));
    keyhash.writeExternal(out);
    url.writeExternal(out);
  }

} // END Gnutant


