/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.nest.*;
import anthill.nestor.*;
import anthill.factory.*;
import anthill.antsim.*;
import anthill.util.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  Class <CODE>ScenarioImpl</CODE> is a simple implementation of
 *  interface <CODE>Scenario</CODE> in which a predefined set of
 *  resources is inserted in the network and then a certain number
 *  of requests are performed and simulate. A <CODE>ScenarioImpl</CODE>
 *  is composed of a <CODE>Peernet</CODE> and a <CODE>ResourceSet</CODE> 
 *  object. Each resource included in the <CODE>ResourceSet</CODE> is inserted
 *  before starting the simulation in the <CODE>Peernet</CODE>. Then,
 *  the <CODE>ResourceSet</CODE> is used to generate a certain number
 *  of requests to be presented to the nests composing the network.
 *  
 *  
 *  @author Hein Meling
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class InsantScenario
implements Scenario
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  // NOTE: Are these used, or will they be used?

  /** 
   *  String name of the parameter used to determine the number of rounds which
   *  will be performed in this scenario.
   */
  public static final String SCENARIO_ROUNDS = "ScenarioRounds";

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Random number generator */
  private static Random random = new Random();
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;
  
  /** Number of rounds to be performed */
  protected int rounds;

  /** Stats */
  protected Stats stats;

  /** Peernet on which the genetic code will be evaluated */
  protected Peernet net;
  
  /** Version identifier of the peer network; when init params change, the version increase */
  protected int netVersion;
  
  /** Set of resources stored in the Anthill network */
  protected ResourceSet resources;
  
  /** Version identifier of the resource set; when init params change, the version increase */
  protected int rsetVersion;

  /** Reader for obtaining keywords */
  protected LineNumberReader lr;

  /** List of files to select randomly to compose fake URLs */
  protected ArrayList fileList;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor / reinitialization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new ScenarioImpl.
   *
   *  @param factory  the factory object used to generate the peernet
   *    and the resource set contained in a scenario
   *  @param rounds the number of rounds of the run
   */
  public InsantScenario(Factory factory, int rounds, String kwFile, String rootDir)
    throws FileNotFoundException, StreamCorruptedException, IOException, ClassNotFoundException
  {
    /* Initializes variables */
    this.factory = factory;
    this.rounds = rounds;
    FileReader fr = new FileReader(kwFile);
    lr = new LineNumberReader((Reader) fr);
    File cacheFile = new File(rootDir + File.separator + "InsantFileCache");
    if (cacheFile.exists() && cacheFile.canRead()) {
      /* The cache file exists read the file list from that instead of
       * recursing through the filesystem. */
      FileInputStream fis = new FileInputStream(cacheFile);
      ObjectInputStream ois = new ObjectInputStream(fis);
      fileList = (ArrayList) ois.readObject();
      fis.close();
    } else {
      File f = new File(rootDir);
      if (f == null)
        throw new FileNotFoundException("No root directory found.");
      fileList = getFiles(f, 3);

      /* Cache the file list for future runs of the scenario. */
      FileOutputStream fos = new FileOutputStream(cacheFile);
      ObjectOutputStream oos = new ObjectOutputStream(fos);
      oos.writeObject(fileList);
      oos.flush();
      fos.close();
    }

    /* Invokes clear*/
    clear();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Recursively obtain references to files in the filesystem based
   *  on the starting point <code>f</code>. We limit recursion by the
   *  the <code>depth</code> parameter.
   *
   * @param f the starting point in the filesystem.
   * @param depth used to limit recursion
   * @return an <code>ArrayList</code> of files
   */
  private ArrayList getFiles(File f, int depth)
  {
    if (f == null) 
      return new ArrayList();
    File[] fileList = f.listFiles();
    ArrayList list = new ArrayList();
    for (int i = 0; i < fileList.length; i++) {
      if (!fileList[i].canRead() || fileList[i].length() == 0)
        continue;
      if (fileList[i].isDirectory()) {
        if (depth > 0) {
          /* Only iterate recursively until depth reaches zero. */
          list.addAll(getFiles(fileList[i], depth-1));
        }
      } else {
        list.add(fileList[i]);
      }
    }
    return list;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the object to its initial state. The network is re-initialized
   *  as an empty network.
   */
  public void clear()
  {
    int version; 

    /*
     *  Read parameter ScenarioRounds from the factory object; if the parameter 
     *  is found, the value is updated. Otherwise, the initialization parameter
     *  is still valid.
     */
    try {
      rounds = factory.intValue(SCENARIO_ROUNDS);
    } catch (Exception e) {
    }
    
    /* 
     * Creates a new peernet iff (i) there is no peernet currently; 
     * (ii) the peernet initialization data have changed.
     * Otherwise, clear existing one.
     */
    version = factory.getVersion(Peernet.class);
    if (net == null || netVersion != version ) {
      netVersion = version;
      net = (Peernet) factory.get(Peernet.class);
    } else {
      net.clear();
    }
    
    /* 
     * Creates a new resource set iff (i) there is no resource set currently; 
     * (ii) the resource set initialization data have changed. 
     * Otherwise, clear existing one.
     */
    /*
    version = factory.getVersion(ResourceSet.class);
    if (resources == null || rsetVersion != version) {
      rsetVersion = version;
      resources = (ResourceSet) factory.get(ResourceSet.class);
    } else {
      resources.clear();
    }
    */

    /* Inserts generated resources in nests */
    Nest[] nests = net.getNests();

    /*    
    Resource[] rs = resources.getResources();
    for (int i=0; i < rs.length; i++) {
      int pos = random.nextInt(nests.length);  
      nests[pos].insert(rs[i]);
    }
    */
  }
   
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Scenario
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Set the statistics object which will be used throughout
   *  the scenario evaluation.
   */
  public void setStats(Stats stats)
  {
    this.stats = stats;
  }
  

  /**
   *  Perform the evaluation of this scenario object by simulating it.
   *  The simulation is subdivided in rounds. At each iteration,
   *  a new request is generated and assigned to one of the nest
   *  included in the simulated peer network, which generates an ant
   *  for it. Furthermore, for each  nest, an ant included in the 
   *  manager queue is selected and its <CODE>run()</CODE> method
   *  is executed. Statistics about the simulated execution of the
   *  code are collected in the <CODE>Stats</CODE> object specified
   *  through method <CODE>setStats()</CODE>.
   */
  public void evaluate()
  {
    // Obtain information about the network
    Nest[] nests = net.getNests();
    Manager[] managers = new Manager[nests.length];
    for (int i=0; i < nests.length; i++) {
      managers[i] = nests[i].getManager();
    }
    
    // Execute the required rounds 
    for (int i=0; i < rounds; i++) {
      if ((i / 100) * 100 == i)
        Log.println(500, "Round " + i);

      /* Choose random nest and make a request */
      int nest = random.nextInt(nests.length);
      /* Pick a file from the local filesystem at random */
      File f = (File) fileList.get(random.nextInt(fileList.size()));
      /* Get the next line of keywords from the keyword file */
      String keywords;
      try {
        keywords = lr.readLine();
      } catch (IOException e) {
        /* No more keyword lines in the keyword file, restart from the top. */
        lr.setLineNumber(0);
        /* FIXME   Should find better solution here... */
        keywords = "no more keywords";
      }
      /* Create insert request based on the url from the local filesystem,
       * and keywords from the gnutella keyword monitoring. */
      Object request = null;
      try {
        request = new InsantRequest(f.toURL(), keywords);
      } catch (IOException e) {
        Util.panic(e);
      }
      nests[nest].request(null, request, null);
      stats.inc("RequestedResources");
      Log.println(1000, "Searching: " + request);
      
      /* Each nest executes the code of an ant in its queue */
      for (int j=0; j < managers.length; j++)
        managers[j].execute();
    }
    /* Log the content of the pheromone table at each nest. */
    for (int i = 0; i < nests.length; i++) {
      GnutantPheromone gnutPhero = (GnutantPheromone) 
        ((LocalNest)nests[i]).getPheromone(GnutantPheromone.class);
      Log.println(600, "Pheromone at nest no. " + i);
      Log.println(600, gnutPhero);
    }
  }


  
} // END ScenarioImpl

