/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.util.*;
import java.io.*;
import java.util.*;

/**
 *  Provides a holder for keywords, read from an ASCII file with a set of
 *  keywords on each line.
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class Keywords
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator */
  private static Random random = new Random();

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** List of keywords to select from. */
  private ArrayList keywordList;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Loads the specified keyword file, entirely.
   *
   *  @param kwFile  the filename of the keyword file.
   */
  public Keywords(String kwFile)
    throws FileNotFoundException
  {
    /* We assume that the file does not contain more than MAX_VALUE lines. */
    this(kwFile, Integer.MAX_VALUE);
  }

  /**
   *  Loads the specified keyword file, limited by the number of keyword
   *  lines to load into the keyword list.
   *
   *  @param kwFile  the filename of the keyword file.
   *  @param lines   the number of lines to load; used to limit the
   *    number of keyword lines to load into the <code>keywordList</code>.
   */
  public Keywords(String kwFile, int lines)
    throws FileNotFoundException
  {
    BufferedReader br = new BufferedReader(new FileReader(kwFile));
    keywordList = new ArrayList();
    try {
      String keyline = null;
      /* Get the next line of keywords from the keyword file */
      while (((keyline = br.readLine()) != null) && (keywordList.size() < lines))
        keywordList.add(keyline);
    } catch (IOException e) {
      /* Serious failure */
      Util.panic(e);
    }
    //    Log.println("Read " + keywordList.size() + " lines from: " + kwFile);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the keyword stored in the given position.
   *  <P>
   *  Note that this method may return a <code>null</code> object
   *  if the index position is beyond that held within the keywordList.
   *
   *  @param index  position into the keywordList.
   *  @return a <code>String</code> containing a set of keywords
   *   associated with the given index position in the keywordList.
   */
  public String getKeyword(int index)
  {
    return (String) keywordList.get(index % keywordList.size());
  }

  /**
   *  Returns a random keyword from the keywordList.
   *
   *  @return a <code>String</code> containing a set of keywords.
   */
  public String getRndKeyword()
  {
    return getKeyword(random.nextInt(keywordList.size()));
  }

} // END Keywords
