/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.storage.*;
import anthill.util.*;
import java.io.*;
import java.util.*;

/**
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class SearchGnutant
implements Ant, Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Max number of times to backtrack before giving up. */
  private static final int MAXRESULTS = 4;

  /** Max number of times to backtrack before giving up. */
  private static final int BACKTRACK_LIMIT = 10;

  /** Backtracking on or off. */
  private static final boolean backtracking = true;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // State fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Variable path contains the list of visited nodes, allowing the ant
   *  to avoid visiting the same nests multiple times and backtrack to
   *  the originating nest.
   */
  private ArrayList path;

  /** The identifier of the request */
  private RequestId rid;

  /** The query string associated with this ant */
  private String query;

  /** The hash value of the keyword searched by this ant */
  private KeyHash keyhash;

  /** Resource to be returned */
  private HashSet resources;

  /** Number of backtracks performed so far by this ant. */
  private int backtracks;

  long rsize;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiate a new SearchGnutant by setting the query to be satisfied and the
   *  hash value to be searched.
   */
  public SearchGnutant(RequestId rid, String query, KeyHash keyhash)
  {
    /* Store initialization parameters */
    this.rid = rid;
    this.query = query;
    this.keyhash = keyhash;
    backtracks = 0;
    path = new ArrayList();
    resources = new HashSet();
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Ant
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  // Documentation inherited from interface
  public String getSpecies()
  {
    return "Gnutant";
  }

  
  /**
   *
   */
  public void init(AntView view)
  {
  }
  
  /**
   *  This method encapsulate the computation performed by the ant.
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
    path.add(view.getNestId());

    /*
     *  Obtains the local pheromone contained in this nest; if not found,
     *  creates a new pheromone object and installs it locally
     */
    GnutantPheromone pher = (GnutantPheromone) view.getPheromone();
    if (pher == null) {
      pher = new GnutantPheromone(view);
      view.setPheromone(pher);
    }

    /* Perform a local check of the storages */
    checkStorage(view);

    if (rsize > MAXRESULTS) {
      Resource[] res = new Resource[resources.size()];
      resources.toArray(res);
      view.addAnt(new BackwardGnutant(keyhash, res, path));
      return;
    }   
    
    /* 
     *  Get the a set of potential next nests --
     *  based on the existence of the given keyhash
     *  or according to the closeness relation to the keyhash 
     */
    ArrayList nxtList = pher.getNextNest(keyhash, path);
    for (int i = 0; i < nxtList.size(); i++) {
      try {
        view.move((NestId) nxtList.get(i));
        /* Successful move to new nest */
        return;
      } catch (IOException e) { 
        /* Try move to an alternative nest */
        continue;
      }
    } // end of for

    /*
     *  We could not move due to a dead end; backtracking
     */
    if (backtracking) {
      /*
       *  Note that we keep dead end nests in the path,
       *  since that simplifies the use of getNextNest().
       */
      int pathsize = path.size();
      if (pathsize > 1 && backtracks < BACKTRACK_LIMIT) {
        backtracks++;
        try {
          /*
           *  The last entry in path (pathsize-1) is the current nest
           *  because we just added it above. Therefore, we move the the
           *  nest before (pathsize-2) this one, and remove it from the
           *  path since it will be added again after the move.
           */
          view.move((NestId) path.remove(pathsize-2));
          /* Successful move to the previous nest */
          return;
        } catch (IOException e) { 
          /* Should never happen in the simulation */
        }
      }

      /*
       *  No more backtracking possible; commit suicide.
       */
      Log.print(400, "#");
    }
  }

  /**
   *  
   */
  private void checkStorage(AntView view)
  {
    QueryStorage docStorage = (QueryStorage) view.getStorage(Storage.DOC_STORAGE);
    QueryStorage urlStorage = (QueryStorage) view.getStorage(Storage.URL_STORAGE);

    /* Create an empty array list for the new resources */
    HashSet newResources = new HashSet();
    
    /* Add resources found in the doc storage */
    if (docStorage != null) {
      Resource[] docs = docStorage.getResources(query);
      if (docs != null) {
        for (int i=0; i < docs.length; i++) {
          if (resources.add(docs[i])) {
            newResources.add(docs[i]);
            rsize = rsize + docs[i].size();
          }
        }
      }
    }
    
    /* Add resources found in the url storage */
    if (urlStorage != null) {
      Resource[] urls = urlStorage.getResources(query);
      if (urls != null) {
        for (int i=0; i < urls.length; i++) {
          if (resources.add(urls[i])) {
            newResources.add(urls[i]);
            rsize = rsize + urls[i].size();
          }
        }
      }
    }

    /*  
     *  If there are new resources, send a ReplyGnutant.
     */
    if (newResources.size() > 0) {
      Resource[] news = new Resource[newResources.size()];
      newResources.toArray(news);

      /*
       *  Report the found resources directly to the originating nest.
       */
      ReplyGnutant rgant = new ReplyGnutant(rid, news, (NestId) path.get(0));
      view.addAnt(rgant);
    }

    /*
     *  Once the TTL have been exhausted we do a backward walk of
     *  the path in order to update the distributed index.
     */
    if (view.getTTL() == 0) {
      Resource[] res = new Resource[resources.size()];
      resources.toArray(res);
      /*
       *  BackwardGnutant is used to update the distributed index.
       *  Since we have reached the TTL we don't need to clone the
       *  path even though BackwardGnutant removes path entries as
       *  it backtracks.
       */
      view.addAnt(new BackwardGnutant(keyhash, res, path));
    }
  }

//         /*
//          *  BackwardGnutant is used to update the distributed index.
//          *  We need to clone the path since BackwardGnutant removes
//          *  path entries as it backtracks.
//          */
        //        view.addAnt(new BackwardGnutant(keyhash, news, (ArrayList) path.clone()));

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    // FIXME To be implemented
  }
  
  public void readExternal(ObjectInput in)
  {
    // FIXME To be implemented
  }
  
} // END SearchGnutant


