/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.nest.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  An <CODE>UrlResource</CODE> is an object containing the URL from which the
 *  content of a resource may be downloaded. When a <CODE>UrlResource</CODE>
 *  is inserted in a <CODE>Storage</CODE> object, the storage may perform a 
 *  local copy of the resource by downloading the contents through the
 *  input stream obtained through the <CODE>getInputStream()</CODE> method.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision$
 */
public class UrlResource
implements Resource, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** String identifier of the resource. */
  private String name;

  /** Url to this resource. */
  private String url;

  /** Probabilistic document key for the specified Url. */
  private DocumentKey dockey;

  /** Mapping between attribute names and attribute values */
  private Map map;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public UrlResource()
  {
    dockey = new DocumentKey();
    map = Collections.synchronizedMap(new HashMap());
  }

  /**
   *  Constructs a new <CODE>UrlResource</CODE> identified by the
   *  specified <code>name</code>. This will compute the hash and the length of
   *  the referenced file.
   *
   *  NON-SIMULATION VERSION
   *
   *  @param name identifier of the resource
   *  @param url  url from which the resource may be downloaded
   */
  public UrlResource(URL url)
    throws IOException
  {
    this.name = url.getFile();
    this.url = url.toString();
    dockey = new DocumentKey(url);
    map = Collections.synchronizedMap(new HashMap());
  }

  /**
   *  Constructs a new <CODE>UrlResource</CODE> identified by the
   *  specified <code>name</code>. This is for simulation and will generate a
   *  random document identifier for the referenced file.
   *
   *  SIMULATION VERSION.
   *
   *  @param url  url from which the resource may be downloaded
   */
  public UrlResource(String url)
  {
    this.url = url;
    this.name = getFile();
    /*
     * We use file part of the URL to compute the <code>DocumentKey</code>.
     * This is because using random DocumentKeys prevents clustering
     * in the same document in a UrlResourceSet (i.e., it would generate
     * different DocumentKeys for the same document.) Using the file part
     * of the url we can generate Ids deterministically.
     */
    dockey = new DocumentKey(name);
    map = Collections.synchronizedMap(new HashMap());
  }

  /**
   *  Constructs a new <CODE>UrlResource</CODE> identified by the
   *  specified <code>name</code> and <code>dockey</code>.
   *
   *  SIMULATION VERSION.
   *
   *  @param url  url from which the resource may be downloaded
   *  @param dockey  <code>DocumentKey</code> for the specified URL.
   */
  public UrlResource(String url, DocumentKey dockey)
  {
    this.url = url;
    this.name = getFile();
    this.dockey = dockey;
    map = Collections.synchronizedMap(new HashMap());
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the string identifier of the object.
   */
  public String getName()
  {
    return name;
  }


  /**
   *  Returns the value associated to the specified attribute
   *  name, or null if there is no attribute with that name.
   *
   *  @param name the name of the attribute
   */
  public String getAttribute(String name)
  {
    return (String) map.get(name);
  }


  /**
   *  Return an input stream to download the contents of the
   *  resource. Depending on the implementation, the contents
   *  may be read from the local storage, if the resource has
   *  been copied locally; or downloaded remotely, otherwise.
   *
   *  @return an input stream to read the contents of a resource;
   */
  public InputStream getInputStream()
    throws IOException
  {
    URL Url = new URL(url);
    return Url.openStream();
  }


  /**
   *  Return the lenght of the content of the resource, if available.
   *
   *  @return the lenght of the content of the resource
   */
  public long getContentLength()
  {
    return (long) dockey.getLength();
  }

  public long size()
  {
    return 1 + map.size();
  }

  /**
   *  Return the document key for this Url resource as the key.
   */
  public Object getKey()
  {
    return dockey;
  }

  /**
   *  Return the stringified url for this Url resource.
   */
  public String getURL()
  {
    return url;
  }

  /**
   * Returns the file part of this URL.
   *
   * @return a <code>String</code> value
   */
  public String getFile()
  {
    int index = url.lastIndexOf("/");
    return (index != -1) ? url.substring(index+1) : url;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Check whether this <CODE>UrlResource</CODE> is equal to the specified object.
   */
  public boolean equals(Object obj)
  {
    if (obj != null && obj instanceof UrlResource) {
      UrlResource urlRsrc = (UrlResource) obj;
      if (!dockey.equals(urlRsrc.dockey))
        return false;
      if (!name.equals(urlRsrc.name))
        return false;
      if (!url.equals(urlRsrc.url))
        return false;
      if (!map.equals(urlRsrc.map))
        return false;
      return true;
    } else {
      return false;
    }
  }

  /**
   *  Returns a hash code value for the object.
   */
  public int hashCode()
  {
    return dockey.hashCode();
  }
  
  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(": ");
    buf.append(name);
    buf.append("\n  ");
    buf.append(dockey);
    buf.append("\n  [URL: ");
    buf.append(url);
    buf.append("]\n]");
    return buf.toString();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Insert a mapping between the specified attribute
   *  name and the specified attribute value.
   *
   *  @param name the attribute name
   *  @param value the attribute value
   */
  void setAttribute(String name, String value)
  {
    map.put(name, value);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals this <CODE>UrlResource</CODE> to the specified output stream.
   */
  public void readExternal(ObjectInput in)
    throws IOException
  {
    // Read main fields
    name = in.readUTF();
    url = in.readUTF();
    dockey.readExternal(in);

    // Read attributes
    int size = in.readInt();
    for (int i=0; i < size; i++) 
      map.put(in.readUTF(), in.readUTF());
  }

    
  /**
   *  Marshals this <CODE>UrlResource</CODE> to the specified output stream.
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    // Write main fields
    out.writeUTF(name);
    out.writeUTF(url);
    dockey.writeExternal(out);

    // Write attributes
    out.writeInt(map.size());
    Iterator i = map.keySet().iterator();
    while (i.hasNext()) {
      String key = (String) i.next();
      out.writeUTF(key);
      out.writeUTF((String) map.get(key));
    }
  }

} // END UrlResource

