/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.nest.*;
import anthill.util.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class UrlResourceSet
implements Resource, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** String identifier of the resource. */
  private String name;

  /** Probabilistic document identifier common to all the URLs in the set. */
  private DocumentKey dockey;

  /** Set of URLs to this resource. */
  private Set urls;

  /** Mapping between attribute names and attribute values */
  private Map map;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public UrlResourceSet()
  {
    dockey = new DocumentKey();
    map = Collections.synchronizedMap(new HashMap());
  }


  /**
   *  Constructs a new <CODE>UrlResource</CODE> identified by the
   *  specified name. This will compute the hash and the length of
   *  the referenced file.
   *
   *  @param name identifier of the resource
   *  @param url  url from which the resource may be downloaded
   */
  UrlResourceSet(UrlResource url)
  {
    name = url.getName();
    dockey = (DocumentKey) url.getKey();
    urls = new HashSet();
    urls.add(url.getURL());
    map = Collections.synchronizedMap(new HashMap());
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the string identifier of the object.
   */
  public String getName()
  {
    return name;
  }


  /**
   *  Returns the value associated to the specified attribute
   *  name, or null if there is no attribute with that name.
   *
   *  @param name the name of the attribute
   */
  public String getAttribute(String name)
  {
    return (String) map.get(name);
  }


  /**
   *  Return an input stream to download the contents of the
   *  resource. Depending on the implementation, the contents
   *  may be read from the local storage, if the resource has
   *  been copied locally; or downloaded remotely, otherwise.
   *
   *  @return an input stream to read the contents of a resource;
   */
  public InputStream getInputStream()
    throws IOException
  {
    // FIXME  Should be smart here. Download from several of the urls specified in the set.
    // for now we just get the first one returned by the iterator.
    String url = null;
    Iterator iterator = urls.iterator();
    if (iterator.hasNext())
      url = (String) iterator.next();
    else
      throw new IOException("No URLs in the set.");
    URL Url = new URL(url);
    return Url.openStream();
  }


  /**
   *  Return the lenght of the content of the resource, if available.
   *
   *  @return the lenght of the content of the resource
   */
  public long getContentLength()
  {
    return (long) dockey.getLength();
  }


  public long size()
  {
    return urls.size() + map.size();
  }


  /**
   *  Return the document identifier for this Url resource as the key.
   */
  public Object getKey()
  {
    return dockey;
  }


  public boolean addURL(UrlResource url)
  {
    if (dockey.equals((DocumentKey) url.getKey()))
      urls.add(url.getURL());
    else
      throw new IllegalArgumentException(
        "The specified URL's identifier does not match this UrlResourceSet.");
    return true;
  }

  public Set getUrls()
  {
    return urls;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Check whether this <CODE>UrlResourceSet</CODE> is equal to the specified object.
   */
  public boolean equals(Object obj)
  {
    if (obj != null && obj instanceof UrlResourceSet) {
      UrlResourceSet urlSet = (UrlResourceSet) obj;
      if (!dockey.equals(urlSet.dockey))
        return false;
      if (!name.equals(urlSet.name))
        return false;
      if (!urls.equals(urlSet.urls))
        return false;
      if (!map.equals(urlSet.map))
        return false;
      return true;
    } else {
      return false;
    }
  }

  /**
   *  Returns a hash code value for the object.
   */
  public int hashCode()
  {
    return dockey.hashCode();
  }
  
  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(": ");
    buf.append(name);
    buf.append("\n  ");
    buf.append(dockey);
    buf.append("\n");
    Iterator iterator = urls.iterator();
    while (iterator.hasNext()) {
      buf.append("  URL: ");
      buf.append(iterator.next());
      buf.append("\n");
    }
    buf.append("]");
    return buf.toString();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Insert a mapping between the specified attribute
   *  name and the specified attribute value.
   *
   *  @param name the attribute name
   *  @param value the attribute value
   */
  void setAttribute(String name, String value)
  {
    map.put(name, value);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals this <CODE>UrlResourceSet</CODE> to the specified output stream.
   */
  public void readExternal(ObjectInput in)
    throws IOException
  {
    // Read main fields
    name = in.readUTF();
    dockey.readExternal(in);

    // Read URL set
    int size = in.readInt();
    for (int i = 0; i < size; i++) {
      urls.add(in.readUTF());
    }

    // Read attributes
    size = in.readInt();
    for (int i = 0; i < size; i++) 
      map.put(in.readUTF(), in.readUTF());
  }

    
  /**
   *  Marshals this <CODE>UrlResourceSet</CODE> to the specified output stream.
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    // Write main fields
    out.writeUTF(name);
    dockey.writeExternal(out);

    // Write URL set
    out.writeInt(urls.size());
    Iterator i = urls.iterator();
    while (i.hasNext()) {
      out.writeUTF((String)i.next());
    }

    // Write attributes
    out.writeInt(map.size());
    i = map.keySet().iterator();
    while (i.hasNext()) {
      String key = (String) i.next();
      out.writeUTF(key);
      out.writeUTF((String) map.get(key));
    }
  }

} // END UrlResource

