/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package gnutant;

import anthill.*;
import anthill.storage.*;
import anthill.storage.impl.*;
import anthill.factory.*;

import java.util.*;

/**
 *  Implementation of an Least Recently Used (LRU) based URL storage.
 *
 *  @author Hein Meling
 *  @version $Revision$
 */
public class UrlStorage
extends LRUQueryStorage
implements QueryStorage
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Contructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public UrlStorage(Factory factory, String storageName, String controllerName)
  {
    super(factory, storageName, controllerName);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Overriding public methods of the LRUQueryStorage
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Adds a resource to the storage and associates it to the specified
   *  key. If the remaining capacity of the storage is not sufficient to
   *  store the resource, a <CODE>CapacityException</CODE> is thrown
   *  containing information about the remaining capacity. Method
   *  <CODE>size()</CODE> may be invoked on the resource
   *  to determine the amount of space required by the resource. 
   *  In order to be stored on persistent storage, resource must 
   *  implement the <CODE>Serializable</CODE> interface.
   *
   *  This is a specialized implementation that will add URLs pointing
   *  to a file with the same <CODE>resourceKey</CODE> to a URL resource
   *  set before adding that URL set to the storage.
   *
   *  @param resourceKey  the key identifying the URL resource
   *  @param urlResource  the URL resource to be added
   *  @throws CapacityException  raised if the given resource object
   *    exceeds the total available capacity for this storage.
   */
  public void addResource(Object resourceKey, Resource urlResource)
    throws CapacityException
  {
    UrlResourceSet urlRsrcSet = null;
    if (list.containsKey(resourceKey)) {
      urlRsrcSet = (UrlResourceSet) getResource(resourceKey);
      urlRsrcSet.addURL((UrlResource) urlResource);
    } else {
      urlRsrcSet = new UrlResourceSet((UrlResource) urlResource);
    }
    /* We use the LRUQueryStorage addResource implementation, thus avoiding
     * to deal with capacity checks and LRU issues. */
    super.addResource(resourceKey, urlRsrcSet);
  }

  // HEIN FIXME: This should perhaps be implemented more efficiently.
  public Resource[] getResources(Object query) 
  {
    Resource[] urlResourceSet = super.getResources(query);
    ArrayList rsrcs = new ArrayList();
    for (int i = 0; i < urlResourceSet.length; i++) {
      DocumentKey key = (DocumentKey) urlResourceSet[i].getKey();
      Set urls = ((UrlResourceSet) urlResourceSet[i]).getUrls();
      Iterator iter = urls.iterator();
      while (iter.hasNext()) {
        UrlResource r = new UrlResource((String) iter.next(), key);
        rsrcs.add(r);
      }
    }
    Resource[] res = new Resource[rsrcs.size()];
    rsrcs.toArray(res);
    return res;
  }

} // END UrlStorage

