/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package random;

import anthill.*;
import anthill.util.*;
import anthill.storage.*;
import java.io.*;

/**
 *  This class contains the implementation of a simple ant which performs
 *  a random walk through the network. Ants move forward until they reach
 *  the searched resource; then, they move backward along the incoming
 *  path. When they reach the requestor nest, they copy the resource 
 *  locally.
 *
 *  @author Alberto Montresor
 *  @version $Revision$
 */
public class RandomAnt
implements Ant, Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static java.util.Random rand;
  static {
    rand = new java.util.Random();
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The identifier of the request */
  private RequestId rid;

  /** The identifier of this ant */
  private AntId aid;
  
  /** The query associated to this ant */
  private String query;
  
  /** The resource found */
  private Resource resource;

  /** 
   *  True if this ant is returning to the previous nest because it has reached
   *  a node already visited.
   */
  private boolean backtracking;
  
  /** True if this ant is returning to the requestor nest */
  private boolean forward;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiate a new RandomAnt by setting the query to be satisfied and the
   *  hash value to be searched.
   */
  public RandomAnt(RequestId rid, String query)
  {
    this.rid = rid;
    this.query = query;
    resource = null;
    aid = null;
    backtracking = false;
    forward = true;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Ant
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the identifier of the "species" of this ant.
   *  Different ant species may have access to different
   *  pheromone information in nests.
   */
  public String getSpecies()
  {
    return "RandomAnt";
  }

  public void init(AntView view)
  {
  }

  /**
   *  This method encapsulate the computation performed by the
   *  ant. 
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
    if (forward) {
      goForward(view);
    } else {
      goBackward(view);
    }
  }

  /**
   *  Algorithm used when moving forward
   */
  private void goForward(AntView view)
  {
    if (aid == null) {
      aid = view.createAntId();
    }
    if (backtracking) {
      /*  An ant is backtracking if it reaches a node which
       *  has already visisted and returns back to the previous
       *  node. In the case of RandomAnt, the backtracking operation
       *  cannot be longer than one step (each time we select a different
       *  node casually. 
       */
      backtracking = false;
    } else {
      /*  If the random ant is not backtracking, it has to check if it
       *  has already visited this node.
       */
      boolean firsttime = view.storeReturnNest(aid, view.getLastVisited());
      if (!firsttime) {
        /* We have already visited this nest */
        try {
          backtracking = true;
          view.move(view.getLastVisited());
        } catch (Exception e) {
          // Do nothing;
          return;
        }
        return;
      }
    }
    backtracking = false;
    
    /* Local query of the database */
    resource = view.getResource(query);
    if (resource != null) { 
      /* Found resource; go backward */
      forward = false;
      goBackward(view);
      return;
    } 

    /* Selection of the next hop */
    Neighbour[] neighbours = view.getNeighbours();
    int next = rand.nextInt(neighbours.length);
    try {
      // anthill.util.Log.println("Moving to " + neighbours[next].getId());
      view.move(neighbours[next].getId());
    } catch (Exception e) {
      e.printStackTrace();
      // Do nothing
    }
  }

  
  /**
   *  Algorithm used when moving backward
   */
  private void goBackward(AntView view)
  {
    /** Obtain the previous nest*/
    NestId prev = view.getReturnNest(aid, true);
    
    if (prev == view.getNestId()) {
      /** We have reached the requestor nest */
      view.returnResources(rid, new Resource[] { resource });
      try {
        view.addResource(resource);
      } catch (Exception e) { 
        e.printStackTrace(); 
      }
      /*
      GenericStorage storage = (GenericStorage) view.getStorage(Storage.DOC_STORAGE);
      if (storage != null) {
        try {
          storage.addResource(resource.getKey(), resource);
        } catch (Exception e) { 
          e.printStackTrace(); 
        }
      }
      */
    } else if (prev != null) {
      /** The ants move to the previous hop*/
      try {
        view.move(prev);
      } catch (Exception e) {
        // Do nothing
      }
    } else {
      /* We have lost the trail */
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    // FIXME To be implemented
  }
  
  public void readExternal(ObjectInput in)
  {
    // FIXME To be implemented
  }

 
} // END RandomAnt


